package HSPC::Mutex;

##--------------------------------------------------
## This module purpose is lock some actions.
## I.e lock simultaneous processing of one object
## in different apache childre, taskmanager and so on.
##-------------------------------------------------- 

use strict;
use SWsoft::protect;	# Mark this module for encryption upon build

use Time::HiRes;
use Error qw(:try);


use HSPC::Logger qw(sw_die);
use HSPC::Trace;
use HSPC::WebDB qw(:DEFAULT trans);

use HSPC::Exc::Mutex;

use constant DEFAULT_TIMEOUT         => 60;
use constant DB_NAME                 => 'MUTEX';
use constant ZOMBIE_CLEANUP_INTERVAL => 28;
use constant SLEEP_BETWEEN_TRIES     => 1000; ## milliseconds/

my $webdb;
my $last_cleanup_try = 0;

##--------------------------------------------------
## Usage: HSPC::Mutex->new(mutex=>"Proc doc #12", comment=>"start order saving")
## Description:	constructor, should be used only in trans.
## Parameters:
##     mutex   =>> mutex
##     comment => comment
##     timeout => timeout. If lock not obtained during timeout - die.
##     sleep_time => time in milliseconds to sleep between tries.
## Return value: object
sub new {
	my $class = shift;
	my %h = (
		mutex      => undef,
		comment    => undef,
		timeout    => undef,
		sleep_time => undef,
		@_
	);

	my $mutex   = $h{mutex}   || sw_die("Mutex not specified");
	my $timeout = $h{timeout} || &DEFAULT_TIMEOUT;
	my $comment = $h{comment} || 'No comment';
	my $sleep_time = $h{sleep_time} || &SLEEP_BETWEEN_TRIES;

	my $sysobj  = $ENV{'system_obj'};

	## create webdb object.
	$webdb ||= HSPC::WebDB->new( dbname => &DB_NAME );

	my $time_before_lock = time();
	my $time_after_lock;
	my $connection_id;

	{ ## make lock.
		my $mutex_quote = HSPC::WebDB::quote($mutex);
		my $comment_quote = HSPC::WebDB::quote($comment);

		my $lock_obtained = 0;
		while (1) {
			$connection_id = $webdb->connection_id;
			eval {
				my $rows = $webdb->select_run(
					qq{
						INSERT IGNORE INTO
							mutex
						(mutex, owner_cid, ref_count, timestamp, comment)
						VALUES 
						($mutex_quote, $connection_id, 1, unix_timestamp(), $comment_quote)
					},
				);

				if (!$rows) {
					$rows = $webdb->select_run(
						qq{
							UPDATE
								mutex
							SET
								owner_cid = $connection_id,
								ref_count = ref_count + 1,
								timestamp = unix_timestamp(),
								comment   = $comment_quote
							WHERE
								mutex = $mutex_quote
								AND ( owner_cid = $connection_id OR ref_count = 0 ) 
						}
					);
				}
				if ($rows) {
					$lock_obtained = 1;
				}
			}; ## eval
			$time_after_lock = time();
			if ($lock_obtained) {
				last; ## break while(1);
			} else {
				if ($timeout <= 0) {
					throw HSPC::Exc::Mutex( mutex => $mutex, timeout => 0 );
				} elsif($time_after_lock - $time_before_lock >= $timeout) {
					throw HSPC::Exc::Mutex( mutex => $mutex, timeout => $timeout );
				}

				## if we can't get lock try to make cleanup.
				## time_to_cleanup function guarantee, that only one of concurrent
				## waiting locks will try to cleanup.
				Time::HiRes::usleep( $sleep_time * 1000 ); 

				## real cleanup will execute not every time.
				$class->soft_cleanup_zombies();
			}
		}; ## while
	};

	my $time_to_obtain_lock = $time_after_lock - $time_before_lock;
	
	## prepare object.
	my $self = {};

	bless($self,$class);
	## initialize attributes of obj.
	$self->{mutex}               = $mutex;
	$self->{time_to_obtain_lock} = $time_to_obtain_lock;
	$self->{connection_id}       = $connection_id;
	## Flag indicating that mutex is occupied. Required for proper work of 'release' and 'DESTROY' method.
	$self->{'__mutex_occupied'} = 1;

	## on destroy of this object transaction will be rolled back automatically.
	return $self;
};

##--------------------------------------------------
## destructor
## it deref mutex in database.
sub DESTROY {
	my $self = shift;

	$self->release();
};

## Usage: $self->release()
## Description: 'manually' releases occupied mutex
## Parameters: none
## Return value: true on success, false otherwise 
sub release {
	my $self = shift;

	sw_die(__PACKAGE__." : invalid call to the object method")
		unless ref($self);

	if($self->{'mutex'} 
		&& $self->{'__mutex_occupied'}) 
	{
		my $mutex_quote = HSPC::WebDB::quote($self->{'mutex'});
		my $connection_id = $webdb->connection_id();
	
		sw_die("Mutex connection gone away during mutex lock ($self->{'connection_id'} => $connection_id)")
			if $connection_id != $self->{'connection_id'};

		$webdb->select_run(
			qq{
				UPDATE
					mutex
				SET
					ref_count = ref_count - 1,
					timestamp = unix_timestamp()
				WHERE
					mutex = $mutex_quote
					AND ( owner_cid = $connection_id )
			}
		);
		$self->{'__mutex_occupied'} = 0;
	}

	return !$self->{'__mutex_occupied'};	
}


## usage: HSPC::Mutex->cleanup_zombies
## It cleans up all mutexes, belongs
## to non-existent connectons
sub cleanup_zombies {
	my $class = shift;
	my %h = (
		@_
	);

	my $trace = HSPC::Trace->new("Mutex->cleanup_zombies");

	$webdb ||= HSPC::WebDB->new( dbname => &DB_NAME );

	$webdb->select_run("LOCK TABLE mutex WRITE");
	eval {
		my $processlist = $webdb->select_hashrows( "SHOW PROCESSLIST" ) || [];

		my @proc_id_list = map { $_->{Id} } @$processlist;
		my $proc_id_list_sql = join(',', @proc_id_list);

		$webdb->select_run("DELETE FROM mutex WHERE owner_cid NOT IN ($proc_id_list_sql)");
	};

	$webdb->select_run("UNLOCK TABLES");
};

## usage: HSPC::Mutex->soft_cleanup_zombies
## It cleans up all mutexes, belongs
## to non-existent connectons
## difference between cleanup_zombies and soft_cleanup_zombies
## is: soft_zombe_cleanup checks time of previous cleanup;
sub soft_cleanup_zombies {
	my $class = shift;

	if ( $class->time_to_cleanup() ) {
		$class->cleanup_zombies();
	}
};

## usage: HSPC::Mutex->time_to_cleanup
## This function atomically checks is time from 
## previous cleanup is long ehougth and reset this time
sub time_to_cleanup {
	my $class = shift;

	my $result = 0;
	my $current_time = time();
	if ($last_cleanup_try + &ZOMBIE_CLEANUP_INTERVAL < $current_time) {
		$last_cleanup_try = $current_time;

		$webdb ||= HSPC::WebDB->new( dbname => &DB_NAME );

		my $rows = 0;
		eval {
			$rows = $webdb->select_run(
				qq{
					INSERT IGNORE INTO
						mutex_cleanup
					(id, last_cleanup)
					VALUES
					(1, unix_timestamp())
				}
			);
			if ($rows == 0) {
				$rows = $webdb->select_run(
					qq{
						UPDATE
							mutex_cleanup
						SET
							last_cleanup = unix_timestamp()
						WHERE
							id = 1
							AND last_cleanup + ? < unix_timestamp()
					},
					&ZOMBIE_CLEANUP_INTERVAL
				);
			}
		};
		if ($rows > 0) {
			$result = 1;
		}
	}
	return $result;
};

## resets cleanup time to 0.
sub reset_cleanup_time {
	my $class = shift;
	$webdb ||= HSPC::WebDB->new( dbname => &DB_NAME );

	$webdb->select_run(
		qq{
			UPDATE
				mutex_cleanup
			SET
				last_cleanup = 0
			WHERE
				id = 1
		}
	);
};

1;
__END__
