package HSPC::WebPage;
use strict;
use HSPC::Config;
use HSPC::WebElements;
use HSPC::GeoData;

use HSPC::DataRequest;
use HSPC::DataCheck;

use HSPC::Util::String qw(sw_decode_base64 sw_escape_string sw_trim sw_encode_base64_safe);
use HSPC::Logger qw(sw_die);
use HSPC::Localization::Date qw(date_now date_to_str period_to_str
								time_gmt_to_local time_local_to_gmt time_to_date
								time_to_str split_datetime time_to_secs
								secs_to_time gmt_now);
use HSPC::WebSystemLib;
use HSPC::Localization qw(:DEFAULT string string_lang);
use HSPC::MT::Core::Constants;
##use HSPC::MT::Core::Web;
##use HSPC::MT::Core::Setup;
use SWsoft::protect;	# Mark this module for encryption upon build
use Spreadsheet::WriteExcel;
use POSIX qw(strtod);
use MIME::Base64;
use Storable qw(dclone);
use HSPC::ParamsValidate;
##use HSPC::MT::DM;
=head1 NAME

HSPC::WebPage - Object-oriented Interface for create HTML for
items lists and forms;

=head1 SYNOPSYS

	use HSPC::WebPage;
	use HSPC::MT::News;

	my $page = HSPC::WebPage->new({
		common_url=>qq{/office/news},
		res_itemstitle_id=>'news_articles',
		res_itemtitle_id=>'news_article'
	});


	my $request = $page->prepare_browse(cols=>[
			{title=>'ID',code=>'news_id',type=>'id',for_search=>1},
			{title=>'Title',code=>'title',type=>'q',for_search=>1},
			{title=>'Created',code=>'created_date',type=>'date'},
		]);

	HSPC::MT::News->get_articles(request=>$request);

	$page->header(page_title=>'News Setup');
	$page->title(title=>'Office|News');
	$page->search_bar();
	$page->browse(browse_info=>$request,item_url=>'/act,e');
	$page->footer();

	print $page->{edit_out};


=head1 NAME
	sub new()
		description : make instance of class HSPC::WebPage and init internal data
		agruments:
			common_url=> : string with base URL for current module, for example '/office/news'
			res_itemstitle_id=> :  one item title i18n_string_id, for example 'article'
			res_itemstitles_id=> :  some items title i18n_string_id for example 'articles'
		result: reference to new instance
=cut
# > h, with keys:
#	=> common_url
#	=> res_itemstitle_id
#	=> res_itemtitle_id
#   => lev : empty by default

sub new{
	my $self = {class=>shift,lev=>''};
	bless ($self,$self->{class});
	my $h = shift;
	#
	while( my ($k,$v) = each(%$h) ){
		$self->{$k}=$v;
	}
	#

	$self->{common_url} ||= $ENV{COMPREP} && $ENV{COMPREP}->link;

	my $sys = $ENV{system_obj};
	my $old_up_pc = $sys->upref_pc;
	$self->{pc} = sw_uriparam('pc'.$self->{lev})
		|| ( defined $sys->upref_pc ? $old_up_pc:20 );
	$self->{pc} = 20 unless HSPC::DataCheck::is_num($self->{pc})
		&& $self->{pc}<100 ;

	$sys->upref_pc( $self->{pc} ) if $self->{pc}!=$old_up_pc;

	$self->{from} = sw_uriparam('from'.$self->{lev});
	if($self->{from} ne '' && HSPC::DataCheck::is_num($self->{from}) ){
		$self->{url_from} = "/from$self->{lev},$self->{from}";
	}else{
		$self->{from} = 1;
		$self->{url_from} = undef;
	}

	$self->{got_sticky_search} = 0;
	$self->{list_url_params} = '';
	$self->{list_url_params} .= $self->{url_from};
	$self->{list_url_params} .= $self->{url_pc};

	$self->{sort_code} = sw_uriparam('sort'.$self->{lev});
	$self->{sord} = sw_uriparam('sord'.$self->{lev});

	## Sorting settings
	if ($self->{sort_code}){
		$sys->upref_by_alias(
			name => 'sort_code',
			value => $self->{sort_code},
			alias => $self->_upref_alias,
		);
		$sys->upref_by_alias(
			name => 'sord',
			value => $self->{sord} ? 1 : 0,
			alias => $self->_upref_alias,
		);
		$self->{columns_customized} = 1;
	} else {
		## reset columns order in upref
		if (sw_param('resetcols')){
			$sys->upref_by_alias(
				name => 'columns_order',
				alias => $self->_upref_alias,
				value => ''
			);
			$sys->upref_by_alias(
				name => 'sort_code',
				alias => $self->_upref_alias,
				value => ''
			);
			$sys->upref_by_alias(
				name => 'sord',
				alias => $self->_upref_alias,
				value => ''
			);
		} else {
			$self->{sort_code} = $sys->upref_by_alias(
				name => 'sort_code',
				alias => $self->_upref_alias,
			);
			$self->{sord} = $sys->upref_by_alias(
				name => 'sord',
				alias => $self->_upref_alias,
			);
			$self->{columns_customized} = 1
				if $self->{sort_code} || $self->{sord};
		}
	}

	#
	$self->{url_sort} = '';
	$self->{url_sort} .= "/sort$self->{lev},".$self->{sort_code}
		unless ($self->{sort_code} eq '');
	$self->{url_sord} .= "/sord$self->{lev},".$self->{sord} if $self->{sord};

	$self->{tab} .= sw_uriparam('tab');
	$self->{'__header_nav_path'} = [];
	$self->{'__tabs'} = [];

	return $self;
}

############################ Header part ########################################

#	>= page_title
# 	>= page_head
#	>= page_title_id
#	>= page_head_id
#	<= result
sub header
{
    my $self = shift;
    my %h = @_;
    $h{page_title} = string($h{page_title_id}) if (exists($h{page_title_id}));
    $h{page_head} = string($h{page_head_id}) if (exists($h{page_head_id}));

    $self->{edit_out} .= sw_web_html_header(%h);
}

## Usage: $self->header_nav_path([$@])
## Description: read/write accessor to header navigation path
## Paramaters:
##	path > ref. to array with navigation path, each element is ref. to hash
##		{
##			title_id => link title
##			url => link URL
##		}
## Return value: ref. to array with navigation path
sub header_nav_path {
	my $self = shift;

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	if(@_ && ref(@_[0]) eq 'ARRAY') {
		my $self->{'__header_nav_path'} = [];
		my $path = shift;
		foreach (@{$path}) {
			my %link = %{$_};
			push(@{$self->{'__header_nav_path'}},\%link);
		}
	}

	return [ @{$self->{'__header_nav_path'}} ];
}

## Usage: $self->add_header_nav_path(path=>)
## Description: add new header navigation path.
## Paramaters:
##	path =>> {
##		title_id => path link title
##		url => path link URL
##	}
##  force => if set then do not rewrite elements with equal URL part
## Return value: number of navigation path elements
sub add_header_nav_path {
	my $self = shift;
	my %h = (
		path  => undef,
		force => undef,
		@_,
	);

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	sw_die(__PACKAGE__." : path undefined or invalid")
		unless ref($h{'path'}) eq 'HASH';

	my $idx = -1;
	my %link = %{$h{'path'}};

	for(my $i=0;$i<=$#{@{$self->{'__header_nav_path'}}};$i++) {
		if(
			$self->{'__header_nav_path'}->[$i]->{'url'} eq $h{'path'}->{'url'}
			&& !$h{force}
		) {
			$self->{'__header_nav_path'}->[$i] = \%link;
			$idx = $i;
			last;
		}
	}
	unless($idx>=0) {
		push(@{$self->{'__header_nav_path'}},\%link);
		$idx = $#{@{$self->{'__header_nav_path'}}};
	}

	return $idx + 1;
}

## Usage: $self->add_tab(tab=>)
## Description: add new tab
## Paramaters:
##	tab =>> {
##		caption=> Text string with Page Title, f.e. "General Info"
##			OR
##		caption_id=> Text string with i18n ID for page title
##		page=> Page Code, uses for current page determinating and make URL for this page, f.e. 'gen'
##		url=> Full URL for this page
##		ext_url=> Ext. URL for this page, will be added to url+page
##		target=> url target
##	}
## Return value: index of latest tab, starting from 0
sub add_tab {
	my $self = shift;
	my %h = (
		'tab' => undef,
		@_,
	);

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	sw_die(__PACKAGE__." : path undefined or invalid")
		unless ref($h{'tab'}) eq 'HASH';

	my $idx = -1;
	my %tab = %{$h{'tab'}};

	push(@{$self->{'__tabs'}},\%tab);

	return 1 + $#{@{$self->{'__tabs'}}};
}

sub page_title
{
	my $self = shift;
	my %h = (
		alias  => undef,
		title  => undef,
		params => undef,
		@_,
	);

	sw_die ( 'WebPage->page_title: alias not defined' )
		unless $h{ alias };

	sw_die ( 'WebPage->page_title: title not defined' )
		unless $h{ title };

	my $node = $ENV{COMPREP}
			->node( alias => $h{ alias }, edit => 1 );

	$node->{ title_hash }->{ title } = $h{ title };

	if ( $h{params} && ref( $h{ params } ) eq 'HASH' ) {
	        foreach ( keys %{ $h{ params } } ) {
	                next unless $_;
	                next unless $h{ params }->{ $_ };
	                $node->{ uri_params }->{ $_ } = $h{ params }->{ $_ };
		}
	}
}

#  > class
#	=> title_id
#	=> title
#   =>> refresh_url
# comment_obj_id & comment_obj_type - both undef || both required
# => comment_obj_id
# => comment_obj_type
#   =>> devscreen_id : DevScreenID
#				for layout apps : app_id-page_id#field_id, look HelpManager doc
#				for cgi-scipts : unit_id-app_id-page_id#field_id
#   =>> icon :
#			file name for icon, example : '/pcc/images/i_commercemanager.gif'
#                      ask nomad@ about your icon
#          or
#   		icon_id from customization repository, ask nomad@ about your ID
#   =>  post_root : 1(default)=add "Top" link to path, 0=do not add "Top"
#	=> path : [{title=>,url=>},{title_id=>,url=>}...]
#	=> no_path (ignore path)
#   => pages - if passed, 'tab' is called immediately for these pages
#	<= result
sub title{
	my $self = shift;
	my %h = ( @_ );

	## postpone output for 'tab' call
	$self->{_title_called} = {
		idn_convert => 1,
		@_,
	};

	## in CP we should display onscreen help
	$self->{_devscreen_id} = $h{devscreen_id};

	$self->tab
		if defined $self->{_title_called}->{pages};
}

## > string_id
## < string - string with substituted params
sub _parse_status
{
	my $string_id = shift;
	my $string = string($string_id);
	$string =~ s/%([^%]+)%/sw_param($1)/eg;
	return $string
}## parsed_string

# > class
# => pages, ref to array of hashes
#	    caption=> Text string with Page Title, f.e. "General Info"
#          OR
#	    caption_id=> Text string with i18n ID for page title
#       page=> Page Code, uses for current page determinating and make URL for this page, f.e. 'gen'
#       url=> Full URL for this page
#       ext_url=> Ext. URL for this page, will be added to url+page
#		target=> url target
# => url, base url for all tabs, for
# => cur_page : current page code
# => descr, text for description();
# => no_statuses	- boolean, do not call statuses method;
sub tab{
	my $self = shift;
	my %h = (
		$self->{_title_called} ? %{$self->{_title_called}} : ( ),
		@_,
	);
	$h{url} = $self->get_browse_url().'/tab,' unless $h{url};
	$h{'pages'} ||= $self->{_title_called} && $self->{_title_called}->{pages}
		|| $self->{__tabs};

	my $node = my $node_orig = $ENV{COMPREP} && $ENV{COMPREP}->node;

	if ($node) {
	 	## try to locate node basing on 'tab' URI param, if any
		$node = $ENV{COMPREP}->node(choose => 1, edit => 1,
					alias => $node->{alias} . '_tab_' . $1)
				|| $node
			while $ENV{PATH_INFO} =~ m!/tab,([^/]+)!g;
	 	## try to locate node basing on 'act' URI param, if any
		$node = $ENV{COMPREP}->node(choose => 1, edit => 1,
					alias => $node->{alias} . '_act_' . $1)
				|| $node
			while $ENV{PATH_INFO} =~ m!/a(?:ct)?,([^/]+)!g;
		## inherit missing title hash parameters
		$node->{title_hash}->{$_} ||= $node_orig->{title_hash}->{$_}
			foreach keys %{$node_orig->{title_hash}};

		## convert tab URLs basing on CompRep
		if ($h{pages}) {
			## determine current tab basing on current node alias
			foreach (@{$h{pages}}) {
				$_ = {page => $_}
					unless ref $_;
				if ($node->{alias} =~ /^(.+)_tab_\Q$_->{page}\E/ && $ENV{COMPREP}->node(alias => $1)) {
					$node = $ENV{COMPREP}->node(alias => $1);
					$h{cur_page} = $_->{page};
				}
			}
			## set current tab and switch to tab parent node if current screen is in a tab
			foreach (@{$h{pages}}) {
				## try to locate tab in CompRep
				if (my $node = $ENV{COMPREP}->node(alias => "$node->{alias}_tab_$_->{page}")) {
					## remove tab URI parameter
					substr($_->{url}, index($_->{url}, 'tab,'), 4 + length($_->{page}), '')
						if index($_->{url}, 'tab,') >= $[;
					## set caption unless defined
					$_->{caption_id} ||= $node->{title_id};
					## generate absolute link
					$_->{url} = $ENV{COMPREP}->link(node_id => $node->{node_id})
						. $_->{url};
				}
			}
		}
		return sw_web_forbidden()
			unless $ENV{security_obj}->test_security($node->{alias});
	}

	if ($self->{_title_called}) {
		my ($path, $path_node) = ([], $node);
		unless ($node) {
			$path = $h{path} || $self->header_nav_path;
		} else {
			unless ($h{no_path}) {
				while ($path_node) {
					$path_node = $ENV{COMPREP}->node_parent($path_node);
					unshift @$path, {
						title	=> string($path_node->{title_id}, %{$path_node->{title_hash}}),
						url		=> $ENV{COMPREP}->link(
							node_id	=> $path_node->{node_id},
							uri_params	=> $path_node->{level} < 4
								? { menu_path => 1 }
								: undef
							),
					} unless !$path_node
						|| $path_node->{id} eq '00' && $path_node->{level} == 1
						|| $path_node->{filtered_out};
				}
			} else {
				$path = undef;
			}
		}

		$h{title} = string($node->{title_id}, %{$node->{title_hash}})
			if $node;

		if ($h{idn_convert}) {
			require HSPC::Util::DM qw(prepare_idn_value);
			foreach my $item (@$path) {
				$item->{title} = prepare_idn_value($item->{title})
					if defined $item->{title};
			}
			$h{title} = prepare_idn_value($h{title});
		}
		foreach my $item (@{$h{path}}) {
			if (defined $item->{title}) {
				$item->{title} = sw_escape_string($item->{title});
			}
		}
		$h{title} = sw_escape_string($h{title}) if (defined $h{title});

	    $self->{edit_out} .= HSPC::WebElements->title(
	    	%h,
	    	$node && (
				devscreen_id	=> $ENV{system_obj}->generate_screen_id,
		    	path			=> $path,
		    	title			=> $h{title},
		    	title_id		=> undef,
		    	icon			=> $node->{icon},
			),
		);

		## restore original node located using URI
		$ENV{COMPREP}->node(choose => 1, node_id => $node_orig->{node_id})
			if $node_orig;
	}

	$self->{edit_out}.=HSPC::WebElements->tab(%h);

	$self->description(description=>$h{descr}) if defined $h{descr};

	my $str;
	if( $ENV{layout} ){
		my $l = $ENV{layout};
		if($l->{uri_sl}){
			$self->add_status( string=>_parse_status($l->{uri_sl}) );
			$l->{uri_sl} = undef;
		}
		if($l->{uri_slx}){
			$self->add_status( string=>sw_decode_base64($l->{uri_slx}) );
			$l->{uri_slx} = undef;
		}
		if($l->{uri_sle}){
			## backward compatibility - old-fashioned errors, error code in sle uri
			## (still in use in SACP)
			if ($l->{uri_sle} =~ m|^\d+$|){
				$self->add_error(string_id=>$l->{uri_sle});
				$l->{uri_sle} = undef;
			}else{
				$self->add_error( string=>_parse_status($l->{uri_sle}));
				$l->{uri_sle} = undef;
			}
		}
		if($l->{uri_slerr}){
			$self->add_error( error_id=>$l->{uri_slerr} );
			$l->{uri_slerr} = undef;
		}
		if($l->{uri_slxe}){
			$self->add_error( string=>sw_decode_base64($l->{uri_slxe}));
			$l->{uri_slxe} = undef;
		}
	}else{
		if(sw_uriparam('err')){
			my @errors = sw_uriparam('err');
			$self->add_error(error_id => $_) for @errors;
		}
		## localized status messages
		$self->add_error( string=>_parse_status(sw_uriparam('sle')) )
			if sw_uriparam('sle');
		$self->add_status( string=>_parse_status(sw_uriparam('sl')) )
			if sw_uriparam('sl');

		## non-localized status messages
		$self->add_error(string=>sw_decode_base64(sw_uriparam('slxe')))
			if sw_uriparam('slxe');
		$self->add_status(string=>sw_decode_base64(sw_uriparam('slx')))
			if sw_uriparam('slx');
	}

	$h{no_statuses}	|| $self->statuses();

	### show custom description
	if(  length($self->{_devscreen_id}) ){
		$self->helpbar(position=>'top');	}
	### /show custom description
}

####### Store error in webpage ########
##  Added errors will be showed in tab() method or you can call statuses() for showing errors
##  at any time
##
## => string
##    OR
## => string_id
##    OR
## => error_id
sub add_error{
	my ($self,%h) = @_;

	if( defined $h{error_id} ){
		$self->{error_ids}.= $h{error_id}.',';
	}elsif( defined $h{string_id} ){
		$self->{error_string_ids}.= $h{string_id}.',';
	}else{
		sw_die('string, string_id or error_id should be provided for add_error')
			unless(defined($h{string}));
		#### convert all \n to <br>
		$h{string}=~s|\n|<br>|g unless $h{string}=~m|\<br\>|;
		$self->{errors}.= '<br>' if $self->{errors};
		$self->{errors}.= $h{string};
	}

	return 1;
}

# => string or => string_id # REQ_Y #
sub add_status{
	my ($self,%h) = @_;

	if( $h{string_id} ){
		$self->{statuses_id}.= $h{string_id}.',';
	}else{
		sw_die('string or string_id should be provided for add_status')
			unless(defined($h{string}));
		$self->{statuses}.= '<br>' if $self->{statuses};
		$self->{statuses}.= $h{string};
	}
}

###### check errors existing
## < 1=errors exists
sub has_errors{
	my $self = shift;
	return $self->errors ne '';
}

sub errors{
	my $self = shift;
	return ($self->{errors} ne '') || ($self->{error_string_ids} ne '') || ($self->{error_ids} ne '');
}

# RESET ACCUMULATED ERRORS #
# >> self
sub reset_errors{
	my $self = shift;
	$self->{errors} = undef;
	$self->{error_string_ids} = undef;
	$self->{error_ids} = undef;
}

# POST ACCUMULATED ERRORS & STATUSES AND RESET IT #
# >> self
sub statuses{
	my $self = shift;

	$self->{edit_out}.= HSPC::WebElements->status_message(string=>$self->{statuses}) if $self->{statuses};
	$self->{edit_out}.= HSPC::WebElements->status_message(string_id=>$self->{statuses_id}) if $self->{statuses_id};
	$self->{edit_out}.= HSPC::WebElements->status_message(string=>$self->{errors},is_warning=>1) if $self->{errors};
	$self->{edit_out}.= HSPC::WebElements->status_message(string_id=>$self->{error_string_ids},is_warning=>1) if $self->{error_string_ids};
	$self->{edit_out}.= HSPC::WebElements->status_message(error_id=>$self->{error_ids},is_warning=>1) if $self->{error_ids};

	$self->{statuses} = undef;
	$self->{statuses_id} = undef;
	$self->{errors} = undef;
	$self->{error_string_ids} = undef;
	$self->{error_ids} = undef;
}

############################ /Header part #########################################

############################ Description part #####################################

## >> self
## =>> description
sub description
{
	my $self = shift;
	my %args = (
	    @_
    );
 	return unless $args{description};

	$self->{edit_out} .= HSPC::WebElements->page_description(
		title => $args{description}
	);
}



## >> self
## => message_id
sub close_after_submit_js
{
	my $self = shift;
	my %h = @_;

	$self->edit_view_custom(
		t    =>'close_after_submit_js.tmpl',
		class=>__PACKAGE__,
		message_id=>$h{message_id}
	);

}##/close_after_submin_js


## >> self
## =>> position : help bar position. variants : 'top'
sub helpbar{
	my ($self,%h) = @_;
	sw_die('position=> undefined') unless defined $h{position};
	sw_die('position=> has wrong value') unless  $h{position} eq 'top';

	return if $self->{_helpbar_posted};

	## calc helpbar_location
	my $helpbar_location = $ENV{system_obj}->upref_var('helpbar_location'); ## Defined by user
	$helpbar_location = $ENV{system_obj}->provider_config()->helpbar_location
		if not defined $helpbar_location or $helpbar_location == -1; ## Defined by provider
	$helpbar_location = 1 unless defined $helpbar_location; ## Defined by default
	## /calc helpbar_location

	## is helpbar hidden ?
	return undef unless $helpbar_location;

	## show top helpbar
	if($h{position} eq 'top' && $helpbar_location==1){

		$self->{edit_out} .= HSPC::WebElements->page_description(
			helpbar_devscreen_id=>$self->{_devscreen_id}
		);
	}
	####
}

############################ /Description part ####################################

############################ Body part #########################################

sub search_bar{
	my $self = shift;
	my %h = ( self => $self, data => { self => $self }, show_br => '', @_ );
#	my %h=(self=>$self);			
	my ($search, $filter, $range, $freq);

	$h{title_res} = string($self->{res_itemstitle_id}) 
		if $self->{res_itemstitle_id};

	##$h{base_url}=$self->get_browse_url;
	$h{base_url}=$self->{common_url}.$self->{url_pc}.$self->{url_sort}.$self->{url_range}.$self->{url_frequency}.$self->{url_sord}.$h{ext_url};

	## remember for 'show all' button
	$h{all_url} = $self->{common_url}.'/reset_search,1';
	$h{got_sticky_search} = $self->{got_sticky_search};

	$h{devscreen_id} = $ENV{system_obj}->generate_screen_id;

	if($self->{browse_cols_search}){
		$h{name_bar}=string('search');
		$search = sw_web_parse_template(__PACKAGE__,'search_bar_s.tmpl',\%h);
	}
	if($self->{browse_cols_filter}){
		$h{name_bar}=string('filter');	
		$filter = sw_web_parse_template(__PACKAGE__,'search_bar_f.tmpl',\%h);	
	}	
	
	if($self->{browse_cols_range}){		
		$h{base_url}=$self->{common_url}.$self->{url_pc}.$self->{url_sort}.$self->{url_filter}.$self->{url_frequency}.$self->{url_sord}.$h{ext_url};
		$range = sw_web_parse_template(__PACKAGE__,'browse_ranges_bar.tmpl',\%h);
	}	

	if(ref $self->{browse_cols_frequency}){	
		$h{base_url}=$self->{common_url}.$self->{url_pc}.$self->{url_sort}.$self->{url_filter}.$self->{url_range}.$self->{url_sord}.$h{ext_url};
		$freq = sw_web_parse_template(__PACKAGE__,'browse_frequency_bar.tmpl',\%h);
	}

	if (($search ne '') || ($filter ne '') || ($range ne '') || ($freq ne '')) {
		$h{search}  = $search;
		$h{filter}  = $filter;
		$h{range}   = $range;
		$h{freq}	= $freq;
		$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'search_bar.tmpl',\%h);	
		$self->{edit_out} .= qq~<script language=JavaScript>\n function SelectAll() {return true} \n</script>~ unless $search;
	}
}

## maxwell@ :: "Add New Item" Top Action Bar
# Designed via icon+title -- instead of common button!
#
# => btns      # REQ_Y #   [{..btn hash....}]
# => url	   # REQ_N #   by default = page->get_browse_url()
# => form_type # REQ_N #  'post' default/ 'get'
#  -------------
#  btn hash 
#     	=> icon     : 'new_vps' || 'new_item'
#     	=> title    : 'Skip translation',
#	  	=> title_id : 'skip',
#		=> full_url # REQ_N # full form url
#		=> url_ext  # REQ_N # '/a,e', will be posted after url=>
#		=> target   # REQ_N #
#		=> class    # REQ_N # button class, default = 'SWIButton',
#		=> code     # REQ_N # custom button code, example : a=>qq{onClick="document.forms.item.item_submit=1;"} }]
#   => hidden : ref to hash with keys
#                      hidden_name=>hidden_value
#                      f.e. : {company_id=>1234,mode=>'simple'}
sub actions_bar{
	my $self  = shift;
	my %h = (
		form_type	=> 'post',
		form_name	=> 'new_item_form',
		devscreen_id=> $ENV{system_obj}->generate_screen_id,
		btn_name	=> '',
	@_);
	$h{url} = $self->get_browse_url() unless $h{url};
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'actions_bar.tmpl',\%h);
}


## maxwell@ :: checklist_top_actions
# New! 19-04-2006
# "Action bar for Group Operations"
# Designed via icon+title
# -- instead of common buttons at bottom of listing
# -- instead of separate frame (as for VPSs)!
#
# => btns      # REQ_Y #   [{..btn hash....}]
# btn hash 
#     	=> title     : 'Start',
#	  	=> title_id  : 'act_start', 
#       => name      : 'act_button_name'
#		=> jscode    # REQ_N # custom button code, example : jscode => qq|document.forms.item.item_submit=1;|
#		=> target    : '' / frame
#       => hidden    : ref to hash with keys
#                      hidden_name=>hidden_value
#                      f.e. : {company_id=>1234,mode=>'simple'}
#       => question  : "Are you shure ?"
#		=> chunk     :'Any custom HTML into <..btn..>
#		=> form_name : the name of form called for browse() -> $checkform_name
#		=> for_check : the ID of checked column called for browse() -> $colcode_for_check
sub checklist_top_actions {
	my $self  = shift;	
	my %h = (		
		form_name => 'item_id',
		for_check => 'id',
	@_);	
	$h{icon} = 'act_default' unless $h{icon};
	return sw_web_parse_template(__PACKAGE__,'checklist_top_actions.tmpl',\%h);
}

# draw top actions buttons on view screen
sub checklist_top_actions_view {
	my $self  = shift;	
	my %h = (		
		form_name => 'action_item',##default
		@_);	
	return sw_web_parse_template(__PACKAGE__,'checklist_top_actions_view.tmpl',\%h);
}

# generate paging bar and post it into self
# >>  self
# =>> request : ref to HSPC::DataRequest instance
# < content
sub paging_bar{
	my ($self,%h)=@_;
	die unless $h{request};

	# show top_paging #
	my $search_url = $self->{url_filter}.$self->{url_range}.$self->{url_frequency};
	return HSPC::WebElements->paging({
		url_pc	=> $self->{common_url}.$search_url.$self->{url_sort}.$self->{url_sord},
		url		=> $self->{common_url}.$search_url.$self->{url_sort}.$self->{url_sord}.$self->{url_pc}.$self->{url_tq}."/from$self->{lev},",
		title_res_id	=> $self->{res_itemstitle_id},
		class_app		=> $self->{class_app},
		current_item	=> $self->{from},
		total_items		=> $h{request}->{total_items},
		items_onpage	=> $self->{pc},
		url_index		=> $self->{lev},
		hide_columns_custom	=> $h{hide_columns_custom},
	});
}

#  >= cols : [] of hashes
#		=> code         # REQ_Y #  column code
#		=> title_id  	# REQ_N # string_id for column header
#		=> title  	    # REQ_N # text string for column header
#		=> type 	 	# REQ_Y # logic data type, variants:  # REQ
#			char = chars data, align by left, wrapped
#			id = integer data, not wrapped, align by right, sorted by default if no more columns with sort_def=1, for_itemurl=1,show_itemurl=1
#			int = integer data, not wrapped, align by right
#			q = string , align by left, wrapped, width=100%, show_itemurl=1
#			wide = string , align by left, wrapped, width=100%
#			options = strings array : [[value,title_id],[value,title_id]]
#			float = float data,not wrapped, align by right
#			bool = boolean data, align by center
#			live = bool, show red or green lamp
#			ip = IP Address, not wrapped, align by left
#           email = string, align by left + show as mailto link
#           emailtext = string, align by left, nowrap + don't!!! show as mailto link
#           country = string, align by left + filter_options = list of counties
#           state = string, align by left + filter_options = list of US states & Canada provincs
#			date = row date showed in localized text format
#			datetime = row datetime showed in localized text format
#			money = money ( in current version only $ )
#			percent  = percent
#			period  = period
#			unit  = integer data, not wrapped, align by right, with postfix 'U'
#		=> sort_def 	# REQ_N # 2/1/undef - this column will be sorted by default, 1=ASC,2=DESC
#       => align        # REQ_N # column align, variants : left/center/right
#		=> hide			# REQ_N # do not show this column on page
#		=> hide_default	# REQ_N # disable this column on page, but allow user to enable it
#		=> for_check 	# REQ_N # 1/undef  - use value from this column for first column with checkbox for group list operations
#		=> for_where 	# REQ_N # 1/undef  - check column data for value=>
##################
#		=> for_search 	# REQ_N # 1/undef - include this field in search_bar
#		=> search_title # REQ_N # Special Title for Search Box
##################
#		=> for_range  : ref to hash with range attributes.... support only 'id' and 'int col types
#			=> title  : range title
#			=> from : select from
#			=> to : select to
#			=> from_exc : exclude from_value
#			=> to_exc : exclude from_value
##################
#		=> for_frequency : frequency bar... () support only 'date' and 'datetime' col types
#			=> def_year_from - years interval (>=2001 by default)
#			=> def_year_to - years interval (<=2009 by default)
#			=> start_day_of_week : first day of week. default 0 (= sunday)
##################
#		=> for_filter 	# REQ_N # 1/undef - include this field in filter_bar
#               you can use additonal arguments with for_filter=>1
#		=> no_auto_filter # REQ_N # 1/undef - disable auto-filtering by this column
#		=> filter_use_def : insert "All" item in the top of the options list
#		=> filter_options # REQ_N # [[value,title]..]
#		=> filter_localize # 0 by defaut; 1 - translate titles
#		=> filter_width # REQ_N # Filter bar width, in characters, default value=10 characters
#		=> filter_script # REQ_N # Java-script code, calling in onChange
#		=> filter_title # REQ_N # Special Title for Filter Box
##################
#		=> colors		# REQ_N #  hash with value<>color_type links, format
#					{_def_color => color, value1=>color,value2=>color},  we have four color_types: 0=Black,1=Green,2=Red,3=Yellow, _def_color - color for defaut value
#		=> show_itemurl # REQ_N #  1/undef - show as link to item editing
#		=> for_itemurl 	# REQ_N #  1/undef - includes in link to item editing
#		=> itemurl_param_name 	# REQ_N #  parameter name will be set in link
								# if for_item_url is defined
#		=> data_options # REQ_N # [[value,title]..]      #TODO : podumat ob izpolzovanii v kachesteve filter_options pri nezapolnenosti poslednih
#		=> no_select	# REQ_N #  this column will not be present in request_data
#		=> no_wrap 		# REQ_N #
#		=> no_sort		# REQ_N #  this column can`t be sorting
#		=> zeroempty    # REQ_N # show zero as empty place, uses for "money" and "int"
#		=> int_nums		# REQ_N # numbers count after decimal point, for type "int"
#		=> search_width # REQ_N # Search bar width, in characters, default value=10 characters
#		=> value		# REQ_N # Filter by value, can be ref to ARRAY
#		=> link_base    # REQ_N # Show col values as link to any page. Example : '/office/company/id,'
#		=> link_attr    # REQ_N # Attribute Name, which value will be appended to every link_base=>
#		=> link_cond_attr # REQ_N # Link condition attribute name, which defines
#							show link or not. Note: should be integer because if it boolean
#							then its value will be changed to image and cannot be checked
#							properly
#		=> default		# REQ_N # Default value, i case if value is invalid (e.g. invalid date)
#		=> enum_titles  # REQ_N #
#                        ref to hash {cell_value1=>title1,cell_value2=>title2}
#		=> precision	# precision
#		=> icon         # (picture id for icon)
#		=> localize_content # 0 by default, 1 means string_id is expected
#		=> filter_localize	# 0 by default, 1 means string_id is expected
#		=> data_localize	# 0 by default, 1 means string_id is expected
#		=> item_url_format       : ref to hash with link definition, format
#				=> url_extension : An text extension that will be added to common_url.
#					Ex: '/tab,default'
#				=> full_url      : The beginning of url that will replace common_url value.
#					Ex: /pcc/site.cgi/bld/rcv/docs/a,ar_doc_list_screen
#				=> url_params    : Required. ref to array of hashes
#					=> param_col_code : column code from which value of parameter will be taken
#					=> param_value    : value of parameter can be provided directly from definition
#					=> param_name     : name of parameter. Should be specified if differes from param_col_code
#					=> mandatory      : do not place link if value of this column is false
#			}
##  =>not_escape if true value will not be escaped for html
#
# < info_hash {
#		sort_col=>
#		sort_revers=>
#		from=>
#		count=>
#		filters=> {code=>value,code1=>value1}
#		cols=>
#	}
sub prepare_browse{
	my $self = shift;
	my %h = (@_ );

	$self->{browse_cols} = $h{cols} || sw_die('cols param missing');

	undef $self->{url_filter};
	undef $self->{url_range};
	undef $self->{browse_cols_search};
	undef $self->{url_frequency};
	undef $self->{columns_customized};
	# handle columns #
	my $browse_col_sort_def;
	my $browse_col_sort;
	my $browse_cols_filter;
	my $browse_cols_range;
	my $browse_cols_frequency;
	my $browse_cols_search;
	my $browse_cols_filters;
	my $colcode_for_check;
	my $col_id;
	my $index=0;

	## Show all pages at once
	if (sw_uriparam('excel')){
		$self->{from} = 1;
		$self->{pc} = 0x6FFFFFFF;
	}
	foreach my $col (@{$h{cols}}){
		sw_die("col should be hash ref") unless ( ref($col) eq 'HASH' );
		$browse_col_sort_def = $col if $col->{sort_def};
		$col_id = $col if $col->{type} eq 'id';
		$browse_col_sort = $col if $self->{sort_code} && $col->{code} eq $self->{sort_code};
		$colcode_for_check = $col->{code} if $col->{for_check};

		$col->{title} .= string( $col->{title_id} ) if $col->{title_id};

		# init modes (align,no_wrap,width) for type#
		$col->{align} = " align=$col->{align} " if defined $col->{align};

		if($col->{type}  eq 'id'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'int'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'money'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'percent'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'period'){
			$col->{align} = ' align=left ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'q'){
			$col->{align} = ' align=left ' unless defined $col->{align};
			$col->{width} = ' width = 100% ';
		}elsif($col->{type}  eq 'wide'){
			$col->{align} = ' align=left ' unless defined $col->{align};
			$col->{width} = ' width = 100% ';
		}elsif($col->{type}  eq 'date'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'datetime'){
			$col->{align} = ' align=right ' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'unit'){
			$col->{align} = ' align=right ' unless defined $col->{align};
		}elsif($col->{type}  eq 'emailtext'){
			$col->{align} = ' align=left' unless defined $col->{align};
			$col->{no_wrap} =  1;
		}elsif($col->{type}  eq 'live'){
			$col->{align} = ' align=center ' unless defined $col->{align};
		}elsif($col->{type}  eq 'icon'){
			$col->{align} = ' align=center ' unless defined $col->{align};
		}elsif($col->{type}  eq 'ip'){
			$col->{align} ||= ' align=left ';
			$col->{no_wrap} ||=  1;
		}elsif($col->{type}  eq 'bool'){
			$col->{align} = ' align=center ' unless defined $col->{align};
		}

		## Localize options in filter and data
		if($col->{filter_localize}){
			## Copy options
			my @options;
			my $i = 0;
			foreach my $option (@{$col->{filter_options}}){
				$options[$i]->[0] = $option->[0];
				$options[$i]->[1] = string($option->[1]);
				$i++
			}
			$col->{filter_options} = \@options;
		}
		if($col->{data_localize}){
			## Copy options
			my @options;
			my $i = 0;
			foreach my $option (@{$col->{data_options}}){
				$options[$i]->[0] = $option->[0];
				$options[$i]->[1] = string($option->[1]);
				$i++
			}
			$col->{data_options} = \@options;
		}


		# set for_itemurl #
		$col->{show_itemurl} = 1 if (not defined $col->{show_itemurl}) && ($col->{type} eq 'id' || $col->{type} eq 'q');
		$col->{for_itemurl} = 1 if $col->{type} eq 'id';

		# default filter options for country types
		if ($col->{type} eq 'country' && $col->{for_filter} && ! $col->{filter_options}) {
			# pre-fill with countries
			$col->{filter_options} = [['', string('all')], @{HSPC::GeoData->get_countries()}];
		}

		# default filter options for state types
		if ($col->{type} eq 'state' && $col->{for_filter} && ! $col->{filter_options}) {
			# pre-fill with states
			$col->{filter_options} = [@{HSPC::GeoData->get_states()}];
			$col->{filter_options}->[0]->[1] = string('all');
		}

		$col->{chunk_wrap} = $col->{no_wrap}?' nowrap ':'';


		# check filter_flags #
		if($col->{for_filter} || $col->{for_search}){
			if($col->{type} ne 'date'){
			#### usual filter
				my $code_f = $col->{code}.$self->{lev}.'_f';

				my $form_value = $self->sticky_param (
					arg_name	=> $code_f,
					is_undef	=> 1
				);

				## if we need to set/replace value for filter
				if((defined $form_value && ref $col->{value} eq 'ARRAY' &&
					scalar grep { $_ eq $form_value } @{$col->{value}}) || !defined $col->{value}) {
					$col->{value} = $form_value;
					$self->{url_filter} .= "/$col->{code}$self->{lev}_f,".
						$self->convert_quotes($form_value) if length($form_value);
				}
				push(@$browse_cols_filter,$col) if $col->{for_filter};
				push(@$browse_cols_search,$col) if $col->{for_search};
				push(@$browse_cols_filters,$col);

			}else{
			#### date filter		
				$col->{value} = $self->sticky_param(
					uri_name	=> $col->{code}.$self->{lev}.'_f',
					arg_name	=> ('datebar_filter', 'date'),
					secs		=> 1,
				);

				$self->{url_filter}.="/$col->{code}$self->{lev}_f,".time_to_secs($col->{value})
					if defined($col->{value}) && $col->{value} ne '';

				push(@$browse_cols_filter,$col) if $col->{for_filter};
				push(@$browse_cols_filters,$col);

			}

		}

		# check range_flags #
		if($col->{for_range}){
			if($col->{type} ne 'datetime' && $col->{type} ne 'date'){
				$col->{value_from} = $self->sticky_param(
					arg_name	=> $col->{code}.$self->{lev}.'_r_from',
				);
				$col->{value_from} = $col->{for_range}->{from} if $col->{value_from} eq '';
				$col->{value_to} = $self->sticky_param(
					arg_name	=> $col->{code}.$self->{lev}.'_r_to',
				);
				$col->{value_to} = $col->{for_range}->{to} if $col->{value_to} eq '';

				$self->{url_range}.="/$col->{code}$self->{lev}_r_from,".$col->{value_from} if defined($col->{value_from}) && $col->{value_from} ne '';
				$self->{url_range}.="/$col->{code}$self->{lev}_r_to,".$col->{value_to} if defined($col->{value_to}) && $col->{value_to} ne '';
			}else{
				# big ugly date range bar
				$col->{for_range} = {} unless ref($col->{for_range});

				$col->{value_from} = $self->sticky_param(
					uri_name	=> $col->{code}.$self->{lev}.'_r_from',
					arg_name	=> ['datebar_from', 'datebar'],
					secs		=> 1,
				) || $col->{for_range}->{from};

				$col->{value_to} = $self->sticky_param(
					uri_name	=> $col->{code}.$self->{lev}.'_r_to',
					arg_name	=> ['datebar_to', 'datebar'],
					secs		=> 1,
				) || $col->{for_range}->{to};

				if ( !defined($col->{reset_value_to}) && !defined($col->{value_to})){
					if ($col->{type} eq 'date') {
						$col->{value_to} = time_local_to_gmt(time_to_date(time_gmt_to_local(gmt_now())));
					} else {
						$col->{value_to} = gmt_now();
					}
				};
				$self->{url_range}.="/$col->{code}$self->{lev}_r_from,".time_to_secs($col->{value_from}) if defined($col->{value_from}) && $col->{value_from} ne '';
				$self->{url_range}.="/$col->{code}$self->{lev}_r_to,".time_to_secs($col->{value_to}) if defined($col->{value_to}) && $col->{value_to} ne '';
			}

			push( @$browse_cols_range,$col) if $col->{for_range};
			push( @$browse_cols_filters,$col);

		}

		# check frequency_flags #
		if($col->{for_frequency}){
			unless(ref($col->{for_frequency})) { $col->{for_frequency}={}; }
			## set >= and <= in sql query
			for ('to_exc','from_exc') { $col->{for_frequency}->{$_} = 1; }

			$col->{for_frequency}->{def_type} ||= 0;
			$col->{for_frequency}->{def_time_from} = date_now()." 00:00:00";
			$col->{for_frequency}->{def_time_to} = date_now()." 23:59:59";
			$col->{for_frequency}->{start_day_of_week} ||= 0;

			## calculate from date
			my $uri_from = sw_uriparam($col->{code}.$self->{lev}.'_freq_from');
			$uri_from = secs_to_time($uri_from) unless $uri_from eq '';
			$col->{value_from} = sw_argparam('frequency_type_from');
			if($col->{value_from}) {
				$col->{value_from} = time_local_to_gmt($col->{value_from});
			}
			else { $col->{value_from} = $uri_from;}
			$col->{value_from} ||= time_local_to_gmt($col->{for_frequency}->{def_time_from});


			## calculate to date
			my $uri_to = sw_uriparam($col->{code}.$self->{lev}.'_freq_to');
			$uri_to = secs_to_time($uri_to) unless $uri_to eq '';
			$col->{value_to} = sw_argparam('frequency_type_to');
			if($col->{value_to}) {
				$col->{value_to} = time_local_to_gmt($col->{value_to});
			}
			else { $col->{value_to} = $uri_to;}
			$col->{value_to} ||= time_local_to_gmt($col->{for_frequency}->{def_time_to});

			## calculate type (weekly or daily or ...)
			my $uri_type = sw_uriparam($col->{code}.$self->{lev}.'_freq_type');
			$col->{for_frequency}->{type} = sw_argparam('frequency_bar_type');
			unless(length($col->{for_frequency}->{type})) {
				$col->{for_frequency}->{type} = $uri_type;
			}
			$col->{for_frequency}->{type} = $col->{for_frequency}->{def_type} unless (length($col->{for_frequency}->{type}) > 0);

			## store url
			$self->{url_frequency}.="/$col->{code}$self->{lev}_freq_from,".time_to_secs($col->{value_from}) if defined($col->{value_from}) && $col->{value_from} ne '';
			$self->{url_frequency}.="/$col->{code}$self->{lev}_freq_to,".time_to_secs($col->{value_to}) if defined($col->{value_to}) && $col->{value_to} ne '';
			$self->{url_frequency}.="/$col->{code}$self->{lev}_freq_type,".$col->{for_frequency}->{type} if defined($col->{for_frequency}->{type}) && $col->{for_frequency}->{type} ne '';

			push( @$browse_cols_frequency,$col);
			push( @$browse_cols_filters,$col);

		}

		# this column will be presents in data #
		if(!$col->{no_select}){
			$col->{index} = $index;
			$index++;

			# Needed for state type of column
			$self->{country_col_index} = $index if $col->{type} eq 'country';
		}
	}

	# prepare current sort
	if(not $browse_col_sort){
		if($browse_col_sort_def && !$browse_col_sort && $browse_col_sort_def->{sort_def}==2){
			$self->{sord} = 1;
		}
		$browse_col_sort = $browse_col_sort_def if $browse_col_sort_def;
		$browse_col_sort = $col_id if  (not $browse_col_sort_def) && $col_id;
	}

	# quote url
	$self->{url_frequency}=~s|%|\%25|g;
	$self->{url_range}=~s|%|\%25|g;
	$self->{url_filter}=~s|%|\%25|g;


	# store vars #
	$self->{browse_col_sort} = $browse_col_sort;
	$self->{browse_colcode_for_check} = $colcode_for_check;
	$self->{browse_cols} = $h{cols};
	$self->{browse_cols_filter} = $browse_cols_filter;
	$self->{browse_cols_range} = $browse_cols_range;
	$self->{browse_cols_frequency} = $browse_cols_frequency;
	$self->{browse_cols_search} = $browse_cols_search;
	$ENV{system_obj}->store_upref_by_alias( $self->_upref_alias );

	# return browse_info
	my $info=HSPC::DataRequest->new(
		sort_col_code => $self->{browse_col_sort}?$self->{browse_col_sort}->{code}:'',
		sort_revers   => $self->{sord},
		from          => $self->{from},
		count         => $self->{pc},
		filters       => $browse_cols_filters,
		cols          => $self->{browse_cols}
	);

	#### request tuning
	foreach my $col ( @{ $info->{cols} } ){
		if( $col->{type} eq 'state'  ){
			unless( $info->col('state_alt') ){
				$info->add_col(code=>'state_alt',type=>'char',hide=>1);
			}
		}
	}

	return $info;
};

#	<= result
sub footer{
	my $self = shift;

	$self->{edit_out} .= sw_web_html_footer(@_);
}

# > self
# >= string_id
# >= string
# >= is_warning (1,0)
sub status_message{
	my $self=shift;

	$self->{edit_out} .= HSPC::WebElements->status_message(@_);
}

# > self
sub get_browse_url{
	my $self=shift;
	return $self->{common_url}.$self->{url_pc}.$self->{url_from}.$self->{url_sort}.$self->{url_sord}.$self->{url_filter}.$self->{url_range}.$self->{url_frequency};
}

## $self->get_filter_url()
## Description: getter for page filter url
## Parameters: none
## Return value: filter portion of page browse url
sub get_filter_url {
	my $self = shift;

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	return $self->{'url_filter'} || '';
}

## $self->get_search_url()
## Description: getter for page search url
## Parameters: none
## Return value: search portion of page browse url
sub get_search_url {
	my $self = shift;

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	return $self->get_filter_url().$self->{'url_range'}.$self->{'url_frequency'};
}

# > self
# < resulted HTML content
sub get_content{
	my $self = shift;
	return $self->{edit_out};
}

# => subtitle_res
# => subtitle
sub edit_subtitle{
	my $self=shift;
	my $item = $self->{obj}->{item};
	#
	my %h=(
		self=>$self,
	@_);
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_subtitle.tmpl',\%h);
}

# => title_id
# => title
# => content : normal font text under the title
# => info_blocks[{}]
# =>	title  : text in the grey bar (i.e. 'Common Reasons', 'Possible Solutions')
# =>	items[..] : lines of bulleted list

sub post_info_text{
	my $self=shift;
	my %h=(
			self=>$self,
	    @_);

	$h{title} .= string($h{title_id}) if $h{title_id};

	## escaping HTML symbols
	unless ($h{not_escape} == 1) {
		##$h{content} = $self->($h{content});
		foreach my $block(  @{ $h{info_blocks}} ){
			##$block->{title} = $self->convert_quotes($block->{title});
			foreach my $item ( @{  $block->{items} } ){
				##$item = $self->convert_quotes($item);
			}
		}
	}

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'post_info_text.tmpl',\%h);
}

# => title_id
# => title
# => id
##=>tag_id item by default
# => view_url
# => disable_content : do not post any content in this method
# => disable_header
# => left_offset : push offset from left side, set zero to remove this offset
# => form_width=> x/undef : undef=100%
# => form_cols=> how much columns will be in form ; default = 2
# => url_ext  : additional URL part
# => form_url   # REQ_N #
# => target :  taget attribute of FORM tag, empty by default
# => security_policy  : read_only form if user does`t such permission , BTW: this check uses security_class from constructor
# => is_multipart : 1/0, default=0
# => is_transparent :1/0  default=0 (1 - no grey background, cells instead)
# => object #REQ_N# ref to object MT::Abstract, used for form data and limits
sub edit_open{
	my $self=shift;

	sw_die("ERROR: error in page->edit_open(), page must be in free mode, current mode is $self->{mode}") if defined $self->{mode};
	$self->{mode} = 'edit';

	my %h = (
		self                 => $self,
		form_warning_id      => $self->{form_warning_id},
		form_warning_comment => $self->{form_warning_comment},
		form_cols            => 2,
		left_offset          => 1,
		url_ext              => "/a$self->{lev},s",
		tag_id               => 'item',
		form_width           => 550,
		@_
	);
	if ($h{target}){
		$h{target}='target="'.$h{target}.'"';
	}
	$self->{form_tag_id}      = $h{tag_id};
	$self->{form_width_real}  = $h{form_width};

	## defence from dummies
	$h{form_cols} = 2 unless $h{form_cols};

	$self->{read_only} = 1 unless !$h{security_policy} ;
	$self->{_edit_form_cols} = $h{form_cols};
	$self->{is_form_transparent} = $h{is_transparent};
	#
	if((not defined $h{disable_content}) || ($h{disable_content}==0)){
		#
		$h{title} .= string($h{title_id}) if $h{title_id};
		$h{enctype}=q{enctype="multipart/form-data"} if $h{is_multipart};
		$self->{_form_multipart} = 1 if $h{is_multipart};

		if(!$h{form_url}){
			my $url = $self->{common_url}.$self->{url_sort}.$self->{list_url_params}.$self->{url_filter}.$self->{url_range}.$self->{url_frequency}.$self->{prefix_save};
			$url.="/id$self->{lev},$h{id}" if $h{id};
			$h{form_url} = $url.$h{url_ext};
		}
		$h{tag_type} = 'form';
		#
		$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_open.tmpl',\%h);
	}
	$self->{edit_object} = $h{object};
	$self->{edit_req}=[];
	$self->{edit_lists}=[];
	$self->{edit_disable_views} = undef;
	$self->{edit_enable_views} = undef;

}######################## /Form Header & Footer ################################

######################## Form Close Part ######################################

sub edit_table_close{
	my $self = shift;
	my %h= (@_);
	$self->_need_mode('edit');
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_table_close.tmpl',\%h);
}

# => height : default = 10
sub edit_table_break{
	my $self = shift;
	my %h=(
		height=>10,
		new_table_class=>'LightLines',
		border_after_break=>0,
		cellpadding => 5,
	@_);

	$h{new_table_class} = 'formborder' if $h{border_after_break};
		
	$self->_need_mode('edit');
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_table_break.tmpl',\%h);
}



##tag_id default item
# => btn_del : 1
# => upd_id : res_id : default='Update', set to '0' for removing update button from page
# => del_id : resoures_string_id for del_button
# => cancel_id : string_id for "Cancel" button , set to zero for removing cancel button from page
# => btns : [{
#		title =>> 'Skip translation
#		name  =>> 'btn_skip'
#		type  =>> 'button'||'submit'
#		class =>  'SWIButtonX'|'SWIButton'|'SWButton'|undef, def=undef
#		check_form = >1/0|undef
#		code  => qq{onClick="document.forms.item.item_submit=1;"}
#		}]
# => no_btns : do not post buttons code
# => no_table_close : do not call edit_table_close
# => no_form_close : do not close form in edit_close.tmpl
# => btns_table_width : string, default='520'
# => btn_offset : push extra! offset from left side (for case when form's left_offset is '0', default='0'
# => code_custom : JS code to add into OnSubmit function
#
sub edit_close{
	my $self = shift;

	$self->_need_mode('edit');

	die qq[ERROR: page->edit_close() can be called only after page->edit_open(), current mode is '$self->{mode}'] if $self->{mode} ne 'edit' && $self->{mode} ne 'grid';

	my %h = (
		upd_id           => 'update',
		del_id           => 'delete',
		cancel_id        => 'cancel',
		confirm_question => string('sure_delete'),
		self             => $self,
		hidden           => $self->{edit_hidden},
		btn_offset       => '0',
		form_cols        => $self->{_edit_form_cols},
		tag_id           => 'item',
		form_width_real	 => $self->{form_width_real},
		@_
	);

	## TODO: pass exact form_name to edit_close (on_submit js) //maxwell@
	## tag_id           => $self->{form_tag_id} || 'item',

	{ ## quote titles
		foreach my $rec( @{$self->{edit_req}}){
			$rec->[2] = $self->escape_quotes($rec->[2]);
		}
	}


	# show close table #
	$self->edit_table_close() unless $h{no_table_close};

	$h{tag_type} = 'form';
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_close_btns.tmpl',\%h) unless $h{no_btns};
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_close.tmpl',\%h);

	if(defined $self->{_edit_out_afterform}){
		$self->{edit_out}.=$self->{_edit_out_afterform};
		$self->{_edit_out_afterform} = undef;
	}
	if ($self->{'has_aster'})
	{## search for the last occurence of the tag;
		my $ref	=\$self->{'edit_out'};
		my $pos = rindex($$ref, '<!-- /TAG type="edit_open" id="item" -->');
		if ($pos >= 0)
		{## parse the explanation text and insert the asterisk explanation
		 ## right before the tag;
			substr($$ref, $pos, 0)	= sw_web_parse_template(__PACKAGE__, 'aster_explained.tmpl',\%h);
		}
		## drop the flag
		$self->{'has_aster'}	= 0;
	}

	$self->{mode} = undef;
}


# > view_name=>value,view_name1=>value1,
sub edit_hidden{
	my ($self,%h) = @_;

	foreach(keys %h){
		push @{$self->{edit_hidden}},{view_name=>$_,value=>$h{$_}};
		#$self->{edit_hidden}->{ $_ } = $h{$_};
	}
};

# =>> content   : any HTML content
# => after_form : 1/undef, 1 = post content after form closed, undef = immediately
sub edit_append{
	my ($self,%h) = @_;
	if($h{after_form} == 1){
		$self->{_edit_out_afterform}.=$h{content};
	}else{
		$self->{edit_out}.=$h{content};
	}
}

#	=>	title_id
#	=>  title
#
# CLASSIC INTERFACE
#	=>>	view_name
##=>> value if multi_lang == 1 shold be hash {lang=>value}
##=>>multi_lang 1/0
##=>>default_lang
#	=>>	max_length
# /CLASSIC INTERFACE
# MODERN ABSTRACT INTERFACE
#   =>> obj_attr=>string with object attribute name
#         auto-setting view_name=>,value=>,max_length=>,not_empty=> argmuments
#   =>  object : use this ref instead ref from edit_open(object=>)
# /MODERN ABSTRACT INTERFACE
#	=>	not_empty
#	=>	size
#	=>	prefix
#	=>	postfix
#   =>  descr
#	=>	read_only
#	=>  is_password
#   =>  disabled : 1 or undef
#	=>	is_aster  show red asterisk, auto set for not_empty=>1
#	=>	format    :  description is before _edit_format function
#   =>  js_on_change : any JS code for <input ... onChange=""> code
#   =>  js_on_click : any JS code for <input ... onClick=""> code
#   =>  js_on_focus: any JS code for <input ... onFocus=""> code
#   =>  allow_trailspaces  : allow to input spaces at the start and at the end of line
#   =>  title_not_quote
## => langs
sub edit_view_input{
	my $self    = shift;
	my %h = (
		self => $self,
		size =>  30,
		multi_lang=>0,
		default_lang=>undef,
		langs=>undef,
		idn_convert => 1,
	@_);

	if(defined $h{obj_attr}){
		my $object =  $h{object}||$self->{edit_object};
		die unless $object;

		$h{view_name} = $h{obj_attr};

		if ($object->can($h{obj_attr})) {
			$h{value} = eval '$object->' . $h{obj_attr};
			sw_die("no such method $h{obj_attr} for the object") if $@;
		} else {
			$h{value} = $object->{$h{obj_attr}};
		}

		my $col = $object->_get_class_info->{props}->{ $h{obj_attr} };

		if(not defined $h{max_length}){
			my $max = $col->{max_length};
			unless(defined $max){
				sw_die("ERROR: page->edit_view_input(), in CLASS_INFO undefined  \"max_length\" value for attribute $h{obj_attr}  ");
			}
			$h{max_length} = $max;
		}
		$h{not_empty} = $col->{required} unless defined $h{not_empty};
	}

	$h{input_type} = $h{is_password}?'password':'text';
	$h{title}.= string($h{title_id}) if $h{title_id};

	$self->row_open();
	$self->cell_text(
		value     => $h{title},
		tag_type  => 'cell_title',
		tag_id    => $h{view_name},
		width     => 165,
		is_aster  => defined $h{is_aster} ? $h{is_aster} : ( $h{not_empty} && !( $h{read_only} ) ),
		not_quote => $h{title_not_quote},
	);
	$self->cell_input(
		value             => $h{value},
		check_title       => $h{not_empty} ? $h{title} : '',
		view_name         => $h{view_name},
		max_length        => $h{max_length},
		read_only         => $h{read_only},
		disabled          => $h{disabled},
		size              => $h{size},
		input_type        => $h{input_type},
		prefix            => $h{prefix},
		suffix            => $h{postfix},
		descr             => $h{descr},
		is_password       => $h{is_password},
		width             => 345,
		js_on_change      => $h{js_on_change},
		js_on_click       => $h{js_on_click},
		js_on_focus       => $h{js_on_focus},
		allow_trailspaces => $h{allow_trailspaces},
		format            => $h{format} || undef,
		cols              => $self->{_edit_form_cols} - 1,
		multi_lang        => $h{multi_lang},
		default_lang      => $h{default_lang},
		langs             => $h{langs},
		idn_convert       => $h{idn_convert},
		type              => $h{type},
		vendor_id         => $h{vendor_id},
	);
	$self->row_close();
}

# => title_id
# => title
#
# /CLASSIC INTERFACE
#   =>  view_name
#   =>  value
#   =>  not_empty
# /CLASSIC INTERFACE
#
# OPTIONS AS 2D-ARRAY
# =>> options   # REQ_Y # ref to [[value,res_id],[value1,res_id1]]
# /OPTIONS AS 2D-ARRAY
# OPTIONS AS HASHES ARRAY
# =>> options       : ref to hash
#	 =>> data  		: ref to hashes array
#	 =>> attr_value : value attribute name
#	 =>> attr_title : title attribute name
# /OPTIONS AS HASHES ARRAY
#
# => selector_title # To replace default title '-- Select --'
# => read_only
# => options_plain=>1/0
##=>disabled
# => design_width
# => javascript
#   =>  title_not_quote
# => no_default : if set no default string '-- Select --' will be added to the list of options
sub edit_view_combo{
	my $self = shift;
	my %h=(
		self=>$self,
    @_);

	$h{title}.= string($h{title_id}) if $h{title_id};

	my $aster = ( $h{not_empty} && (not $h{read_only}) );
	$aster=0 if (defined $h{value} && $h{value} ne '') && ref($h{options}) eq 'ARRAY' && defined $h{options}->[0]->[0] && $h{options}->[0]->[0] ne '' ;
	$aster=0 if (defined $h{value} && $h{value} ne '') && ref($h{options}) eq 'HASH' && defined $h{options}->{data}->[0]->{$h{options}->{attr_value}} && $h{options}->{data}->[0]->{$h{options}->{attr_value}} ne '';
	$self->row_open();
	$self->cell_text(
		value     => $h{title},
		tag_type  => 'cell_title',
		tag_id    => $h{view_name},
		width     => 165,
		not_quote => $h{title_not_quote},
		is_aster  => $aster ? 1 : undef
	);

	$self->cell_combo(
		value          => $h{value},
		check_title    => $h{not_empty} ? $h{title} : undef,
		read_only      => $h{read_only},
		view_name      => $h{view_name},
		options        => $h{options},
		options_plain  => $h{options_plain},
		design_width   => $h{design_width},
		width          => 345,
		javascript     => $h{javascript},
		js_on_change   => $h{js_on_change},
		cols           => $self->{_edit_form_cols} - 1,
		selector_title => $h{selector_title},
		disabled       => $h{disabled},
		no_default     => $h{no_default},
	);
	$self->row_close();
}

#	=>	title_id
#	=>	title
# CLASSIC INTERFACE
#	=>>	view_name
##=>> value if multi_lang == 1 shold be hash {lang=>value}
##=>>multi_lang 1/0
##=>>default_lang
#	=>>	max_length # REQ_Y #
# /CLASSIC INTERFACE
# MODERN ABSTRACT INTERFACE
#   =>> obj_attr=> string with object attribute name
#         auto-setting view_name=>,value=>,max_length=>,not_empty=> argmuments
#   =>  object : use this ref instead ref from edit_open(object=>)
# /MODERN ABSTRACT INTERFACE
#	=>	size
#	=>	not_empty
#	=>	rows	    : area height in rows # NOT_REQ, default = 20
#	=>	read_only
#	=>	disabled
#	=>  text_below
#   =>  title_not_quote
## langs => [['en','English'], ... ]
## langs_localized => 1/0 - convert strings in template
##	=>	rich_editor 1/0 (for HTML creation in WYSIWYG mode)
sub edit_view_textarea{
	my $self=shift;
	$self->_need_mode('edit');

	my %h = (
		rows => 20,
		size => 46,
		self => $self,
		multi_lang=>0,
		default_lang=>undef,
		langs=>undef,
		langs_localized=>0,
		@_
	);


	if(defined $h{obj_attr}){
		my $object =  $h{object}||$self->{edit_object};
		die unless $object;

		$h{view_name} = $h{obj_attr} unless defined $h{view_name};
		$h{value} = $object->{  $h{obj_attr}   }; #TODO :use real setter
		my $max = $object->_get_class_info->{props}->{ $h{obj_attr} }->{max_length};
		sw_die(
			qq[ERROR: page->edit_view_textare(), in CLASS_INFO undefined ].
			qq[ "max_length" value for attribute $h{obj_attr}  ] )
			unless defined $max;
		$h{max_length} = $max;
		$h{not_empty} = $object->_get_class_info->{props}->{ $h{obj_attr} }->{required} unless defined $h{not_empty};
	}


	$h{read_only} = $h{read_only}||$self->{read_only};
	$h{title}.= string($h{title_id}) if $h{title_id};

	if(defined $h{title}){
		$self->row_open();
		$self->cell_text(
			value     => $h{title},
			cols      => $self->{_edit_form_cols},
			tag_type  => 'cell_title',
			tag_id    => $h{view_name},
			is_aster  => $h{not_empty},
			not_quote => $h{title_not_quote}
		);
		$self->row_close(close=>0);
	}
	$self->row_open();
	$self->cell_input(
		value		=> $h{value},
		view_name	=> $h{view_name},
		tag_type	=> 'cell_textarea',
		check_title	=> $h{not_empty} ? $h{title} : undef,
		rows		=> $h{rows},
		size		=> $h{size},
		max_length	=> $h{max_length},
		cols		=> $self->{_edit_form_cols},
		warn_title	=> $h{title},
		read_only	=> $h{read_only},
		disabled	=> $h{disabled},
		multi_lang	=> $h{multi_lang},
		default_lang=> $h{default_lang},
		langs		=> $h{langs},
		langs_localized=>$h{langs_localized},
		rich_editor	=> $h{rich_editor},
		title		=> $h{title},
	);
	$self->row_close(open=>defined $h{title}?0:1);
	if(defined $h{text_below}){
		$self->row_open();
		$self->cell_text(value=>$h{text_below},cols=>$self->{_edit_form_cols});
		$self->row_close();
	}
}
#	=>	title_id
#	=>	title
# 	=>	spec_back
#	=>	is_big # REQ_N # show value=> on full width
#	=>	value
#   =>  not_quote
sub edit_view_text{
	my $self    = shift;
	my %h = (
		self => $self,
		idn_convert => 1,
	@_);
	$h{title}.= string($h{title_id}) if $h{title_id};

	$self->row_open();
	if($h{is_big} || (not defined $h{title})){
		$self->cell_text(
			value     => $h{value},
			tag_type  => 'cell_text',
			tag_id    => $h{view_name},
			cols      => $self->{_edit_form_cols},
			not_quote => $h{not_quote},
			idn_convert => $h{idn_convert}
		);
	}else{
		$self->cell_text(
			value     => $h{title},
			tag_type  => 'cell_title',
			tag_id    => $h{view_name},
			not_quote => $h{not_quote},
			idn_convert => $h{idn_convert},
			nowrap		=> $h{nowrap},
		);
		$self->cell_text(
			value     => $h{value},
			tag_type  => 'cell_text',
			tag_id    => $h{view_name},
			cols      => $self->{_edit_form_cols} - 1,
			not_quote => $h{not_quote},
			idn_convert => $h{idn_convert}
		);
	}
	$self->row_close();
}

#	=>	value
#   =>  right_line
#   =>  not_quote
#   =>  link_url
#   =>  link_target
#   =>  button_title   : text on the button
#   =>  button_name
#   =>  button_class
##  Note! right_line and button cannot be in same edit_view_info element

sub edit_view_info{
	my $self    = shift;	

	my %h = (
		button_class => 'SWButtonX',
		button_title => string('edit'),
		buttons      => undef,
		@_
	);

	my $cols = $self->{_edit_form_cols};

	$self->row_open();

	if( $h{right_line} ) {
		$self->cell_text(
			value      => $h{value},
			cols       => $cols,
			right_line => 1,
			not_quote  => $h{not_quote}
		);
	} elsif( $h{button_name} ) {
		$cols = $self->{_edit_form_cols} - 1;
		$self->cell_text(
			bg_gray     => 1,
			value       => $h{value},
			cols        => $cols,
			not_quote   => $h{not_quote},
			link_url    => $h{link_url},
			link_target => $h{link_target}
		);
		$self->cell_button(
			bg_gray    => 1,
			align      => 'right',
			view_name  => $h{button_name},
			value      => $h{button_title},
			class      => $h{button_class},
			js_onclick => $h{js_onclick},
		);
	} elsif( $h{buttons} ) {
		$cols = $self->{_edit_form_cols} - scalar( @{$h{buttons}} );
		$self->cell_text(
			bg_gray     => 1,
			value       => $h{value},
			cols        => $cols,
			not_quote   => $h{not_quote},
			link_url    => $h{link_url},
			link_target => $h{link_target}
		);
		foreach my $btn ( @{$h{buttons}} ) {
			$self->cell_button(
				bg_gray    => 1,
				align      => 'right',
				view_name  => $btn->{button_name},
				value      => $btn->{button_title},
				class      => $btn->{button_class} || 'SWButtonX',
				js_onclick => $btn->{js_onclick},
			);
		}
	} else {
		$self->cell_text(
			bg_gray     => 1,
			value       => $h{value},
			link_url    => $h{link_url},
			link_target => $h{link_target},
			cols        => $cols,
			not_quote   => $h{not_quote}
		);
	}
	$self->row_close();
}


#	=>	title_id
#	=>  title
# CLASSIC INTERFACE
#	=>	view_name
#	=>	value
# /CLASSIC INTERFACE
# MODERN ABSTRACT INTERFACE
#   =>  obj_attr=> # REQ_Y # string with object attribute name
#   =>  object : use this ref instead ref from edit_open(object=>)
# /MODERN ABSTRACT INTERFACE
#	=>	read_only
# 	=>  full_row
#   =>  spec_back
#	=>	enable_views
#	=>	disable_views
# 	=>  js_on_change : any JS code for <input ... onChange=""> code
# 	=>  js_on_click : any JS code for <input ... onClick=""> code
# 	=>  js_on_focus : any JS code for <input ... onFocus=""> code
#   =>  title_not_quote
sub edit_view_check{
	my $self = shift;
	my %h = (@_);

	$h{read_only} = 1 if $self->{read_only};
	$h{title}.= string($h{title_id}) if $h{title_id};

	if($h{obj_attr}){
		my $object =  $h{object}||$self->{edit_object};
		die unless $object;

		$h{view_name} = $h{obj_attr};
		$h{value} = $object->{  $h{obj_attr}   }; #TODO :use real setter
	}

	if($h{full_row}){
		$self->row_open();
		$self->cell_check(
			align=>'left',
			close=>0,
			cols=>$self->{_edit_form_cols},
			%h
		);
		$self->cell_text(
			value=>$h{title},
			tag_type=>'cell_title',
			tag_id  =>$h{view_name},
			open=>0,
			not_quote=>$h{title_not_quote}
		);
		$self->row_close(bg_gray=>$h{spec_back});
	}else{
		$self->row_open();
		$self->cell_text(
			value=>$h{title},
			tag_type=>'cell_title',
			tag_id  =>$h{view_name},
			width=>155,
			not_quote=>$h{title_not_quote}
		);
		$self->cell_check(
			cols=>$self->{_edit_form_cols}-1,
			align=>'left',%h
		);
		$self->row_close(bg_gray=>$h{spec_back});
	}
}


# =>  title_id
# =>  title
# =>> view_name
# =>> value
# =>  is_checked
# =>  enable_views : array of view_names
# =>  disable_views : array of view_names
#### javascript-code #######
# =>  js_code : any JavaScript-code, will be posted in <input...._here_...>
#         or
# =>  js_on_change : any JS code for <input ... onChange=""> code
# =>  js_on_click : any JS code for <input ... onClick=""> code
# =>  js_on_focus : any JS code
#   =>  title_not_quote
sub edit_view_radio{
	my $self = shift;
	$self->_need_mode('edit');
	my %h = (
		self => $self,
	@_);
	$h{read_only} = 1 if $self->{read_only};
	$h{title}.= string($h{title_id}) if $h{title_id};
	#
	$self->row_open();
	$self->cell_radio(align=>'left',cols=>$self->{_edit_form_cols},close=>0,%h);
	$self->cell_text(align=>'left',tag_type=>'cell_title',tag_id=>$h{'view_name'},
		value=>'&nbsp;'.$h{title},open=>0,not_quote=>$h{title_not_quote});
	$self->row_close();
}

#	=>	view_name
#	=>	value
#	=>	replace
sub edit_view_hidden{
	my $self    = shift;

	my %h = (@_);

	if($h{replace}){
		my @recs = grep{ $_->{view_name} eq $h{view_name} } @{$self->{edit_hidden}};
		if($#recs>=0){
			$recs[0]->{value}=$h{value};
			return;
		}
	}
	push @{$self->{edit_hidden}},\%h;
	#$self->{edit_hidden}->{ $h{view_name} } = $h{value};
};

## >
##	title_id=>
##	title=>
##	view_name=>
##	value=>
##	title_no_wrap=>
##	read_only=>
##		hide_time=>
##	class_res=> : empty=$self->{obj}->{class}
## =>  title_not_quote
## => type - default date
sub edit_view_datetime{
	my $self=shift;

	my %h=(self=>$self,@_);

	$h{title}.= string($h{title_id}) if $h{title_id};

	$self->row_open();
	## Title
	$self->cell_text(
		value    =>$h{title},
		tag_type =>'cell_title',
		tag_id   =>$h{view_name},
		width    =>165,
		not_quote=>$h{title_not_quote},
	);
	## Date
	$self->cell_datetime(
		width=>345,cols=>$self->{_edit_form_cols}-1,%h
	);
	$self->row_close();
}##/edit_view_datetime

## => view_name
## => type 			: default is 'date'
## => value			: current value
## => read_only		: only show current value
## => hide_time		: if true the calendar will not include a time selector
##					default false
## => format   		: the date format that will be used to display the date in displayArea, 
##					default  - from user locale
## => first_day     : numeric: 0 to 6.  "0" means display Sunday first, "1" means display Monday first, etc.
## => start_date	: suppress to select dates before start_date
## => start_year	: the range of years available 
## => end_year		: 	Default: [curent_year - 2 , current year + 10]
## => step          : configures the step of the years in drop-down boxes; default: 1
sub cell_datetime {

	my $self=shift;

	my %th=(
		self		=> $self,
		cols		=> 1,
		start_date	=> undef,
		start_year	=> undef,
		end_year	=> undef,
		format		=> undef,
		first_day	=> $ENV{'system_obj'}->provider_config->def_first_day_week,
		read_only	=> $self->{read_only},
		type		=> 'date',
		@_
	);

	if( not defined $th{value} or $th{value} eq '') {
		$th{value} = time_gmt_to_local(gmt_now());
	}

	my %current = split_datetime($th{value});
	@th{keys %current} = values %current;

	$th{class_res} = $self->{obj}->{class} if $th{class_res} eq '';

	## limit starting date
	if ($th{start_date}){
		my %start = split_datetime($th{start_date});
		$th{'start_date_'.$_} = $start{$_}
			foreach keys %start;
	}

	my $cur_year = (localtime)[5]+1900;
	my @years;
	if ($th{start_year} && $th{end_year}) {
		push(@years,$th{start_year}..$th{end_year});
	} elsif ($th{start_year}) {
		push(@years,$th{start_year});
		push(@years,$cur_year - 2..$cur_year + 10);
	} elsif ($th{end_year}) {
		push(@years,$th{end_year});
		push(@years,$cur_year - 2..$cur_year + 10);
	} else {		
		push(@years,$cur_year - 2..$cur_year + 10);
		push(@years,$th{year} - 2) if $th{year} > 0;
	}
	push(@years,$th{start_date_year}) if $th{start_date_year};

	@years = sort {$a <=> $b} @years;
	$th{range} = '['.shift(@years).','.pop(@years).']';

	if($th{read_only}){
		my $view_name = $th{view_name};
		if(not $th{hide_time}){
			$self->edit_view_hidden(view_name=>"hour_$view_name",value=>$th{hour});
			$self->edit_view_hidden(view_name=>"min_$view_name",value=>$th{min});
		}
		$self->edit_view_hidden( view_name=>"month_$view_name",value=>$th{month} );
		$self->edit_view_hidden( view_name=>"day_$view_name",value=>$th{day} );
		$self->edit_view_hidden( view_name=>"year_$view_name",value=>$th{year} );
	}

	$self->edit_view_hidden(
		view_name=>"_type_".$th{view_name},
		value    =>$th{type}
	);

   	if ($th{hide_time}){
   		$th{format} ||= $LOCALE{local_lang_conf}->{dateformat};
   		$th{current} ||= date_to_str($th{value});
   	} else {
   		$th{format} ||= $LOCALE{local_lang_conf}->{datetimeformat};
   		$th{current} ||= time_to_str(time_local_to_gmt($th{value}));
	}
             
	if ( $th{read_only} ) {
		$th{content} = $th{current};
	} else {
		$th{cell_datetime_count}=$self->{cell_datetime_count}++;
		$th{content} .= sw_web_parse_template(
			__PACKAGE__,'cell_datetime.tmpl',\%th
		);
	}
	push @{$self->{grid}->{cols}},\%th;
}##/cell_datetime


# >
#		title_id=>
#		title=>
#		view_name=>
#		value=>
#		title_no_wrap=>
#		read_only=>
#		class_res=> : empty=$self->{obj}->{class}
#		hide_time=>
#       hide_seconds=>
#   =>  title_not_quote
sub edit_view_period{
	my $self=shift;
	my %th=(
		@_
	);

	$th{title}.= string($th{title_id}) if $th{title_id};

	$self->row_open();
	$self->cell_text(
		value     => $th{title},
		width     => 165,
		not_quote => $th{title_not_quote},
		tag_type  => 'cell_title',
		tag_id    => $th{'view_name'},
	);
	$self->cell_period(%th);
	$self->row_close;

	return undef;

}


# =>	title_id
# =>	title
# =>  view_name
# =>  not_empty (1,undef)
# =>  size
#   =>  title_not_quote
# Comments:
#  do not forget about edit_open(is_multipart=>1)
sub edit_view_file{
	my $self    = shift;

	die 'ERROR : You can use edit_view_file() only in multipart form. User edit_open(is_multipart=>1).'unless $self->{_form_multipart};

	my %h = (
		self => $self,
		size => 30,
	@_);

	$h{title}.= string($h{title_id}) if $h{title_id};

	if($h{new}){
		$self->row_open();
		$self->cell_text(value=>$h{title},tag_type=>'cell_title',tag_id=>$h{view_name},
			is_aster=>$h{not_empty},not_quote=>$h{title_not_quote});
		$self->cell_file(
			view_name=>$h{view_name},
			size=>$h{size},
			cols=>$self->{_edit_form_cols}-1,
			check_title=>$h{not_empty}?$h{title}:undef
		);
		$self->row_close();
		return;
	}

	$self->_need_mode('edit');

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_view_file.tmpl',\%h);

	if($h{not_empty} && $h{view_name} ne ''){
		push( @{$self->{edit_req}}, [$h{view_name},'',$h{title}]);
	}
}

# => t : template_name
# => class : template_class, if empty, then take $page->{class_app}
# OR
# just HTML to put into stream
sub edit_view_custom{
	my $self    = shift;
	$self->_need_mode('edit');

	$self->grid_push;

    if($#_){
    	my %h = (@_);
    	$self->{edit_out} .= sw_web_parse_template($h{class}||$self->{class_app},$h{t},\%h);
    }else{
        $self->{edit_out} .= $_[0];
    }
}

# >>  self
#  =>> obj : ref to HSPC::MT::Core::Address instances
#  =>> view_name
#  =>  not_empty : 1/undef
#  =>  data_prefix : string with prefix, for example : 'admin_'
#  =>  read_only : 1/undef
#  =>  cdesign_width
#  =>  full_address : 1= show city,address1,address2,phone inputs
##  =>  hide_state_alt : 1 = hide
##  =>  hide_phone_ext: 1 = hide
##  =>  hide_fax_ext: 1 = hide
##	=> hide = [], ref to array with hide elements
##  => max_length => {}, ref on hash with elements max length
##		ex. max_length => {address1 => 100, address2 => 100}
sub edit_view_geo{
	my $self = shift;

	my %h = (
	    hide => [],
		max_length => undef,
	     @_
	);
	my $obj = $h{obj};
	unless( defined $obj || defined $h{view_name}){
		$ENV{system_obj}->warn_calls_stack();
		die 'ERROR : edit_view_geo() undefined view_name=>' unless defined $h{view_name};
		die 'ERROR : edit_view_geo() undefined obj=>' unless defined $h{obj};
	}
	my $max_length = $h{max_length};
	## if element max_length undef, will be used default value
	my $default_max_length = {
		address1    => 120,
		address2    => 120,
		state_alt   => 60,
		city        => 60,
	};

	if (defined $h{max_length}){
		foreach my $key (keys(%{$default_max_length})){
			## if element is not set, set it in default value
			$max_length->{$key} ||= $default_max_length->{$key};
		}
	} else {
		$max_length = $default_max_length;
	}

	use HSPC::GeoData;
	## set default country
	##unless( defined $obj->{$h{namecountry}} ){
		##my $address = HSPC::Web->web_get_provider_address();
		##$obj->{$h{namecountry}} = $address->country;
	##}

	if( $h{full_address} ){
	 	$self->row_open();
		$self->cell_text(
			value=>string('address_line_1'),
			tag_type=>'cell_title',
			tag_id  =>$h{view_name}.'__address1',
			not_quote=>$h{title_not_quote},
			is_aster=>defined $h{is_aster}?$h{is_aster}:($h{not_empty} && !($h{read_only}))
		);

 		$self->cell_input(
 			view_name=>$h{view_name}.'__address1',
	 		value=>$obj->{address1},
 			check_title=>$h{not_empty}?string('address_line_1'):'',
 			max_length=>$max_length->{address1},size=>30, class=>'SWi300'
 		);

	 	$self->row_close();

		$self->edit_view_hidden(title_id=>'address_num',
			view_name=>$h{view_name}.'__house_num',
			value=>$obj->{house_num},
			replace => 1,
		);

		$self->edit_view_hidden(title_id=>'address_suffix',
			view_name=>$h{view_name}.'__house_suff',
			value=>$obj->{house_suff},
			replace => 1,
		);
		unless(scalar (grep {$_ eq 'address2'} @{$h{hide}})){
			$self->edit_view_input(title_id=>'address_line_2',
				view_name=>$h{view_name}.'__address2',
				value=>$obj->{address2},
				max_length=>$max_length->{address2},size=>30
			);
		}
		unless(scalar (grep {$_ eq 'city'} @{$h{hide}})){
			$self->edit_view_input(title_id=>'city',
				view_name=>$h{view_name}.'__city',
				value=>$obj->{city},
				max_length=>$max_length->{city},size=>20,
				not_empty=>$h{not_empty}
			);
		}
	}

	## if state not hidden, show it
	unless(scalar (grep {$_ eq 'state'} @{$h{hide}})){
	    # view state US and Canada #
	    $self->edit_view_combo(title=>string('state_uscanada'),
		    view_name=>$h{view_name}.'__state',
	    	    read_only=>$h{read_only},
		    value=>$obj->{state},
		    options_plain=>1,
		    not_empty=>0,
		    design_width=>150,
		    options=>HSPC::GeoData->get_states($h{allow_to_select_all}),
		    );
	    # view state other countries #
	    $self->edit_view_input(title=>string('state_other'),
		    view_name=>$h{view_name}.'__state_alt',
		    value=>$obj->{state_alt},
		    max_length=>$max_length->{state_alt},size=>15,not_empty=>0
	    ) unless($h{hide_state_alt});
	}
	## if state not hidden, show it
	unless(scalar (grep {$_ eq 'country'} @{$h{hide}})){
	    my $selector_title = $h{allow_to_select_all}? string('all_countries'):'';
	    $self->edit_view_combo(title=>string('country'),
		    view_name=>$h{view_name}.'__country',
		    read_only=>$h{read_only},
		    value=>$obj->{country}||undef,
		    options_plain=>1,
		    allow_to_select_all=>$h{allow_to_select_all},
		    not_empty=>$h{not_empty},
		    design_width=>$h{design_width},
		    options=>HSPC::GeoData->get_countries(),
		    selector_title=>$selector_title
		    );}
	if( $h{full_address} ){
		$self->edit_view_input(title_id=>'zip_code',
			view_name=>$h{view_name}.'__zip',
			value=>$obj->{zip},
			max_length=>15,size=>15,not_empty=>$h{not_empty}
		);
		$self->edit_view_phone(
			view_name	=> $h{view_name}.'__phone',
			title_id	=> 'phone',
			element		=> 'phone',
			not_empty	=> $h{not_empty},
			obj			=> $obj,
			hide_ext	=> $h{hide_phone_ext}
		);
		$self->edit_view_phone(
			view_name	=> $h{view_name}.'__mobile',
			title_id	=> 'mobile',
			element		=> 'mobile',
			obj			=> $obj,
			hide_ext 	=> 1,
		) unless scalar (grep {$_ eq 'mobile'} @{$h{hide}});
		$self->edit_view_phone(
			view_name	=> $h{view_name}.'__fax',
			title_id	=> 'fax',
			element		=> 'fax',
			obj			=> $obj,
			hide_ext	=> $h{hide_fax_ext}
		) unless scalar (grep {$_ eq 'fax'} @{$h{hide}});
	}

	# STORE CONTROL TYPE #
	$self->edit_view_hidden(view_name=>'_type_'.$h{view_name},value=>'geo_'.($h{full_address}?'full':'short'));
}


## >>  self
## =>> view_name
## =>> title_id/title
## =>> obj : ref to HSPC::MT::Core::Address or HSPC::MT::Core::AbstractAccount instance
## =>  data_prefix : string with prefix, for example : 'admin_'
## => not_empty
##   =>  title_not_quote
## => hide_ext 1 = hide
## =>	not_empty 1 = not empty 'country_code' and 'number' (claasic view)
##		not_empty => ref on ARRAY with not empty elements
##			ex. country_code, area_code, number, extension
sub edit_view_phone{
	my $self = shift;
	my %h = (
		not_empty => undef,
		@_
	);
	die unless defined $h{obj};

	my $not_empty = [
		'country_code',
		'number',
	];
	if (ref $h{not_empty} eq 'ARRAY'){
		$not_empty = $h{not_empty};
	} elsif (!$h{not_empty}){
		$not_empty = [];
	}
	$h{title} = string($h{title_id}) if defined $h{title_id};
	$h{element} ||= 'phone';

	require HSPC::MT::Core::Phone;
	$h{phone_hash} = HSPC::MT::Core::Phone->phone_as_hash(
		$h{obj}->{"$h{data_prefix}$h{element}_src"},
		is_keep_delimeters	=>	1
	);

	$self->row_open();
	$self->cell_text(value=>$h{title},valign=>'top',not_quote=>$h{title_not_quote});
	$self->cell_grid_open();
	$self->cell_text(
		value		=>	string('country_code'),
		is_aster	=>	(grep { $_ eq 'country_code'} @{$not_empty}) ? 1 : 0,
		align		=>	'bottom',
		nowrap		=>	1
	);
	$self->cell_skip(width=>2);
	$self->cell_text(
		value		=>	string('area_code'),
		nowrap		=>	1,
		is_aster	=>	(grep { $_ eq 'area_code'} @{$not_empty}) ? 1 : 0
	);
	$self->cell_skip(width=>5);
	$self->cell_text(
		value		=>	string('num'),
		is_aster	=>	(grep { $_ eq 'number'} @{$not_empty}) ? 1 : 0,
		valign		=>	'bottom',
		nowrap		=>	1
	);
	unless($h{hide_ext}){
		$self->cell_text(
			value		=>	string('extention'),
			nowrap		=>	1,
			is_aster	=>	(grep { $_ eq 'extension'} @{$not_empty}) ? 1 : 0
		);
	}
	$self->row_close(cellspacing=>1);
	$self->cell_input(
		view_name	=>	$h{view_name}.'_country_code',
		value		=>	$h{phone_hash}->{country_code},
		max_length	=>	3,
		prefix		=>	'+',
		size		=>	6,
		valign		=>	'bottom',
		check_title	=>	(
			(
				grep { $_ eq 'country_code'} @{$not_empty})
				? $h{title}.'>'.string('country_code')
				: ''
			)
	);
	$self->cell_skip(width=>2);
	$self->cell_input(
		view_name	=>	$h{view_name}.'_area_code',
		value		=>	$h{phone_hash}->{area_code},
		max_length	=>	8,
		valign		=>	'bottom',
		size		=>	6,
		check_title	=>	(
			(
				grep { $_ eq 'area_code'} @{$not_empty})
				? $h{title}.'>'.string('area_code')
				: ''
			)
	);
	$self->cell_skip(width=>5);
	$self->cell_input(
		view_name	=>	$h{view_name}.'_number',
		value		=>	$h{phone_hash}->{number},
		max_length	=>	10,
		valign		=>	'bottom',
		size		=>	10,
		check_title	=>	(
			(
				grep { $_ eq 'number'} @{$not_empty})
				? $h{title}.'>'.string('num')
				: ''
			)
	);
	unless($h{hide_ext}){
		$self->cell_input(
			view_name	=>	$h{view_name}.'_extension',
			value		=>	$h{phone_hash}->{extension},
			max_length	=>	6,
			valign		=>	'bottom',
			size		=>	6,
			check_title	=>	(
				(
					grep { $_ eq 'extension'} @{$not_empty})
					? $h{title}.'>'.string('extention')
					: ''
				)
		);
	}
	$self->row_close(cellspacing => 1);
	$self->cell_grid_close();
	$self->row_close();

	# STORE CONTROL TYPE #
	$self->edit_view_hidden(
		view_name	=>	'_type_'.$h{view_name},
		value		=>	'phone'
	);
	$self->edit_view_hidden(
		view_name	=>	'_dataprefix_'.$h{view_name},
		value		=>	$h{data_prefix}
	);

} ## edit_view_phone


# > self
#   =>> view_name
#	=>>	title_id/title
#	=>	not_empty
#	=>  read_only
#	=>  value_title
#	=>  view_name_title
#	=>  value_first
#	=>  view_name_first
#	=>  value_middle
#	=>  view_name_middle
#	=>  value_last
#	=>  view_name_last
######################
#   =>  obj #REQ_N#
#   =>  data_prefix
#	=>  api_ver=> 1 for normal API !!!!
######################
#	=>	is_aster# #REQ_N# show red asterisk, auto set for not_empty=>1
##   =>  title_not_quote
sub edit_view_fio{
	my $self=shift;
	my $titles = [
		['',''],
		[string('mr'),''],
		[string('mrs'),''],
		[string('ms'),''],
		[string('dr'),'']
	];
	foreach(@$titles){
		$_->[1] = $_->[0];
	}

	my %h=(
		# DROP THEN ALL APPS WILL USE api_ver=>1
		view_name_title => 'title',
		view_name_first => 'first',
		view_name_middle => 'middle',
		view_name_insertion => 'insertion',
		view_name_last => 'last',
	@_);

	if($h{api_ver}==1){
		$h{view_name_title} = 'prefix';
		$h{view_name_first} = 'fname';
		$h{view_name_middle} = 'mname';
		$h{view_name_insertion} = 'insertion';
		$h{view_name_last} = 'lname';
	}

	if(defined $h{obj} ){
		my $obj = $h{obj};
		$h{value_title}  = $obj->{ $h{data_prefix}.$h{view_name_title} };
		$h{value_first}  = $obj->{ $h{data_prefix}.$h{view_name_first} };
		$h{value_middle}  = $obj->{ $h{data_prefix}.$h{view_name_middle} };
		$h{value_insertion}  = $obj->{ $h{data_prefix}.$h{view_name_insertion} };
		$h{value_last}  = $obj->{ $h{data_prefix}.$h{view_name_last} };
	}

	if(defined $h{view_name}){
		$h{view_name_title} = $h{view_name}.'__'.$h{view_name_title};
		$h{view_name_first} = $h{view_name}.'__'.$h{view_name_first};
		$h{view_name_middle} = $h{view_name}.'__'.$h{view_name_middle};
		$h{view_name_insertion} = $h{view_name}.'__'.$h{view_name_insertion};
		$h{view_name_last} = $h{view_name}.'__'.$h{view_name_last};
	}

	$h{title}.=string($h{title_id}) if $h{title_id};

	$self->row_open();
	$self->cell_text(value=>$h{title},not_quote=>$h{title_not_quote});
	$self->cell_grid_open();
	$self->cell_text(value=>string('person_title'));
	$self->cell_text(value=>string('person_first'),
		is_aster=>defined $h{is_aster}?$h{is_aster}:($h{not_empty} && !($h{read_only})));
	$self->cell_text(value=>string('person_middle'));
	$self->cell_text(value=>string('insertion'));
	$self->cell_text(value=>string('person_last'),
		is_aster=>defined $h{is_aster}?$h{is_aster}:($h{not_empty} && !($h{read_only})));
	$self->row_close(cellspacing=>1);
	$self->cell_combo(view_name=>$h{view_name_title},
		value=>$h{value_title},
		options=>$titles,
		options_plain=>1,
		align=>'left');

	my $names_size = 9;
	my $browser = $ENV{HSPC_SYSTEM}->browser;
	if (lc($browser->os_string) eq 'linux' && lc($browser->browser_string) eq 'netscape'){
		$names_size = 18;
	}
	$self->cell_input(
		view_name=>$h{view_name_first},value=>$h{value_first},
		check_title=>$h{not_empty}?string('first_name'):'',
		max_length=>60,size=>$names_size, class=>'SWi90');
	$self->cell_input(view_name=>$h{view_name_middle},
		value=>$h{value_middle},
		max_length=>30,
		size=>3, class=>'SWi60');
	$self->cell_input(view_name=>$h{view_name_insertion},
		value=>$h{value_insertion},
		max_length=>10,
		size=>3, class=>'SWi60');
	$self->cell_input(view_name=>$h{view_name_last},value=>$h{value_last},
		max_length=>60,size=>$names_size, class=>'SWi90',
		$h{not_empty}?(check_title=>string('last_name')):'');
	$self->row_close(cellspacing=>0);
	$self->cell_grid_close();
	$self->row_close();


	# STORE CONTROL TYPE #
	if( defined $h{view_name} ){
		$self->edit_view_hidden(
			view_name=>'_type_'.$h{view_name},
			value=>'fio'.($h{api_ver}?$h{api_ver}:'')
		);
		$self->edit_hidden('_typedata_'.$h{view_name}=>$h{data_prefix}) if defined $h{data_prefix};
	}

	return 1;
}


#	=>	title_id
#	=>	title
#	=>	#view_name
#	=>	data	  	 : ref to array with list rows, array consist of [id,title]
## => selected : ref to array of id which will be selected
#	=>	not_empty
#	=>	rows	   	 : area height in rows # NOT_REQ, default = 10
# =>  width : width of area. Possible values are:
#             450|400|350|300|250|200|150|100|80|70|50
#	=>	read_only
## =>  js_on_change : any JS code for <input ... onChange=""> code
## =>  js_on_click : any JS code for <input ... onClick=""> code
## =>  js_on_focus: any JS code for <input ... onFocus=""> code
## =>  multiple 1/0 default 1
#	=>  btns         : ref to array with buttons for list operations, array consist of ref to hash with btn info
####################################
#	btns setup format:
#	  custom button #
#    => #title : button title
#    => name : button name
#    => script : JS-script for button pressed
#    => width : button width in pixels
#    => class : CSS class
# 	  /custom button #
#	  system button #
#    => type : button type,variants
#			'remove' : remove selected items
#			'add' : create popup window for adding new items
#    => popup_url : for type=>'add'  : popup URL
#    => popup_width : for type=>'add'  : popup window width ,default=500
#    => popup_height : for type=>'add'  : popup window height,default=400
#	  /system button #
##   =>  title_not_quote
sub edit_view_list{
	my $self=shift;

	$self->_need_mode('edit');

	my %h=(
		rows=>10,
		self=>$self,
		post_scripts=>$self->{edit_lists_scripts},
		multiple=>1,
    @_);

	$h{read_only} = $h{read_only}||$self->{read_only};
	$h{title}.= string($h{title_id}) if $h{title_id};
	##$h{title} = $self->convert_quotes($h{title}) unless $h{title_not_quote};
	#
	my %selected;
	if ( $h{selected} ) {
	    my $data = $h{data};
		local ($_);
		foreach ( @{$h{selected}} ) {
			$selected{$_} = 'selected';
		}
	}
	$h{selected} = \%selected;
	if ($h{width}) {
			my $new_width = 50;
			foreach my $width ((50,70,80,100,150,200,250,300,350,400,450)) {
					$new_width = $width if ($h{width} >= $width );
			}
			$h{width} = $new_width;
	} else {
			$h{width} = 300;
	}

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_view_list.tmpl',\%h);
	#
	if($h{not_empty} eq '1' && $h{view_name} ne '' && (not $h{read_only}) ){
		push( @{$self->{edit_req}}, [$h{view_name},'',$h{title}]);
	}

	## if($h{read_only}){
	## }else{
		push @{ $self->{edit_lists} },$h{view_name};
		$self->edit_view_hidden(view_name=>'_type_'.$h{view_name},value=>'list');
		#$self->edit_view_hidden(view_name=>qq{$h{view_name}__list},value=>1);
	## }

	$self->{edit_lists_scripts} = 1;
}

#	=>	title_id       #REQ_N#
#	=>  title          #REQ_N#
#	=>	view_name      #REQ_Y#
#	=>	value_title    #REQ_N#
#	=>	value          #REQ_N#
#	=>  not_empty      #REQ_N#
#	=>  read_only      #REQ_N#
#   =>  popup_url      #REQ_Y#
#   =>  popup_height   #REQ_N#
#   =>  popup_width    #REQ_N#
#   =>  js_on_change : any JS code for <input ... onChange="__CODE__"> code
##   =>  title_not_quote
##  =>  value_not_quote
sub edit_view_popup{
	my $self=shift;

	my %h = @_;
	$h{title}.= string($h{title_id}) if $h{title_id};
	$self->row_open();
	$self->cell_text(
		value=>$h{title},
		width=>165,
		is_aster=>$h{not_empty} && !($h{read_only}),
		not_quote=>$h{title_not_quote},
		tag_type=>'cell_title',
		tag_id=>$h{view_name},
	);
	$h{check_title} = $h{not_empty}?$h{title}:'';
	$self->cell_popup(cols=>$self->{_edit_form_cols}-1,%h, not_quote => $h{value_not_quote});
	$self->row_close();
}

#! This function is out-of-date,
#      use sw_argparams_obj()
#
# > obj
# > names ( string or strings array)
sub restore_obj_strings{
	my $self=shift;
	my $o = shift;
	my @names = @_;
	my ($c,$c1);
	#
	foreach my $n (@names){
		$c = sw_argparam($n);
		$o->{$n}=$c;
	}
}
################### /restore_methods for OBJ Part ####################



################### View info part ###################################
## => $form_width : set a custom form width : default="param('form_view_width')"
sub view_info_header{
	my $self    = shift;
	my %h = (
		self => $self,
	@_);

	die if defined $self->{mode};
	$self->{form_width}	= $h{form_width};
	$self->{edit_out}  .= sw_web_parse_template(__PACKAGE__,'view_info_header.tmpl',\%h);
	$self->{mode}		= 'viewinfo';
}

sub view_info_footer{
	my $self    = shift;
	my %h = (
		self => $self,
	@_);

	sw_die(
		'ERROR: error in $page->view_info_footer(),you must call '.
		'$page->view_info_header() before. Current mode is '.$self->{mode})
		unless $self->{mode} eq 'viewinfo';

	$self->{edit_out} .= sw_web_parse_template(
		__PACKAGE__,'view_info_footer.tmpl',\%h
	);
	$self->{mode} = undef;
}

# >>  self
# =>> data : ref to hash/object with keys  "prefix,fname,mname,lname,email,phone"
# =>  data_key : keys prefix , for example 'admin_'
sub view_info_fio{
	my ($self,%h) = @_;
	die unless $h{data};

	my $data = $h{data};
	my $key = $h{data_key};
	my @keys = qw|prefix fname mname insertion lname|;
	my $name = join(' ',
		map {$data->{$key.$_} if defined $data->{$key.$_}} @keys );

	$self->view_info_text(title_id => 'name', value => $name);
	my $gender = $data->{$key.'gender'};
	my $string_gender;
	if ($gender eq 'M') {
		$string_gender = string('male');
	} elsif ($gender eq 'F') {
		$string_gender = string('female');
	} else {
		$string_gender = '';
	}
	$self->view_info_text(title_id => 'gender', value => $string_gender);
	$self->view_info_text(
		title_id => 'email',
		value => $data->{$key.'email'},
		is_email => 1
	);
	$self->view_info_text(title_id => 'phone', value => $data->{$key.'phone'})
		if $data->{$key.'phone'};
	$self->view_info_text(title_id => 'mobile', value => $data->{$key.'mobile'})
		if $data->{$key.'mobile'};
	$self->view_info_text(title_id => 'fax', value => $data->{$key.'fax'})
		if $data->{$key.'fax'};

}

# => title_id	[<txt_{$title_id}>] or title
# => title		: (Optional, default: CompRep)
# => value_comment : additional text, uses for types 'live','bool'
# => transparent (Optional, default FALSE) - when true doesn't reset background colors
# => is_big     : show value on full width
# => is_multiline: translate \n -> <br>
# => is_area    : 0/1 : 1 - place value in readonly textarea!
# => area_name  : undef/any_name : set area_name to place value in textarea!
#    is_area and area_name both! required to place value in textarea!
# => area_rows  : show # of rows in textarea
# => is_folder  : show folder and value on full width : workd with is_big=1
# => link_url   : (Optional, default: CompRep) show value as link to "link_url" - can be an array ref for multiline value
# => comprep	: (Optional) see '_process_comprep'
# => link_target: link target, for example  :  "_blank" - can be an array ref for multiline value
# => as_plain   : show as plain text without HTML formating
# => color		: string : show text in color, variant 'black','green','red' : default = 'black'
# => height     : height in pixels for vertical spacer
# => type       # REQ_N # logic data type, variants: (default char)
#                       char = chars data
#                       date = row date showed in localized text format
#                       money = money ( in current version only $ )
#                       int= integer value
#                       percent  = percent
#                       period  = period
#						bool = bool
#						live = bool, show red or green lamp
# => translate : ref to hash
#  TRANSLATE_DATA AS 2D-ARRAY
#    =>> data   # REQ_Y # ref to [[value1,title1],[value1,title2]]
#    =>  title_plain : title is a plain string, not localized string
#  /TRANSLATE_DATA AS 2D-ARRAY
#  TRANSLATE_DATA AS HASHES ARRAY
#	 =>> data  		: ref to hashes array
#	 =>> attr_value : value attribute name
#	 =>> attr_title : title attribute name
#    =>  title_plain : title is a plain string, not localized string
#  /TRANSLATE_DATA AS HASHES ARRAY
sub view_info_text{
	my $self    = shift;
	my %h = (
		self => $self,
		is_folder=>'',
		prefix=>'',
		postfix=>'',
		is_email=>'',
		is_area =>0,
		area_name =>undef,
		area_rows =>10,
		idn_convert => 1,
		as_plain => 1,
	@_);

	##convert multilang to string
	$h{value} = ''.$h{value}
		unless ( ref( $h{value} ) eq 'ARRAY' );

	##convert title to string
	$h{title} = ''.$h{title};

	if(defined $h{translate}){
		my $is_value_num = is_num($h{value});
		if( ref($h{translate}->{data}) eq 'ARRAY'){
			# TRANSLATE VALUE TO TITLE FROM ARRAY #
			my @titles;
			if( $is_value_num ){
			 	@titles = grep{$_->[0]==$h{value}} @{$h{translate}->{data}};
			}else{
				@titles = grep{$_->[0] eq $h{value}} @{$h{translate}->{data}};
			}
			if($#titles>=0){
				$h{value} = $titles[0]->[1];
			}else{
				$h{value} = '';
			}
		}else{
			# TRANSLATE VALUE TO TITLE FROM ARRAY HASHES #
			my @titles;
			my $attr_value=$h{translate}->{attr_value};
			if( $is_value_num ){
			 	@titles = grep{$_->{$attr_value}==$h{value}} @{$h{translate}->{data}};
			}else{
				@titles = grep{$_->{$attr_value} eq $h{value}} @{$h{translate}->{data}};
			}
			if($#titles>=0){
				$h{value} = $titles[0]->{ $h{translate}->{attr_title}  };
			}else{
				$h{value} = '';
			}
		}
		$h{value} = string( $h{value} ) unless $h{translate}->{title_plain};
	};

	$h{title}.= string($h{title_id}) if $h{title_id};
	if( $h{as_plain} ){
		$h{value}=~s|<|&lt;|g;
		$h{value}=~s|>|&gt;|g;
		$h{value}=~s|\n|\<br\>|g;
	}
	if ($h{type} eq 'money'){
		&sw_die("view_info_text: the vendor_id argument must be provided") unless ($h{vendor_id});
		$h{value}=sw_money_to_str($h{value}, $h{vendor_id});
	}elsif ($h{type} eq 'percent'){
		require HSPC::Math qw(round2);
		$h{value} = sw_float_to_str(round2($h{value}), $h{vendor_id});
	}elsif ($h{type} eq 'float'){
		require HSPC::Math qw(round2);
		$h{value} = sw_float_to_str(round2($h{value}), $h{vendor_id});
	}elsif ($h{type} eq 'period'){
		$h{value}=period_to_str($h{value});
	}elsif ($h{type} eq 'date'){
		$h{value}=date_to_str($h{value});
	}elsif ($h{type} eq 'datetime'){
		$h{value}=time_to_str($h{value});
	}elsif ($h{type} eq 'bool'){
		$h{value}=$h{value}? '<img_ok_sign>' : '<img_no_sign>';
		$h{value}.=$h{value_comment} if defined $h{value_comment};
	}elsif ($h{type} eq 'live'){
		$h{value_pre}=$h{value}?'<img_alert_ok>':'<img_alert_error>';
		if( defined $h{value_comment} ){
			if( !($h{value_comment}=~m/^ / || $h{value_comment}=~m/^\&nbsp\;/) ){
				$h{value_pre}.='&nbsp;';
			}
		}
		$h{value}=$h{value_comment};
#	}elsif ($h{type} eq 'memory'){
#		$h{value}.='&nbsp;MB';
	}

	_process_comprep($h{comprep}, \$h{link_url}, \$h{title});

	## Multi-line value or not ?
	require HSPC::Util::DM qw(prepare_idn_value);
	if (ref($h{value}) eq 'ARRAY' && ref($h{link_url}) eq 'ARRAY'){
		for (my $i=0 ; $i < scalar @{$h{value}}; $i++){
			my $current_url = $h{link_url}->[$i];
			my $current_target = $h{link_target}->[$i];
			$h{value}->[$i] = qq|<a href="$current_url"|
				.(defined $current_target ? qq| target="_$current_target" |:'')
				.'>'.$h{value}->[$i].'</a>'
				if( $current_url );
			$h{value}->[$i] = prepare_idn_value($h{value}->[$i])
				if $h{idn_convert};
		}
	} else {
		if(defined $h{link_url}){
			$h{value} = $h{value_pre}.qq|<a href="$h{link_url}"|
			.(defined $h{link_target}? qq| target="_$h{link_target}" |:'')
			.qq|>$h{value}</a>|;
		}else{
			$h{value} = $h{value_pre}.$h{value};
		}
		$h{value} = prepare_idn_value($h{value}) if $h{idn_convert};
	}

	if ($h{is_multiline}) {
		$h{value} =~ s/\n/\<br\>/msg;
	}

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,$h{is_big}?'view_info_text_big.tmpl':'view_info_text.tmpl',\%h);
}

## => title_id
## => title
## => title_url
## => comprep	: (Optional) see '_process_comprep'
## => edit_url
## => text_title_edit - button title
## => $is_collaps : "open/hide" button : default="1" show button :: "0" - hide button
## => $form_width : set a custom form width : default="param('form_view_width')"
my $grid_count = 0;
sub view_info_folder{
	my $self    = shift;
	my %h = (
		self => $self,
		text_title_edit => string('edit'),
		is_collaps => 1,
	@_);

	$h{title}.= string($h{title_id}) if $h{title_id};
	$h{form_is_collaps} = $h{is_collaps};
	$self->{form_width}  = $h{form_width};
	
	## several objects of webpage may be created
	## thus we can't use webpage object to store counter
	my $loc_grid_count = 0;
	if($ENV{MOD_PERL}) {
		$loc_grid_count = Apache->request->pnotes('web_page_grid_count') || 0;
		$loc_grid_count++;
		Apache->request->pnotes(web_page_grid_count => $loc_grid_count);
	}
	else {
		$loc_grid_count=++$grid_count;
	}
		
	$h{grid_count}=$loc_grid_count;
	
	$h{devscreen_id}=$ENV{system_obj}->generate_screen_id;

	_process_comprep($h{comprep}, \$h{title_url}, \$h{title});

	# BAZA, don't put this c_roak here ever again. There is nothing wrong with puting separators "view_info_folder" onto edit forms. Example - deffect #2240
	# c-roak 'ERROR: error in $page->view_info_folder(),you must call $page->view_info_header() before' unless $self->{mode} eq 'viewinfo';
	
	## multilang error => force stringify;
	$h{title} = "$h{title}";
	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'view_info_folder.tmpl',\%h);
}

## => title_id
## => title
## => title_url
## => edit_url
## => text_title_edit - button title
sub view_info_plain{
	my $self    = shift;
	my %h = (
		self => $self,
		value => undef,
		@_
	);

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'view_info_plain.tmpl',\%h);
}

# => left_buttons - arrays [name, title, class, type,js_code],[name, title, class, type,js_code]...
# => right_buttons - arrays [name, title, class, type],[name, title, class, type]...
#      type  : 'button'||'submit'
#	   class  : 'SWIButtonX'|'SWIButton'|'SWButton'|undef, def=undef
# => show_cancel - flag show/ not show  , button view_name='item_cancel'
# => show_edit - flag show/ not show  , button view_name='item_edit'
# => show_delete - flag show/ not show , button view_name='item_delete'
# => cancel_class - SWIButtonX by default
# => delete_class - SWIButtonX by default
# => edit_class - SWButtonX by default
# => no_multipart - flag, if on - do not add 'enctype="multipart/form-data"' in form
# => read_only
# => url_ext
# => form_method -  get or post (by default)
# => form_url   # REQ_N #
# comments to left_buttons & right_buttons
#      js_code = any JS-code, for example qq{onClick="document.forms.item.item_submit=1;"}

sub view_info_button{
	my $self  = shift;
	my %h = (
		self => $self,
		text_title_cancel => string('cancel'),
		cancel_class => 'Button',
		text_title_delete => string('delete'),
		delete_class => 'Button',
		text_title_edit => string('edit'),
		edit_class => 'ButtonOKWide',
		confirm_question => string('sure_delete'),
		hidden=>$self->{edit_hidden},
		form_method=>'POST',
	@_);
	$h{form_url} = $self->get_browse_url.$h{url_ext} unless $h{form_url};
	$h{disabled} = 'Disabled'	if ($h{read_only});

	die qq[ERROR: error in page->view_info_button(), page must be in free mode, current mode is $self->{mode}] if defined $self->{mode};

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'view_info_button.tmpl',\%h);
}

# => button - array [js_code, title]
# => read_only
# => url_ext
# => form_url   # REQ_N #
sub view_info_single_button {
    my $self  = shift;
	my %h = (
		self   => $self,
		hidden => $self->{edit_hidden},
		@_
	);

    die q[ERROR: error in $page->view_info_single_button(),you must call $page->view_info_header() before] unless $self->{mode} eq 'viewinfo';

    $h{form_url} = $self->{common_url}.$h{url_ext} unless $h{form_url};
    $h{disabled} = 'Disabled'       if ($h{read_only});
    $self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'view_info_single_button.tmpl',\%h);
}

## >> self
## => title
## => title_id default close
sub view_info_close_button {
	my $self = shift;
	my %h = (
		title_id=>'close',
		@_
	);
	my $title;
	$title = $h{title} if ( $h{title} );
	$title = string($h{title_id}) unless ($title);
	my $th = {};
	$th->{title} = $title;
	$self->{edit_out} .= sw_web_parse_template(
		__PACKAGE__,'view_info_close_button.tmpl',$th
	);
}##/view_info_close_button


# > self
# > obj
# => namestate : state input attr name
# => namestate_alt  : custom state attr name
# => namecountry : country attr name
# => full_address
sub view_info_geo{
	my $self = shift;
	my $obj = shift;
	my %h=(
		namestate=>'state',
		namestate_alt=>'state_alt',
		namecountry=>'country',
		nameaddress1=>'address1',
		nameaddress2=>'address2',
		namehousenum=>'house_num',
		namehousesuff=>'house_suff',
		namezip=>'zip',
		namecity=>'city',
		namephone=>'phone',
		namemobile=>'mobile',
		namefax=>'fax',
		hide => [],
	@_);

	if( $h{full_address} ){
		my $full_address = $obj->{"$h{nameaddress1}"} . ' ' . $obj->{"$h{namehousenum}"}
			. $obj->{"$h{namehousesuff}"};
		$self->view_info_text(title_id=>'op_address',view_name=>$h{viewname_address1},value=>$full_address);
		$self->view_info_text(title_id=>'address_line_2',view_name=>$h{viewname_address2},value=>$obj->{"$h{nameaddress2}"})
			unless scalar (grep {$_ eq 'address2'} @{$h{hide}});
		$self->view_info_text(title_id=>'city',value=>$obj->{"$h{namecity}"})
			unless scalar (grep {$_ eq 'city'} @{$h{hide}});
	}
	## if state not hidden, show it
	unless(scalar (grep {$_ eq 'state'} @{$h{hide}})){
	    if($obj->{$h{namestate}}){
		    $self->view_info_text(
				title_id=>'state_uscanada',
				value=>HSPC::GeoData::get_state_name(
								$obj->{$h{namecountry}},
								$obj->{$h{namestate}})
			);
	    } elsif ($obj->{$h{namestate_alt}}){
		    $self->view_info_text(title_id=>'state',value=>$obj->{$h{namestate_alt}});
	    }
	}
	## if country not hidden, show it
	unless(scalar (grep {$_ eq 'country'} @{$h{hide}})){
	    $self->view_info_text(title_id=>'country',value=>HSPC::GeoData::get_country_name( $obj->{$h{namecountry}} ) );
	}
	if( $h{full_address} ){
		$self->view_info_text(title_id=>'zip_code',value=>$obj->{"$h{namezip}"});
		$self->view_info_text(title_id=>'phone',value=>$obj->{"$h{namephone}"});
		$self->view_info_text(title_id=>'mobile',value=>$obj->{"$h{namemobile}"})
			unless scalar (grep {$_ eq 'mobile'} @{$h{hide}});
		$self->view_info_text(title_id=>'fax',value=>$obj->{"$h{namefax}"})
			unless scalar (grep {$_ eq 'fax'} @{$h{hide}});
	}

}

################### /View info part ##################################


# >> self
# >  status_id  : undef|'' = do not add /sl to URL
# >  is_error
# >  url_ext
# >  status : non-localized message
sub redirect{
	my $self=shift;
	my ($status_id, $is_error, $url_ext, $status_msg) = @_;
	my $url = &sw_web_server_protocol() . '://' . &sw_web_server_name() . $self->get_browse_url();

	$url .= $url_ext;

	if( $status_id ne '' ) {
		$url .= ($is_error ? '/sle,' : '/sl,').$status_id;
	} elsif($status_msg ne '') {
		$url .= ($is_error ? '/slxe,' : '/slx,').sw_encode_base64_safe($status_msg);
		$url =~ s/\n//g;
	}

	return sw_web_redirect($url) unless $ENV{layout};
	return sw_layout_redirect($url);
}

sub redirect_comprep {
	my $self = shift;
	validate_params(@_, [
		({ type => SCALAR | UNDEF, optional => 1 }) x 5,
	]);

	$self->{common_url} = $ENV{COMPREP}->link(alias => shift);
	$self->redirect(@_);
}

# > self
# > item_object : with method get_officeitem_url($page)
# > status_id
# > is_error
# > url_ext
sub redirect_item{
	my $self=shift;
	my $obj = shift;
	my $status_id = shift;
	my $is_error = shift;
	my $url = &sw_web_server_protocol() . '://' . &sw_web_server_name . $obj->get_officeitem_url($self) . shift;
	$url.=qq{/sle,$status_id} if $is_error;
	$url.=qq{/sl,$status_id} unless $is_error;

	return sw_web_redirect($url) unless $ENV{layout};
	return sw_layout_redirect($url);
}

#	=> title_data_id
#	=> title_data
#	=> title_options_id
#	=> title_options
#	=> view_name
#	=> not_empty
#	=> data : [value,...,]
#	=> data_limit : allow to select only data_limit items, undefined by default
#	=> data_limit_warning : warning message text, not required, will be showing system message for undefined data_limit_warning
#	=> options  : [[value,title],[value,title]]
#   => options_plain=>1/0
#   => ordered  : data and options are ordered , show buttons "Up" "Down", default =1
##	=> title_data_not_quote : 1/undef
##	=> title_options_not_quote : 1/undef
##  => skip_cols : columns to skip before control
sub edit_view_lists2{
	my $self    = shift;
	my %h = (
		self => $self,
		ordered  => 1,
	@_);

	my $options = $h{options};
	my $data;
	my $index;

	unless($h{options_plain}==1){
		map{$_->[1]=string($_->[1])} @$options;
	}

	if($h{data}){
		my @options = map {$_} @$options;
		$options = \@options;
		foreach my $component_id(@{$h{data}}){
			$index = -1;
			foreach(my $i = 0;$i<=$#{$options};$i++){
				if($options->[$i]->[0] eq $component_id){
					$index = $i;
					last;
				};
			}
			if($index<0){
				sw_die("ERROR : No title for value=$component_id}");
			}
			push @$data,[$component_id,$options->[$index]->[1]];
			splice (@$options, $index, 1);
		}
	}
	$h{title_data}.=string($h{title_data_id}) if $h{title_data_id};
	$h{title_options}.=string($h{title_options_id}) if $h{title_options_id};

	$h{title_data} = $self->convert_quotes($h{title_data}) unless $h{title_data_not_quote};
	$h{title_options} = $self->convert_quotes($h{title_options}) unless $h{title_options_not_quote};

	$h{data_limit_warning} = $self->escape_quotes( $h{data_limit_warning} ) if defined $h{data_limit_warning};


	$h{data} = $data;
	$h{options} = $options;

	if($h{new}){
		if ($h{skip_cols}) {
			$self->cell_skip($h{skip_cols}>1 ? (cols=>$h{skip_cols}) : ());
		} else {
			$self->row_open();
		}
		$self->cell_text(value=>$h{title_options});
		$self->cell_text(value=>'&nbsp;');
		$self->cell_text(value=>$h{title_data});
		$self->row_close();
		return;
	}

	$self->_need_mode('edit');


	if($h{not_empty} eq '1' && $h{view_name} ne '' && (not $h{read_only}) ){
		push( @{$self->{edit_req}}, [$h{view_name},'', $h{title_data} ]);
	}


	push @{ $self->{edit_lists} },$h{view_name};
	$self->{_edit_lists_move} = 1 if $h{ordered}==1;

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_view_lists2.tmpl',\%h);
}

#	=> title_data_id
#	=> title_data
#	=> title_options_id
#	=> title_options
#	=> view_name
#	=> not_empty
#	=> data : [value,...,]
#	=> data_limit : allow to select only data_limit items, undefined by default
#	=> data_limit_warning : warning message text, not required, will be showing system message for undefined data_limit_warning
#	=> options  : [[value,title],[value,title]]
#   => options_plain=>1/0
#   => ordered  : data and options are ordered , show buttons "Up" "Down", default =1
##	=> title_data_not_quote : 1/undef
##	=> title_options_not_quote : 1/undef
##  => skip_cols : columns to skip before control
sub cell_lists2{
	my $self    = shift;
	my %h = (
		self => $self,
		ordered  => 1,
		cols=>1,
		align=>'left',
 	@_);

	my $options = $h{options};
	my $data;
	my $index;

	unless($h{options_plain}==1){
		map{$_->[1]=string($_->[1])} @$options;
	}

	if($h{data}){
		my @options = map {$_} @$options;
		$options = \@options;
		foreach my $component_id(@{$h{data}}){
			$index = -1;
			foreach(my $i = 0;$i<=$#{$options};$i++){
				if($options->[$i]->[0] eq $component_id){
					$index = $i;
					last;
				};
			}
			if($index<0){
				sw_die("ERROR : No title for value=$component_id}");
			}
			push @$data,[$component_id,$options->[$index]->[1]];
			splice (@$options, $index, 1);
		}
	}

	$h{title_data}.=string($h{title_data_id}) if $h{title_data_id};
	$h{title_options}.=string($h{title_options_id}) if $h{title_options_id};

	$h{title_data} = $self->convert_quotes($h{title_data}) unless $h{title_data_not_quote};
	$h{title_options} = $self->convert_quotes($h{title_options}) unless $h{title_options_not_quote};

	$h{data_limit_warning} = $self->escape_quotes( $h{data_limit_warning} ) if defined $h{data_limit_warning};


	$h{data} = $data;
	$h{options} = $options;

	if($h{not_empty} eq '1' && $h{view_name} ne '' && (not $h{read_only}) ){
		push( @{$self->{edit_req}}, [$h{view_name},'', $h{title_data} ]);
	}


	push @{ $self->{edit_lists} },$h{view_name};
	$self->{_edit_lists_move} = 1 if $h{ordered}==1;

	$self->row_open();

	$h{content} = sw_web_parse_template(__PACKAGE__,'cell_lists2.tmpl',\%h);

	push @{$self->{grid}->{cols}},\%h;
}

# => button - hash [name, title, js_code]
# => read_only
# => url_ext
# => form_url   # REQ_N #
sub edit_view_single_button {
        my $self  = shift;
        $self->_need_mode('edit');

        my %h = (
                self => $self,
                hidden=>$self->{edit_hidden},
        @_);
        $h{form_url} = $self->{common_url}.$h{url_ext} unless $h{form_url};
        $h{disabled} = 'Disabled'       if ($h{read_only});
        $self->{edit_out} .= sw_web_parse_template(__PACKAGE__,'edit_view_single_button.tmpl',\%h);
}


# Displays dashboard.
#
# Arguments:
#	url					=> 'url'			# (Optional) Base URL. $self->get_browse_url() is default
#	summary_title		=> 'string',		# (Optional, default: CompRep) Text string for title of summary section.
#	summary_url			=> 'url',			# (Optional, default: CompRep) Add-on to url parameter for link on title of summary section. Default is blank string.
#	summary_icon		=> 'id',			# (Optional, default: CompRep) Icon for title of summary.
#   summary_comprep		=> (Optional) see '_process_comprep'
#	summary	=> [							# Array of summary records. No summary is displayed if offited
#		{
#			title		=> 'string',		# (Optional, default: CompRep) Text string for summary records.
#			url_ext		=> 'url',			# (Optional) Add-on to url parameter for link on this record. Default is blank string.
#			full_url    => 'any_full_url',	# (Optional, default: CompRep) Url parameter for link on this action. Note that url_ext has more priority than full_url.
#			target      => '_blank'			# <a ... target="???">  target value, empty by default
#			value		=> value,			# Summary value, string or integer or bool
#                  or
#			count		=> number,			# Summary counter value,use value
#			value_title  => value,			# title for type="live"/"bool"
#			type		=> 'type',			# Summary value type,variants, default=char
#                                               live : boolean, show green or red lamp, align by right
#												bool : boolean, yes or no sign, , align by right
#												char : any HTML code, align by right
#												align => 'left' || 'right	
#			comprep		=> (Optional) see '_process_comprep'
#		},
#		...
#	],
#	action_title		=> 'string',		# Text string for title of action section.
#	action	=> [							# Array of action records. No action section is displayed if offited
#		{
#			title		=> 'string',		# Text string for action.
#			target      => '_blank'			# <a ... target="???">  target value, empty by default
#			full_url    => 'any_full_url',	# (Optional) Url parameter for link on this action.
#			url_ext		=> 'url',			# (Optional) Add-on to url parameter for link on this action. Default is blank string.
#			       or
#			comprep		=> (Optional) see '_process_comprep'
#		},
#		# OR #
#		see '_process_comprep'
#		...
#	]
#
# Returns:	nothing
#
sub dashboard {
	my $self  = shift;
	my %h = (
		url		=> $self->{common_url},
		action	=> [],
		summary	=> [],
		@_
	);

	## preprocess summary header: convert CompRep-related parameters
	_process_comprep($h{summary_comprep}, \$h{summary_url}, \$h{summary_title}, \$h{summary_icon});
	undef $h{url}
		if $h{summary_url};

	## preprocess actions: convert CompRep-related parameters
	for (my $i = 0; $i < @{$h{action}}; ++$i) {
		unless (ref $h{action}->[$i] eq 'HASH') {
			my $action = {title => '', full_url => ''};
			my $is_allowed = _process_comprep($h{action}->[$i],
				\$action->{full_url}, \$action->{title});
			$h{action}->[$i] = $action;
			splice @{$h{action}}, $i--, 1
				unless $is_allowed;
		} elsif($h{action}->[$i]->{comprep}) {
			my $is_allowed = _process_comprep($h{action}->[$i]->{comprep},
				\$h{action}->[$i]->{full_url}, \$h{action}->[$i]->{title});
			splice @{$h{action}}, $i--, 1
				unless $is_allowed;
		}
	}

	## preprocess summary: convert CompRep-related parameters
	for (my $i = 0; $i < @{$h{summary}}; ++$i) {
		my $is_allowed = _process_comprep($h{summary}->[$i]->{comprep},
			\$h{summary}->[$i]->{full_url}, \$h{summary}->[$i]->{title});
		splice @{$h{summary}}, $i--, 1
				unless $is_allowed;
	}

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__, 'dashboard.tmpl', \%h);		
}

##############################################################

sub row_open{
	my $self = shift;
	sw_die("ERROR : You can call edit_open() before row_open()") unless $self->{mode}=~m/grid|edit/;

	return if $self->{grid}->{row_size}>=0;

	$self->_grid_open() unless defined $self->{grid};
	$self->_row_reset();
}

# >=self_table : always open table
# >=open
# >=close
# >=cellspacing, def_value=3
# >=cellpadding, def_value=>0
sub row_close{
	my $self = shift;

	return undef unless defined $self->{grid}->{cols};

	my %h=(
		cols=>$self->{grid}->{cols},
		cellspacing=>undef,
		cellpadding=>undef,

	@_);
	$h{cellspacing} = '3' unless defined $h{cellspacing};
	$h{cellpadding} = '0' unless defined $h{cellpadding};

	$h{size} = $self->_row_size();

	push @{$self->{grid}->{rows}},\%h;
	$self->_row_reset();
}


# =>> view_name
# =>> value
# =>  align, default=center
# =>  js_onclick
# =>  check_form : check form filling on click
# =>  class  : button class, default = 'SWIButton' <<<
# =>  btn_type : : button type, default = 'submit'
sub cell_button{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align => 'center',
		class => 'Button',
		btn_type => 'submit',
		tag_id => 'item',
	@_);
	
	my $tag_id = $th{tag_id};
	
	if( $th{check_form}==1 ){
		$th{js_onclick}.= qq?
document.getElementById('__caller_index').value = get_index_${tag_id}(document.getElementById('$th{view_name}')); document.getElementById('__button_skip').value = 0;
		?;
	}else{
		$th{js_onclick}.= qq?
document.getElementById('__caller_index').value = get_index_${tag_id}(document.getElementById('$th{view_name}')); document.getElementById('__button_skip').value = 1;
		?;
	}
	sw_die('ERROR : you must define view_name=> in cell_input() call')
		unless defined $th{view_name};
	sw_die("ERROR : you must define value=> in ".
		"cell_input(view_name=>'$th{view_name}') call")
		unless defined $th{value};

	# PROCESS TEMPLATE #
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_button.tmpl',\%th);

	# PUSH COL #
	push @{$self->{grid}->{cols}},\%th;
};

# =>> view_name
# =>  view_id
# =>  value
# =>  read_only
# =>  enable_views : array of view_names
# =>  disable_views : array of view_names
# =>  select_views : hash of view_names (checkboxes only at the moment)
#     	=>> box_name : view_name
#	  	=>> action : 'clear' (only for checkboxes this time), 'disable', 'enable'
# =>  unselect_views : hash of view_names  (checkboxes only at the moment)
#     	=>> box_name : view_name
#	  	=>> action : 'clear' (only for checkboxes this time), 'disable', 'enable'
# =>  align : default = center
# =>  disabled : 1 or undef
# =>  is_checked  : if empty, then use value as is_checked flag
# =>  js_on_change : any JS code for <input ... onChange=""> code
# =>  js_on_click : any JS code for <input ... onClick=""> code
# =>  js_on_focus : any JS code for <input ... onFocus=""> code
sub cell_check{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align=>'center',
		data=>{
			self=>$self
		},
	@_);

	$th{view_index} = $self->{_edit_view_index}++;
	$th{src_is_checked} = $th{is_checked};
	$th{is_checked} = ($th{value}?1:0) unless defined $th{is_checked};

	sw_die('ERROR: you must define view_name=> for WebPage->cell_check() call')
		unless defined $th{view_name};

	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_check.tmpl',\%th);

	push @{$self->{grid}->{cols}},\%th;

	if( $th{is_checked} && $#{$th{disable_views}}>=0){
		foreach ( @{$th{disable_views}} ){
			$self->{edit_disable_views}->{$_} = 1;
		}
	}
	if( !$th{is_checked} && $#{$th{enable_views}}>=0){
		foreach (@{$th{enable_views}}){
			if( $th{is_checked} ) {
				## view should not be disabled in any case
				$self->{edit_enable_views}->{$_} = 1;
			} elsif (!$self->{edit_enable_views}->{$_}) {
				$self->{edit_disable_views}->{$_} = 1;
			}
		}
	}

	# STORE CONTROL TYPE #
	$self->edit_view_hidden(view_name=>'_type_'.$th{view_name},value=>'check');
}

# => check_title|check_title_id # REQ_N #
# => view_name # REQ_Y #
# OPTIONS AS 2D-ARRAY
# =>> options   # REQ_Y # ref to [[value,res_id],[value1,res_id1]]
# /OPTIONS AS 2D-ARRAY
# OPTIONS AS HASHES ARRAY
# =>> options       : ref to hash
#	 =>> data  		: ref to hashes array
#	 =>> attr_value : value attribute name
#	 =>> attr_title : title attribute name
# /OPTIONS AS HASHES ARRAY
# => options_plain
# => align, default=center
# => value
# => selector_title : use it to replace default '-- Select --' at the top of the list;
# => read_only
# => design_wifdth
# => javascript
# =>  js_on_change : any JS code for <select ... onChange=""> code
# =>  js_on_click : any JS code for <select ... onClick=""> code
# =>  js_on_focus : any JS code for <select ... onFocus=""> code
# => disbled
# => no_default : if set no default string '-- Select --' will be added to the list of options
sub cell_combo{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align => 'left',
		idn_convert => 1,
	@_);

	sw_die('ERROR: you must define view_name=> for WebPage->cell_combo() call')
		unless defined $th{view_name};
	sw_die('ERROR: you must define options=> for WebPage->cell_combo() call')
		unless defined $th{options};
	$th{check_title} = string( $th{check_title_id} ) if $th{check_title_id};

	# copy options #
	require HSPC::Util::DM qw(prepare_idn_value);
	if(ref($th{options}) eq 'ARRAY'){
		my @options = @{$th{options}};
		@options = map {[$_->[0],prepare_idn_value($_->[1])]} @options
			if ($th{idn_convert});
		$th{options} = \@options;
	}else{
		my @options;
		my $data = $th{options}->{data};
		my $attr_value = $th{options}->{attr_value};
		my $attr_title = $th{options}->{attr_title};
		foreach(@$data){
			push @options,[
				$_->{$attr_value},
				$th{idn_convert} ?
					prepare_idn_value($_->{$attr_title}) : $_->{$attr_title}
			];
		};
		$th{options} = \@options;
	}
	map{$_->[1] = string($_->[1])} @{$th{options}}
		if not defined $th{options_plain};
	#

	# CHECK FOR EMPTY #s
	if($th{check_title} &&  (not $th{read_only})){
		push( @{$self->{edit_req}}, [$th{view_name},'',$th{check_title}]);
	}

	if (!$th{no_default}) {
		my $selector_title = $th{selector_title} || string('select_dd');
		if( ($th{options}->[0]->[0] ne '') &&
		 ( ( $th{value} eq '') || (not defined $th{check_title}) )
		){
			splice @{$th{options}},0,0,[undef,$selector_title];
		}
	}

	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_combo.tmpl',\%th);
	push @{$self->{grid}->{cols}},\%th;
}


# => check_title|check_title_id # REQ_N #
# => view_name   # REQ_Y #
# => size
# =>  js_on_change : any JS code for <input ... onChange=""> code
# =>  js_on_click : any JS code for <input ... onClick=""> code
# =>  js_on_focus : any JS code for <input ... onFocus=""> code
# =>  js_on_keydown : any JS code for <input ... onKeyDown=""> code
sub cell_file{
	my $self = shift;

	die 'ERROR : You can use edit_view_file() only in multipart form. User edit_open(is_multipart=>1).'unless $self->{_form_multipart};

	my %th=(
		cols=>1,
		size => 30,
		align => 'left',
	@_);

	sw_die('ERROR: you must define view_name=> for WebPage->cell_file() call')
		unless defined $th{view_name};

	$th{check_title} = string( $th{check_title_id} ) if $th{check_title_id};

	# CHECK FOR EMPTY #s
	if($th{check_title} &&  (not $th{read_only}) ){
		push( @{$self->{edit_req}}, [$th{view_name},'',$th{check_title}]);
	}

	# GENERATE AND STORE CONTENT #
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_file.tmpl',\%th);
	push @{$self->{grid}->{cols}},\%th;
}

# CLASSIC INTERFACE
# =>> view_name
##=>> value if multi_lang == 1 shold be hash {lang=>value}
##=>>multi_lang 1/0
##=>>default_lang
# =>> max_length (not mandatory for textarea)
# => warn_title (for textarea, if max_length is defined)
# /CLASSIC INTERFACE
# MODERN ABSTRACT INTERFACE
# =>> obj_attr=>string with object attribute name
#         auto-setting view_name=>,value=>,max_length=>,not_empty=> argmuments
# =>> object : use this ref instead ref from edit_open(object=>)
# /MODERN ABSTRACT INTERFACE
# =>  check_title|check_title_id # REQ_N #
# =>  read_only
# =>  is_password
# =>  size     : default value is 30 chars
# =>  rows     : must be defined for textarea
# =>  prefix
# =>  suffix
# =>  descr
# =>  class
# =>  width    : control width in pixels
# =>  wrap     : <textarea wrap="..."> value, variants : undef/1/soft/hard
# =>  format   : description is before _edit_format function;
# =>  js_on_change : any JS code for <input ... onChange=""> code
# =>  js_on_click : any JS code for <input ... onClick=""> code
# =>  js_on_focus : any JS code for <input ... onFocus=""> code
# =>  js_on_keydown : any JS code for <input ... onKeyDown=""> code
# =>  allow_trailspaces  : allow to input spaces at the start and at the end of line
##=> langs - [['en','English'], ... ]
##=> langs_localized => 1/0 - convert strings in template
##	=>	rich_editor 1/0 (for HTML creation in WYSIWYG mode, requires rows > 1)
sub cell_input{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align => 'left',
		multi_lang=>0,
		default_lang=>undef,
		langs=>undef,
		data=>{
			self=>$self
		},
		idn_convert => 1,
		langs_localized => 0,
		@_
	);

	## check vendor's RichEditor option
	$th{rich_editor} &&= HSPC::MT::Core::Setup
		->new(provider_id => $ENV{security_obj}->account_no)
		->enable_richeditor;

	$th{value} = dclone($th{value}) ## save original hash! keep it untouched!
		if ref($th{value}) eq 'HASH';
	if ( $th{multi_lang} ) {
		$th{value} = $th{value}->get_hash ## support for object as value
			if $th{value} && ref $th{value} eq 'HSPC::MT::LangValue';
		$th{langs} = sw_sorted_lang_options($ENV{security_obj}->account_no()) unless (
			$th{langs}
		);
		$th{langs} = [ grep { $_->[0] } @{$th{langs}} ]; ## throw away missing languages
		$th{default_lang} = $th{langs}->[0]->[0];
		if ( !$th{not_quote} && ref($th{value}) eq 'HASH' ) {
			foreach my $l (keys(%{$th{value}})) {
				$th{value}->{$l} = $self->convert_quotes($th{value}->{$l});
			}
		}
	} else {
		$th{value} = $self->convert_quotes($th{value}) unless $th{not_quote};
	}

	sw_die(
		"HSPC::WebPage->cell_input: default_lang is required when multi_lang is true"
	) if ( $th{multi_lang} && !defined($th{default_lang}) );
	

	if(defined $th{obj_attr}){
		my $object =  $th{object}||$self->{edit_object};
		die unless $object;

		$th{view_name} = $th{obj_attr};
		$th{value} = $object->{  $th{obj_attr}   }; #TODO :use real setter
		my $col = $object->_get_class_info->{props}->{ $th{obj_attr} };

		if(not defined $th{max_length}){
			my $max = $col->{max_length};
			unless($max){
				sw_die(
					qq[ERROR: page->edit_view_input(), in CLASS_INFO ].
					qq[undefined or zero "max_length" value for attribute $th{obj_attr}  ]);
			}
			$th{max_length} = $max;
		}
	}


	$th{size} = $th{rows}?46:30 unless defined $th{size} ;
	$th{read_only} = 1 if $self->{read_only};

	sw_die('ERROR : you must define view_name=> in cell_input() call')
		unless defined $th{view_name} || $th{read_only};
	sw_die("ERROR : 'max_length' must be defined and non-zero in ".
		"cell_input(view_name=>'$th{view_name}') call")
		unless $th{max_length} || $th{read_only};

	$th{input_type} = $th{is_password}?'password':'text';
	$th{check_title} = string( $th{check_title_id} ) if $th{check_title_id};

	# PROCESS FORMAT IF SPECIFIED #
	if(exists($th{format}) && ref($th{format}) && !int $th{rows}){
		$self->{_edit_formats}  ||= {};
		$th{format}->{field_title} ||= $th{check_title};
		$self->_edit_format(@th{qw(view_name format)});
	}

	my $need_convert = 0;
	# convert float value to float format #
	if ($th{format} && $th{format}->{type} eq 'float'
		&& defined($th{format}->{precision})
		&& defined($th{value})
		&& $th{value} ne ''
	){
		$th{value} = sprintf("%.".$th{format}->{precision}."f", $th{value});
		$need_convert = 1;
	} elsif (defined $th{type}  && $th{type} eq 'percent') {
		require HSPC::Math qw(round2);
		$th{value} = round2($th{value});
		$need_convert = 1;
	} elsif (defined $th{type}  && $th{type} eq 'float') {
		$need_convert = 1;
	}

	if ($need_convert) {
		$self->edit_view_hidden(view_name=>'_vendorid_'.$th{view_name},value=>$th{vendor_id});
	}

	# CHECK FOR EMPTY #
	if(($th{check_title} ne '') &&  (not $th{read_only}) ){
		push( @{$self->{edit_req}}, [$th{view_name},'',$th{check_title}]);
	}

	# PROCESS FORMAT IF SPECIFIED #
	if(exists($th{format}) && ref($th{format}) && !int $th{rows}){
		$self->{_edit_formats}  ||= {};
		$th{format}->{field_title} ||= $th{check_title};
		$self->_edit_format(@th{qw(view_name format)});
		$th{step} = $th{format}->{step} > 1 ? $th{format}->{step} : undef;
	}

	# STORE CONTROL TYPE #
	push @{$self->{_edit_inputs}},$th{view_name};
	my $control_type;
	if ( defined($th{rows}) ) {
		$control_type = 'textarea';
	} else {
		$control_type = 'input';
	}
	$th{type} = 'lang_'.$control_type if ( $th{multi_lang} );
	$self->edit_view_hidden(view_name=>'_type_'.$th{view_name},value=>$th{type});
	if( $th{allow_trailspaces} ){
		$self->edit_view_hidden(view_name=>'_sys_'.$th{view_name}.'_allow_trailspaces',value=>1);
	}
	## Update value only for inputbox
	unless($th{rows}) {
		$th{value} =~ s/"/&quot;/gs;
	}
	require HSPC::Util::DM qw(prepare_idn_value);
	$th{value} = prepare_idn_value($th{value}) if ($th{idn_convert});
	# PROCESS TEMPLATE #
	my $tmpl = 'cell_input.tmpl';
	$tmpl = 'cell_input_multi.tmpl' if ( $th{multi_lang} );
	$th{content} = sw_web_parse_template(__PACKAGE__,$tmpl,\%th);

	# PUSH COL #
	push @{$self->{grid}->{cols}},\%th;
};

# => value
# => is_aster
# => bold
# => cols
# => width : guaranteed screen width
# => not_quote
# => image_id
sub cell_text{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align=>'left',
		link_target=>'_self',
		idn_convert => 1,
	@_);

	## form contain asterix;
	require HSPC::Util::DM qw(prepare_idn_value);
	$self->{'has_aster'}	= 1	if $th{'is_aster'};
	$th{value} = $self->convert_quotes($th{value}) unless $th{not_quote};
	$th{value} = prepare_idn_value($th{value}) if ($th{idn_convert});

	$self->{'nowrap'} = $th{nowrap} if $th{nowrap};

	$th{value} = "<b>$th{value}</b>" if $th{bold};
	$th{value} = "<i>$th{value}</i>" if $th{italic};
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_text.tmpl', {cell => \%th});

	push @{$self->{grid}->{cols}},\%th;
}

# => value
# => is_aster
# => cols
# => width : garanted screen width
sub cell_content{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols=>1,
		align=>'left',
	@_);

	## form contain asterix;
	$self->{'has_aster'}	= 1	if $th{'is_aster'};

	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_text.tmpl', {cell => \%th});

	push @{$self->{grid}->{cols}},\%th;
}


# => check_title|check_title_id # REQ_N #
# => view_name   # REQ_Y #
# => align, default=center
# => read_only
# =>  js_on_add : any JS code trggering by ADD click
# =>  js_on_remove : any JS code  trggering by REMOVE click
# => selected : ref to array of id which will be selected
# DATA AS 2D-ARRAY
# =>> data : ref to array with list rows, array consist of [id,title]
# /OPTIONS AS 2D-ARRAY
# DATA AS ARRAY OF REFAS TO HASHES
# =>> data       : ref to hash with next attributes
#	 =>> data  		: ref to hashes array
#	 =>> attr_id : value attribute name
#	 =>> attr_title : title attribute name
# /DATA AS HASHES ARRAY
# => rows : area height in rows # NOT_REQ, default = 10
# => btns : ref to array with buttons for list operations, array consist of ref to hash with btn info
# => list_width : list width, in pixels,  default = 200
####################################
#	btns setup format:
#	  custom button #
#    => #title : button title
#    => name : button name
#    => script : JS-script for button pressed
#    => b_width : pixels, 80 by default.
#    => b_class : CSS,  SWIButtonX by default
# 	  /custom button #
#	  system button #
#    =>> type : button type,variants
#			'remove' : remove selected items
#			'add' : create popup window for adding new items
#			'up' : move selected items up
#			'down' : move selected items down
#    => popup_url : for type=>'add'  : popup URL
#    => popup_width : for type=>'add'  : popup window width ,default=500
#    => popup_height : for type=>'add'  : popup window height,default=400
#	  /system button #
sub cell_list{
	my $self = shift;
	my %th=(
		cols=>1,
		align => 'left',
		rows=>5,
		list_width=>200,
		post_scripts=>$self->{page}->{edit_lists_scripts},
	@_);

	my %selected;
	if ( $th{selected} ) {
	    my $data = $th{data};
		local ($_);
		foreach ( @{$th{selected}} ) {
			$selected{$_} = 'selected';
		}
	}
	$th{selected} = \%selected;

	sw_die('ERROR : you must define view_name=> in cell_list() call')
		unless defined $th{view_name};

	$th{check_title} = string( $th{check_title_id} ) if $th{check_title_id};

	if(not defined $th{data}){
	}elsif(ref($th{data}) eq 'ARRAY'){
	}else{
		# CONVERT DATA FROM HASH TO ARRAY
		my @data;
		my $data_hashes = $th{data}->{data};
		my $attr_id = $th{data}->{attr_id};
		my $attr_title = $th{data}->{attr_title};

		sw_die('$attr_id or $attr_title not undefined') unless $attr_id && $attr_title;

		foreach(@$data_hashes){
			push @data,[
				$_->{$attr_id},$_->{$attr_title}
			];
		};
		$th{data} = \@data;
	}

	# CHECK FOR EMPTY #
	if($th{check_title} &&  (not $th{read_only}) ){
		push( @{$self->{edit_req}}, [$th{view_name},'',$th{check_title}]);
	}

	# POST CONTROL_TYPE #
	if($th{read_only}){
	}else{
		push @{ $self->{edit_lists} },$th{view_name};
		# STORE CONTROL TYPE #
		$self->edit_view_hidden(view_name=>'_type_'.$th{view_name},value=>'list');
		#$self->edit_view_hidden(view_name=>qq{$th{view_name}__list},value=>1);
	}

	$self->{edit_lists_scripts} = 1;
	my @order_btns = grep $_->{type} eq 'up' || $_->{type} eq 'down',@{$th{btns}};
	$self->{_edit_lists_move} = 1 if $#order_btns == 1;

	# GENERATE AND STORE CONTENT #
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_list.tmpl',\%th);
	push @{$self->{grid}->{cols}},\%th;
}

# =>> view_name
# =>  view_id
# =>  align, default=center
# =>  value
# =>  disabled
# =>  is_checked
# =>  enable_views : array of view_names
# =>  disable_views : array of view_names
# =>  select_views : hash of view_names (checkboxes only at the moment)
#     	=> box_name : view_name
#	  	=> action : 'clear', 'enable', 'disable','select'
#### javascript-code #######
# =>  js_code : any JavaScript-code, will be posted in <input...._here_...>
#         or
# =>  js_on_change : any JS code for <input ... onChange=""> code
# =>  js_on_click : any JS code for <input ... onClick=""> code
# =>  js_on_focus : any JS code for <input ... onFocus=""> code
#### /javascript-code #######
sub cell_radio{
	my $self = shift;

	my %th=(
		cols=>1,
		align=>'center',
		view_index=>$self->{_edit_view_index}++,
	@_);

	die 'ERROR : you must define view_name=> in cell_radio() call' unless defined $th{view_name};
	die 'ERORR : call_radio(), you can`t use js_code=> and other js_on_*=> arguments togather' if defined $th{js_code} && (defined $th{js_on_change}||defined $th{js_on_focus} || defined $th{js_on_click});

	if( $th{is_checked} && $#{$th{disable_views}}>=0){
		foreach ( @{$th{disable_views}} ){
			$self->{edit_disable_views}->{$_} = 1;
		}
	}
	if( !$th{is_checked} && $#{$th{enable_views}}>=0){
		foreach (@{$th{enable_views}}){
			$self->{edit_enable_views}->{$_} = 1;
		}
	}

	# GENERATE AND STORE CONTENT #
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_radio.tmpl',\%th);
	push @{$self->{grid}->{cols}},\%th;
}


#	=>>	view_name
#       =>>     title
#	=>	value
#	=>	read_only
#	=>	hide_time
#	=>	hide_seconds
#	=>	hide_day
#	=>	hide_month
#	=>	hide_year
#
sub cell_period{
	my $self = shift;
	$self->row_open();

	my %th=(
		cols         => 1,
		align        => 'left',
		hide_time    => 0,
		hide_seconds => 0,
		hide_day     => 0,
		hide_month   => 0,
		hide_year    => 0,
		check_title  => '',
		title        => 'Period',
		format       => {
				type     => 'float', precision => 0,
				interval => [ [ 1, 100000, 1, 1 ] ]
			},
		@_
	);

	$th{check_title} = $th{title};
	$th{value} = sprintf("%.".$th{format}->{precision}."f", $th{value});
	
	# PROCESS FORMAT IF SPECIFIED #
	if( exists( $th{ format } ) && ref( $th{ format } ) ) {
		$self->{_edit_formats}  ||= {};
		$th{format}->{field_title} ||= $th{check_title};
		$self->_edit_format(('number_'.$th{view_name}), $th{format});
		$th{step} = $th{format}->{step} > 1 ? $th{format}->{step} : undef;
	}

	if ($th{value} % 31104000 eq 0) {
		$th{interval} = 5;
		$th{number} = $th{value} / 31104000;
	}elsif ($th{value} % 2592000 eq 0) {
		$th{interval} = 4;
		$th{number} = $th{value} / 2592000;
	}elsif ($th{value} % 86400 eq 0) {
		$th{interval} = 3;
		$th{number} = $th{value} / 86400;
	}elsif ($th{value} % 3600 eq 0) {
		$th{interval} = 2;
		$th{number} = $th{value} / 3600;
	}elsif ($th{value} % 60 eq 0) {
		$th{interval} = 1;
		$th{number} = $th{value} / 60;
	}else {
		$th{interval} = 0;
		$th{number} = $th{value};
	}

	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_period.tmpl',\%th);

	push @{$self->{grid}->{cols}},\%th;
}

#   =>  check_title|check_title_id # REQ_N #
#	=>>	view_name      #REQ_Y#
#	=>	value_title    #REQ_N#
#	=>	value          #REQ_N#
#   =>  is_show_value  ## hidden / text input
#   =>  popup_url      #REQ_Y#
#	=>  read_only      #REQ_N#
#   =>  popup_height   #REQ_N#
#   =>  popup_width    #REQ_N#
#   =>  js_view_code   ## only if is_show_value, its adding for onKeyDown, onKeyPress handlers
#   =>  js_prefix_code #REQ_N#
#   =>  js_on_change : any JS code for <input ... onChange=""> code
sub cell_popup{
	my $self = shift;

	my %th = (
		cols=>1,
		align=>'left',
		popup_height=>450,
		popup_width=>400,
		is_show_value=>0,
	@_);

	die 'ERROR : you must define view_name=> in cell_popup{() call' unless defined $th{view_name};

	# quote js-code #
	## commented due to MAC Safari support
	## $th{js_prefix_code} = _s2js($th{js_prefix_code}) if defined $th{js_prefix_code};
	$th{js_on_change} = _s2js($th{js_on_change}) if defined $th{js_on_change};
	$th{js_view_code} = _s2js($th{js_view_code}) if defined $th{js_view_code};

	# expand URL #
	$th{popup_url}.="/owner_form,item/owner_view,$th{view_name}/owner_list,0";

	# MAKE HIDDEN VALUE #
	$self->edit_hidden($th{view_name}=>$th{value}) unless $th{is_show_value};

	$th{value_title} = $self->convert_quotes($th{value_title}) unless $th{not_quote};

	# GENERATE AND STORE CONTENT #
	$th{content} = sw_web_parse_template(__PACKAGE__,'cell_popup.tmpl',\%th);
	push @{$self->{grid}->{cols}},\%th;

	# CHECK FOR EMPTY #
	my $check_name = $th{is_show_value} ? $th{view_name}:'title_'.$th{view_name};
	if(($th{check_title} ne '') &&  (not $th{read_only}) ){
		push( @{$self->{edit_req}}, [$check_name,'',$th{check_title}]);
	}
}

# => width : garanted screen width
sub cell_skip{
	my $self = shift;

	$self->cell_text(value=>'&nbsp;',not_quote=>1,@_);
}

sub cell_grid_open{
	my $self = shift;

	$self->_grid_open(@_);
}


sub cell_grid_close{
	my $self = shift;

	my $content = $self->grid_content(embedded=>1);
	$self->_grid_close();
	$self->cell_text(value=>$content,not_quote=>1);
}

############################################################################

sub grid_push{
	my $self = shift;

	return if $#{$self->{grid}->{rows}}<0;
	my %th=(
		self	     => $self,
		rows	     => $self->{grid}->{rows},
		form_cols  	 => $self->{_edit_form_cols},
		tag_id       => $self->{form_tag_id} || 'item',
		grid_count	 => ++$self->{grid_count},
		devscreen_id => $ENV{system_obj}->generate_screen_id,
		form_width_real	 => $self->{form_width_real},
	);

	$self->{edit_out} .= sw_web_parse_template(__PACKAGE__, 'grid.tmpl', \%th);
	$self->_grid_close();
}

# >= embedded
sub grid_content{
	my $self = shift;

	return if $#{$self->{grid}->{rows}}<0;
	my %th=(
		self	=> $self,
		rows	=> $self->{grid}->{rows},
		grid_count	=> ++$self->{grid_count},
		tag_id  => $self->{form_tag_id} || 'item',
		devscreen_id=> $ENV{system_obj}->generate_screen_id,
		form_width_real	 => $self->{form_width_real},		
		@_
	);
	return sw_web_parse_template(__PACKAGE__, 'grid.tmpl', \%th);
}

############################################################################

sub _row_reset{
	my $self = shift;

	$self->{grid}->{cols} = undef;
	$self->{grid}->{row_size} = 0;
	$self->{mode} = 'grid';
}

sub _row_size{
	my $self = shift;

	my $count=0;
	foreach my $col( @{$self->{grid}->{cols}} ){
		$count+=$col->{cols} if (not defined $col->{open}) || $col->{open}==1;
	}
	return $count;
}

sub _grid_open{
	my $self = shift;

	# PUSH EXISTED GRID TO STACK #
	if($self->{grid}){
		push @{$self->{grids}},$self->{grid};
		$self->{grid} = undef;
	}

	$self->{grid} = {grid_row_size=>0,@_};
}

sub _grid_close{
	my $self = shift;

	# delete current grid #
	$self->{grid} = undef;

	# get grid from stack #
	my $stack_last = $#{$self->{grids}};

	if($stack_last>=0){
		$self->{grid} = $self->{grids}->[ $stack_last  ];
############################################################################
		splice @{$self->{grids}},$stack_last,1;
		$self->{mode} = 'grid';
	}else{
		$self->{mode} = 'edit';
	}
}

############################################################################

# > self
# > mode
sub _need_mode{
	my $self = shift;
	my $mode = shift;

	return if $mode eq $self->{mode};

	if($mode eq 'edit'){
		if( $self->{mode} eq 'grid'){
			$self->grid_push();
		}else{
			#die; #TODO_BAZA : uncomment DIE
		}
	}
}

############################################################################
## Params                :
##       1               - field name;
##       2               - format itself (HASH or some object ref).
## Format params :
##       type            - format type:
##       field_title     - string - field title;
##       precision       - number of decimal places;
##       step            - quantization step of discrete integer value;
##       interval        - optional ref to ARRAY of avaliable intervals.
##							every interval is a ref TO ARRAY with interval info,
##							[	interval_left_border_value;
##								interval_right_border_value;
##								interval_left_border;
##								interval_right_border;
##							] only two first elements are required;
## Possible interval borders:
##	0	- open;
##	1	- closed;
##	2	- not used.
##
## Possible format types :
##       int             - is Integer. Without params;
##       float           - is Float. Uses optional param 'precision' - number of decimal places;
##       regex           -
##
##
use constant	LEFT_BORDER	=> 2;
use constant	RIGHT_BORDER	=> 3;

sub _edit_format{
        my ($self, $name, $format)      = @_[0..2];
        return if($name eq '' || !ref($format));
        return if exists $self->{_edit_formats}->{$name};

        undef $@;
        eval {$format->{type}   = lc $format->{type}};
        sw_die("ERROR in HSPC::WebPage->_edit_format(): wrong format reference type$/") if $@;

SWITCH:{
                if ($format->{type} eq 'int'){
                        last SWITCH;
                }
                if ($format->{type} eq 'float') {
                        $format->{precision}    = abs(int $format->{precision}) if exists $format->{precision};
                        last SWITCH;
                }
                if ($format->{type} eq 'regex') {
                        last SWITCH;
                }
				# Unknown format
                sw_die("ERROR in HSPC::WebPage->_edit_format(): unknown format type $format->{type}$/");
        }

        if (exists($format->{interval}) && 'ARRAY' eq ref($format->{interval}) && @{$format->{interval}}) {
                my(@temp, $unparsed, $i);
                foreach (@{$format->{interval}}){
                        $!      = 0;
                        @$_     = sw_trim(@$_);
                        while (@$_ > 1 && @$_ < 4){
                                push @$_, 1;
                        }

                        if (@$_ == 4){
                                for $i (0..1){
                                        ($_->[$i], $unparsed)   = strtod($_->[$i]);
                                        ($! || $unparsed) && next;
                                }
##                                next if $_->[0] > $_->[1];
                                for $i (LEFT_BORDER, RIGHT_BORDER){
                                        $_->[$i]        = abs int $_->[$i];
					($_->[$i] > 2)	&& ($_->[$i]	= 0);
##                                      ($_->[$i]== 2)  && ($_->[5-$i]  = 0);
                                }
				($_->[LEFT_BORDER] == 2 && $_->[RIGHT_BORDER] == 2)	&& next;
				($_->[LEFT_BORDER] != 2 && $_->[RIGHT_BORDER] != 2)	&& ($_->[LEFT_BORDER] = $_->[RIGHT_BORDER] = 1);
                                push @temp, $_;
                        } elsif (@$_ == 1){
                                ($_->[0], $unparsed)    = strtod($_->[0]);
                                ($! || $unparsed) && next;
                                push @temp, $_;
                        } else {
				next;
                        }
                }
                if (@temp){
                        $format->{interval}     = [@temp];
                } else {
                        delete $format->{interval};
                }
        }

        $self->{_edit_formats}->{$name} = $format;
}

# > original js_code
# < quoted js_code
sub _s2js{
	local($_,$1,$`,$',$&);
	my $jsesc={"\\"=>"\\\\",
	        "\n"=>"\\n",
	        "\r"=>"\\r",
	        "\t"=>"\\t",
	        "?"=>"\\?",
	        "\""=>"\\\"",
	        "\'"=>"\\\'",
	        ">"=>"\\x3e",
	        "<"=>"\\x3c"
	       };
	$_=$_[0];
	s/([\\?\n\r\t<>"'])/${$jsesc}{$1}/ge;
	return $_;
}

sub convert_quotes{
	my $self = shift;
    my $line = shift;
    $line=~s/"/\&quot;/g;##"
    $line=~s/'/\&#39;/g;##'
    $line=~s|\<txt_(.+?)\>|~~txt_$1~~|g;
    $line=~s|\<|\&lt\;|g;
    $line=~s|\>|\&gt\;|g;
    $line=~s|~~txt_(.+?)~~|<txt_$1>|g;
    return $line;
}

## Convert: "backslash" ' and " special for HSPC::Billing::HPPopup screens!
sub convert_quotes_popup{
	my $self = shift;
    my $line = shift;
    $line =~ s/\"/\\\"/g;
    $line =~ s/\'/\\\'/g;
    $line=~s|\<txt_(.+?)\>|~~txt_$1~~|g;
    $line=~s|\<|\&lt\;|g;
    $line=~s|\>|\&gt\;|g;
    $line=~s|~~txt_(.+?)~~|<txt_$1>|g;
    return $line;
}

## >> self
## >> value
sub escape_quotes
{
	my $self = shift;
    my $line = shift;
	$line =~ s|\\|\\\\|g;
    $line =~ s/\"/\\\"/g;
    $line =~ s/\'/\\\'/g;
    return $line;
}

## Try to refill 'link', 'title' and 'icon' with data from CompRep, according to 'comprep' structure:
##	[
##		'alias',	# Node alias in CompRep, link with URI params:
##		p1 => 'v1',	# (Optional) URI param
##		p2 => 'v2',	# (Optional) URI param
##		...
##	],
##	# OR #
##	'alias',		# Node alias in CompRep, link without URI params.
sub _process_comprep {
	validate_params(@_,[
		{ type => SCALAR | ARRAYREF | UNDEF },			## comprep
		{ type => SCALARREF | UNDEF },					## link
		{ type => SCALARREF | UNDEF, optional => 1 },	## title
		{ type => SCALARREF | UNDEF, optional => 1 },	## icon
		]
	);
	return 1
		unless $_[0] && $ENV{COMPREP};

	my $node = $ENV{COMPREP}->node(alias => ref $_[0] ? shift @{$_[0]} : $_[0]);
	if ($node) {
		my $is_allowed = $ENV{security_obj}->test_security($node->{alias});
		${$_[1]} ||= $ENV{COMPREP}->link(node_id => $node->{node_id}, ref $_[0] ? $_[0] : ())
			if defined $_[1] && $is_allowed;
		${$_[2]} ||= string($node->{title_id}, %{$node->{title_hash}})
			if defined $_[2];
		${$_[3]} ||= $node->{icon}
			if defined $_[3];
		return $is_allowed;
	} else {
	 	return undef;
	}

	return 1;
}
## Generate unique identifier to store screen's upref
##
sub _upref_alias {
	my $self = shift;
	return $ENV{COMPREP} && $ENV{COMPREP}->node_alias
		|| $ENV{layout} && $ENV{layout}->devscreen_id()
		|| $ENV{system_obj}->generate_screen_id
			. 
			( 
			$self->{_title_called} 
			? $self->{_title_called}->{devscreen_id}
			: '' 
			);
}

## Manage sticky search form params, try to get value
##  1. user entered
##  2. uri_param, then get form upref 
## >> class
## => arg_name param name
## => uri_name param name
## => secs flag to conver uri_param with secs_to_time
## => is_undef flag to undef '' return value
## < value
sub sticky_param {
	my $self = shift;
	my %h = (
		arg_name	=> undef,
		uri_name	=> undef,
		secs		=> undef,
		is_undef	=> undef,
	@_);
	sw_die('arg_name should be provided for sticky_param')
		unless $h{'arg_name'};

	$h{'uri_name'} ||= $h{'arg_name'};

	my $form_value = ref($h{'arg_name'}) eq 'ARRAY' 
		? sw_argparam(@{$h{'arg_name'}})
		: sw_argparam($h{'arg_name'});
	my $uri_value = sw_uriparam($h{'uri_name'});

	$uri_value = secs_to_time($uri_value) if $h{secs} && $uri_value ne '';

	if ($uri_value){
		$form_value = $uri_value unless defined $form_value;
	} elsif ($ENV{system_obj}->get_user_option(option => 'is_sticky_search_enabled')) {
		my $key_name = ref($h{'arg_name'}) eq 'ARRAY'
			? join '_', @{$h{'arg_name'}}
			: $h{'arg_name'};
		my $upref_value = $ENV{system_obj}->upref_by_alias(
			name	=> $key_name,
			value	=> $form_value,
			alias	=> $self->_upref_alias,
			no_save	=> 1,
		);

		if (sw_uriparam('reset_search') && $upref_value ne '' && !(defined $form_value)){
			$ENV{system_obj}->upref_by_alias(
				name	=> $key_name,
				value	=> '',
				alias	=> $self->_upref_alias,
				no_save	=> 1,
			);
			$upref_value = undef;
		}
		unless (defined $form_value){
			$form_value = $upref_value;
			$self->{got_sticky_search}++ if $upref_value ne ''
		}
	}
	$form_value = undef if $form_value eq '' && $h{is_undef};
	return $form_value;
}

## >> class
## =>>options=>[{key=>,color=>,string_id=>},..]
## => all=>{key=>,string_id=>} default {'','all'}
## < {filter=>[[key,value],...],data=>[[key,value],...]}
sub colored_options {
	my $class = shift;
	my %h = validate_params(@_,
		{options=>{type=>ARRAYREF},all=>{type=>HASHREF,optional=>1}}
	);
	my $ret = {filter=>[['',string('all')]],data=>[]};
	if ( $h{all} ) {
		$ret->{filter} = [[$h{all}->{key},string($h{all}->{string_id})]];
	}
	foreach my $e ( @{$h{options}} ) {
		push @{$ret->{filter}}, [$e->{key},string($e->{string_id})];
		my $value = $class->colored_status(
			color=>$e->{color},string_id=>$e->{string_id}
		);
		push @{$ret->{data}},[$e->{key},$value];
	}
	return $ret;
}##/colored_options


sub colored_status {
	my $class = shift;
	my %h = validate_params(@_,{color=>{type=>SCALAR},string_id=>{type=>SCALAR}});
	my ($status_icon,$color,$str) = (undef,$h{color},string($h{string_id}));
	if ( $h{color} eq 'red' ) {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id =>'alert_error');
	} elsif ( $h{color} eq 'gray') {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id =>'alert_inprogress');
		$color = 'black';
	} elsif ( $h{color} eq 'green') {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id =>'alert_ok');
	} elsif ( $h{color} eq 'yellow') {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id =>'alert_withconflict');
		$color = 'black';
	} else {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id =>'alert_inprogress');
	}

	return qq[$status_icon&nbsp;<font color="$color">].$str.q[</font>];
}##/colored_status

############################################################################
1;

__END__

=head1 SEE ALSO

B<HSPC::Site::Layout>

=cut

#  Privat Object Properties #
	_edit_out_afterform : content, will be published after edit_close()
	_edit_lists_move : publish JS-code for Up&Down buttons
	_edit_inputs : array with names for all inputs
	_edit_view_index :
	_edit_form_cols :

	url_range :

	mode : current mode, variants:
		[undef] : page is ready for any mode
		[viewinfo] : page->view_info_header() .... page->view_info_footer()
		[edit] : page->edit_open() .... page->edit_close()
		[grid] : page->edit_open() -> page->grid_open()...page->grid_close()

	grid: ref to hash with attributes
		rows : ref to array with grid rows, {cols=>,size=>,self_table=>,open=>,close=>}
		cols : ref to array current row data,  {
			align=>,content=>,bg_gray=>,width=>,cols=>,open=>,close=>
			}
		row_size : current row size, from 1

	grids : array of refs to grids for upper level

#  /Privat Object Properties #


