package HSPC::XMLAPI;
use strict;
use warnings;
no warnings qw/redefine uninitialized/;

use SOAP::Lite ();

use HSPC::API ();
use HSPC::API::ConverterFactory ();
use HSPC::Logger ();
use HSPC::ParamsValidate;

################################################################################

package HSPC::XMLAPI::Parser;
our @ISA = qw(SOAP::Parser);
## 'use base' doesn't work here, since SOAP::Parser doesn't compile well

use Encode ();

## disable UTF-8 flag on incoming data
sub decode {
	my $self = shift;

	my $result = $self->SUPER::decode(@_);
	my @to_cleanup = ($result);
	while (my $ref = shift @to_cleanup) {
		if (ref $ref eq 'ARRAY') {
			foreach (@$ref) {
				if (ref $_) {
					push @to_cleanup, $_;
				} else {
					Encode::_utf8_off($_);
				}
			}
		} elsif (ref $ref eq 'HASH') {
			foreach (keys %$ref) {
				if (ref $ref->{$_}) {
					push @to_cleanup, $ref->{$_};
				} else {
					Encode::_utf8_off($ref->{$_});
				}
			}
		} else {
			Encode::_utf8_off($ref);
		}
	}

	return $result;
}

################################################################################

package HSPC::XMLAPI::Serializer;
our @ISA = qw(SOAP::Serializer);
## 'use base' doesn't work here, since SOAP::Serializer doesn't compile well

## perform postprocessing on call results
sub envelope {
	my $self = shift;
	my $r = Apache->request;

	return $self->SUPER::envelope(@_)
		if $_[0] eq 'fault';

	my $logger = HSPC::Logger->new(category => 'xml_api');

	## return cached response, if any
	if (my $response_cached = $r->notes('api_response_cached')) {
		$logger->debug('API request served *from cache*');
		return $response_cached;
	}

	my ($method_key, $method_name) = (shift, shift);

	## convert only results corresponding to API calls
	my $results = \@_;
	if (my $class = $r->notes('api_class')) {
		$results = HSPC::API::ConverterFactory->get_converter(
			package => $class
		)->convert($results);
		unless ($self->uri) {
			(my $namespace = $class) =~ s!::!/!g;
			$self->uri($namespace);
		}
	}

	## append version back to namespace
	if (my $version = $r->notes('api_version')) {
		$self->uri($self->uri . '/' . $version);
	}

	## remove 'Response' suffix, if any, and add '_response' suffix
	substr($method_name, -8, 8, '')
		if substr($method_name, -8, 8) eq 'Response';
	$method_name .= '_response';

	my $envelope = $self->SUPER::envelope($method_key => $method_name,
		SOAP::Data->name($method_name . '_result')->value(@$results));

	## put response to cache, if needed
	if (my $request_cache = $r->pnotes('api_cache')) {
		HSPC::XMLAPI::_cache_set($request_cache->{key} => $envelope, $request_cache->{expires});
		$logger->debug("API request saved *to cache* by key [$request_cache->{key}]"
			. ($request_cache->{expires} && ", expires in $request_cache->{expires}"));
	}

	return $envelope;
}

## correctly handle dead calls
sub on_die {
	my $error = shift;

	## if error is an object or doesn't originate from HSPC::API namespace, proceed with it
	CORE::die $error
		if ref $error;

	## error is not an object, so log application error and produce an object
	my $logger = HSPC::Logger->new(category => 'xml_api');
	my $t = HSPC::Logger->no_trace;
	$logger->error($error);
	CORE::die HSPC::API->fault('ApplicationError',
		'Application error occurred, contact provider for details.');
}

################################################################################

package HSPC::XMLAPI;
use SOAP::Transport::HTTP ();
use base qw(SOAP::Transport::HTTP::Apache);

use SWsoft::protect;

use Apache ();
use Apache::Constants ();

use Socket qw(unpack_sockaddr_in);
use Storable qw(freeze thaw);

use HSPC::Config ();
use HSPC::Localization qw(get_languages set_language);
use HSPC::TaskExec ();
use HSPC::WebDB qw(select_run select_hashrows trans_counter);
use HSPC::WebSecurity ();

use HSPC::MT::Core ();
use HSPC::MT::Core::AbstractAccount ();
use HSPC::MT::Core::Constants qw(SW_HSP_ID SW_RESELLER SW_ACCOUNT_ACTIVE);
use HSPC::MT::TM::Constants ();
use HSPC::MT::TM::Queue ();
use HSPC::MT::GM;

## port used for local requests
use constant LOCAL_PORT => 8080;

## classes to set hooks in for cache reload for corresponding requests
use constant CACHE_HOOKS => {
	q|HSPC::API::HP 1.0 get_sellable_plan_list| => [qw/
		HSPC::MT::Billing::Promotion
		HSPC::MT::Billing::SellablePeriod
		HSPC::MT::BM::HP
		HSPC::MT::BM::HP::Type
		HSPC::MT::BM::Price
		HSPC::MT::EM::Store
	/],
	q|HSPC::API::HP 1.0 get_extended_plan_info| => [qw/
		HSPC::MT::AD::AbstractTemplate
		HSPC::MT::AD::AppCategory
		HSPC::MT::AD::Application
		HSPC::MT::AD::SiteApp
		HSPC::MT::AD::TemplateDep
		HSPC::MT::Billing::Promotion
		HSPC::MT::Billing::SellablePeriod
		HSPC::MT::BM::Attr
		HSPC::MT::BM::HP
		HSPC::MT::BM::HP::Attr
		HSPC::MT::BM::HP::AttrGroup
		HSPC::MT::BM::HP::Category
		HSPC::MT::BM::HP::DNSAssignment
		HSPC::MT::BM::HP::DNSSettings
		HSPC::MT::BM::HP::Fees
		HSPC::MT::BM::HP::PleskClientPerms
		HSPC::MT::BM::HP::PleskDomainPerms
		HSPC::MT::BM::HP::PleskPerms
		HSPC::MT::BM::HP::Questions
		HSPC::MT::BM::HP::Resource::Application
		HSPC::MT::BM::HP::Resource::LicClass
		HSPC::MT::BM::HP::Resource::PleskApplication
		HSPC::MT::BM::HP::Resource::QoS
		HSPC::MT::BM::HP::Resource::TLD
		HSPC::MT::BM::HP::Resource::TLDRegOption
		HSPC::MT::BM::HP::SBPlan
		HSPC::MT::BM::HP::Traffic
		HSPC::MT::BM::HP::Type
		HSPC::MT::BM::HPPromotion
		HSPC::MT::DM::RegOption
		HSPC::MT::DM::TLDconf
		HSPC::MT::EM::Store
		HSPC::MT::LM::AbstractLicenseClass
		HSPC::MT::LM::Plesk::LicClass::Abstract
		HSPC::MT::OM::TraffClass
		HSPC::MT::SB::Plan
	/],
};

## requests to cache with parameters list and cache expiration options:
## <parameter> => undef		- *any* value, *optional* parameter
## <parameter> => <value>	- *exact* value, *required* parameter
## expires					- refer to MySQL documentation on INTERVAL unit types
use constant CACHE_REQUESTS => {
	q|HSPC::API::HP 1.0 get_sellable_plan_list| => [
		{
			parameters	=> {
				type_id		=> undef,
				promo_id	=> undef,
			},
		},
		{
			parameters	=> {
				type_id		=> undef,
				promo_id	=> undef,
				account_id	=> undef,
				sb_sid		=> undef,
				sb_node		=> undef,
			},
			expires		=> '120 MINUTE',
		},
	],
	q|HSPC::API::HP 1.0 get_extended_plan_info| => [
		{
			parameters	=> {
				hp_sid		=> undef,
				promo_id	=> undef,
				period		=> undef,
				is_domain_hp=> undef,
			},
		},
		{
			parameters	=> {
				hp_sid		=> undef,
				promo_id	=> undef,
				account_id	=> undef,
				period		=> undef,
				for_trial	=> undef,
			},
			expires		=> '120 MINUTE',
		},
	],
};

## initialize SOAP server
our $SERVER = __PACKAGE__->new(compress_threshold => 10000);

## get record from cache by key or all records
sub _cache_get(;$) {
	my ($key) = validate_params(@_, [
		{ type => SCALAR, optional => 1 },
	]);

	## delete expired records
	select_run('DELETE FROM `xml_api_cache` WHERE `expires` IS NOT NULL AND `expires` < NOW()');

	if ($key) {
		## get one record by key
		my $row = select_hashrows('SELECT * FROM `xml_api_cache` WHERE `key` = ? LIMIT 1', $key);
		return $row && @$row ? $row->[0] : undef;
	} else {
		## get all records
		my $rows = select_hashrows('SELECT * FROM `xml_api_cache`');
		return $rows && @$rows ? @$rows : ();
	}
}

## add/update record in cache by key or delete record (value => undef)
sub _cache_set($$;$) {
	my ($key, $value, $expires) = validate_params(@_, [
		{ type => SCALAR },
		{ type => SCALAR | UNDEF },
		{ type => SCALAR | UNDEF, optional => 1 },
	]);

	if (defined $value) {
		## prepare date expression for expiration
		$expires &&= "DATE_ADD(NOW(), INTERVAL $expires)";
		$expires ||= 'NULL';

		## add/update record by key
		select_run("REPLACE INTO `xml_api_cache` VALUES (?, ?, $expires)", $key, $value);
	} else {
		## delete record
		select_run('DELETE FROM `xml_api_cache` WHERE `key` = ?', $key);
	}
}

## clear cache completely
sub _cache_cleanup {
	## clear cache
	select_run('DELETE FROM `xml_api_cache`');

	my $logger = HSPC::Logger->new(category => 'xml_api');
	$logger->debug('API cache cleanup succeeded');
}

sub _cache_update_on_request {} ## no need to do anything

sub _cache_update_on_modify(@) {
	## list updated classes
	my %mt = map {$_ => 1} @_;

	## find cacheable requests to reload
	my @requests_cacheable = grep {
		grep {$mt{$_}} @{&CACHE_HOOKS->{$_}}; ## does request rely on class that caused event?
	} keys %{&CACHE_HOOKS};

	## add task for request re-caching unless such already exists
	foreach my $request_cacheable (@requests_cacheable) {
		my $task_options = {
			class		=> __PACKAGE__,
			method		=> '_cache_update_on_modify_task',
			param		=> [$request_cacheable],
			priority	=> 255,
			descr		=> "API re-caching ($request_cacheable)",
			mutex		=> 'xml_api_cache_update',
			redo		=> 1,
			timeout		=> 7200,
		};

		## skip request if task for re-caching already exists
		next
			if @{HSPC::MT::TM::Queue->list(attributes => ['task_id'], filter => {
				description	=> $task_options->{descr},
				mutex		=> $task_options->{mutex},
				status		=> HSPC::MT::TM::Constants::HSPC_TM_TASK_STATUS_PENDING,
			})};

		## add task
		my $logger = HSPC::Logger->new(category => 'xml_api');
		HSPC::TaskExec::task_add(%$task_options)
			and $logger->debug("$task_options->{descr} scheduled due to changes in "
				. (join (', ', @_) || 'database'));
	}
}

sub _cache_update_on_modify_task($;%) {
	my (undef, $request_cacheable) = @_;

	## extract request details
	my ($class, $version, $method) = split(' ', $request_cacheable);

	## temporary storage for "fake" session and security objects (per vendor)
	my $obj_cache = {};

	## look through cached parameters
	foreach my $request_cache_rec (_cache_get()) {
		my $request_pos = index($request_cache_rec->{key}, $request_cacheable);
		next
			unless $request_pos > $[;

		unless (defined $request_cache_rec->{expires}) {
			## expiration flag is *not* set, so need to update cache

			## recreate system object to apply new language options
			$ENV{system_obj} = HSPC::WebSystem->new(reinit_ok => 1);
			delete $ENV{HSPC_CU_SKIN_BID};

			## prepare session and security objects for vendor
			my ($vendor_id, $lang) = split(' ', $request_cache_rec->{key}, 3);
			unless ($obj_cache->{$vendor_id}) {
				HSPC::API->session_open({account_id => $vendor_id});
				$obj_cache->{$vendor_id}->{$_} = $ENV{$_}
					foreach qw/session security_obj/;
			} else {
				$ENV{$_} = $obj_cache->{$vendor_id}->{$_}
					foreach qw/session security_obj/;
			}

			## set language options
			$ENV{session}->{LANG} = $lang;
			set_language($ENV{system_obj}->locale_lang);

			## extract parameters frozen in request cache key
			substr(my $parameters_str = $request_cache_rec->{key}, $[,
				length($request_cacheable) + $request_pos + 1) = '';
			my @parameters = $parameters_str ? thaw($parameters_str) : ();

			## emulate remote call
			my $r = Apache->request;
			$r->notes(api_class => $class);
			$r->notes(api_version => $version);
			my @results = eval {
				$class->$method(@parameters);
			};

			## put response to cache on success
			my $logger = HSPC::Logger->new(category => 'xml_api');
			unless ($@) {
				my $envelope = $SERVER->serializer->prefix('s')->envelope(
					response => $method, @results,
				);
				$logger->debug("API ($version) request ($class->$method) *re-caching* succeeded by key [$request_cache_rec->{key}]");
				_cache_set($request_cache_rec->{key} => $envelope);
				next;
			} else {
				$logger->warn("API ($version) request ($class->$method) *re-caching* failed: $@");
			}
		}

		## expiration flag is set, so cache update is not worth it,
		## or remote call emulation didn't succeed, so remove request from cache
		_cache_set($request_cache_rec->{key} => undef);
	}
}

my @_database_mt_schedule = ();
my %_database_orig_methods = ();

sub _cache_update_on_database {
	my $modify_method = shift;

	no strict 'refs';

	## restore original methods
	*{"HSPC::WebDB::$_"} = $_database_orig_methods{$_}
		foreach qw/_db_commit _db_rollback/;

	if ($modify_method eq '_db_commit') {
		my $logger = HSPC::Logger->new(category => 'xml_api');
		$logger->debug('API requests are about to *re-cache* on transaction commit');

		## force cache update for all classes modified during transaction
		_cache_update_on_modify(@_database_mt_schedule);
	}

	@_database_mt_schedule = ();
	%_database_orig_methods = ();
}

sub _cache_update_on_database_schedule {
	my $mt = shift;

	## if no overridden methods yet, set hooks on commit/rollback - once per transaction!
	foreach my $modify_method (qw/_db_commit _db_rollback/) {
		no strict 'refs';

		next
			if $_database_orig_methods{$modify_method};

		## store reference to original method
		$_database_orig_methods{$modify_method} =
			*{"HSPC::WebDB::$modify_method"}{CODE};

		## override original method with hook forcing cache update
		*{"HSPC::WebDB::$modify_method"} = sub {
			my $result = $_database_orig_methods{$modify_method}->(@_);
			_cache_update_on_database($modify_method);
			return $result;
		};
	}

	## schedule cache update on class
	push @_database_mt_schedule, $mt;
}

## set hooks for cache update and cleanup
BEGIN {
	## clean cache on startup of parent process
	_cache_cleanup()
		if getpgrp(0) == $$;

	## clean cache on server configuration reload
	Apache->server->register_cleanup(\&_cache_cleanup);

	## get list of classes to hook in *without* duplicates
	my @mt_cacheable = keys %{{
		(map {$_ => 1} map @{&CACHE_HOOKS->{$_}}, keys %{&CACHE_HOOKS})}
	};

	foreach my $modify_method (qw/save delete/) {
		foreach my $mt (@mt_cacheable) {
			no strict 'refs';

			next
				if ${"${mt}::_cache_update_on_${modify_method}_set"};

			## ensure packages are preloaded first so methods already exist
			(my $package = $mt . '.pm') =~ s!::!/!g;
			eval {
				require $package;
			} or next;

			## store reference to original method (if any)
			my $orig_method = $mt->can($modify_method) || next;
			## override original method with hook forcing cache update
			*{"${mt}::$modify_method"} = sub {
				my $result = $orig_method->(@_);

				if (trans_counter()) {
					## transaction active, so do cache update only on commit
					_cache_update_on_database_schedule($mt);
				} else {
					## no transaction, so do cache update now
					_cache_update_on_modify($mt);
				}

				return $result;
			};

			## set flag for overridden method (to override only once)
			${"${mt}::_cache_update_on_${modify_method}_set"} = 1;
		}
	}
}

sub _cache_preload() {
	## list all vendors to preload cache for
	my $vendors = HSPC::MT::Core::AbstractAccount->list(
		attributes	=> ['id'],
		filter		=> { type => SW_RESELLER, status => SW_ACCOUNT_ACTIVE },
	);
	unshift @$vendors, [SW_HSP_ID];

	## add task for cache preload
	HSPC::TaskExec::task_add(
		class		=> __PACKAGE__,
		method		=> '_cache_preload_task',
		param		=> [map {$_->[0]} @$vendors],
		priority	=> 128,
		descr		=> 'Preload API cache',
		mutex		=> 'xml_api_cache_preload',
		redo		=> 1,
		timeout		=> 1800,
	)
}

sub _cache_preload_task($@) {
	shift; ## ignore class name

	my $logger = HSPC::Logger->new(category => 'xml_api');
	$logger->info('API requests are being *pre-cached* on startup');

	## preload cache for all available vendors
	my $client = SOAP::Lite->proxy('http://localhost:' . LOCAL_PORT . '/hspc/xml-api');
	$client->on_fault(sub {
		$logger->warn('API request *pre-caching* failed: '
			. (ref $_[1]
				? 'SOAP fault: ' . $_[1]->faultcode . ': ' . $_[1]->faultstring
				: $_[0]->transport->is_success
					? 'bad response: ' . $_[1]
					: 'transport error: ' . $_[0]->transport->status));
	});
	foreach my $vendor_id (@_) {
		my $cfg = HSPC::MT::GM->find_provider_config(id => $vendor_id);
		foreach my $lang_id (map { $_->{lang} } @{$cfg->langs}) {
			## initialize connection with backend
			my $sid = eval {
				$client->ns('HSPC/API/1.0')->session_open({
					account_id	=> $vendor_id,
					lang_id		=> $lang_id,
				})->result->{session_id};
			} or next;
			$client->transport->http_request->header('HSPC-SID' => $sid);

			## request extended info for each of sellable plans (emulate cacheable requests)
			$client->ns('HSPC/API/HP/1.0');
			$client->get_extended_plan_info;
			eval { $client->get_extended_plan_info({hp_sid => $_->{series_key}}) }
				foreach eval { @{$client->get_sellable_plan_list->result->{plan_list}} };
		}
	}

	$logger->info('API requests *pre-caching* done');
}

## preload cache on startup of parent process
_cache_preload()
	if getpgrp(0) == $$;

## set predispatching hook
$SERVER->on_dispatch(sub {
	my $request = shift;

	## extract namespace, method name and parameters list
	my ($namespace, $method, @parameters)
		= ($request->namespaceuriof, $request->dataof->name, $request->paramsin);
	## try to extract required version from namespace
	$namespace =~ s!^(HSPC/API.*?)/(\d+\.\d+)$!$1!;
	my ($class, $version) = ($namespace, $2);
	## convert namespace to class name
	for ($class) { s!^/|/$!!g; s!\.!!g; s!/!::!g }

	## initialize required objects and structures
	$ENV{cookies_in} = {has_cookies => 1};
	$ENV{system_obj} = HSPC::WebSystem->new;

	## extract HSPC-SID from HTTP or SOAP headers and initialize session and security object
	my $r = Apache->request;
	my $sid = $r->headers_in->get('HSPC-SID')
		|| $request->header && $request->header->{'HSPC-SID'};
	if ($sid) {
		$ENV{session} = HSPC::Session::checkSession(undef, undef, undef, $sid);
		$ENV{security_obj} = HSPC::WebSecurity->factory(
			SESSION		=> $ENV{session},
			LANGUAGE	=> $HSPC::Localization::LOCALE{local_lang},
		);
		$ENV{system_obj}->restore_upref;
		$ENV{security_obj}->set_account($ENV{system_obj}->upref_account(
			tool_id => 'xml-api',
		));
		$ENV{security_obj}->check;
		set_language($ENV{system_obj}->locale_lang);
	}

	## distinguish between local and remote requests
	my ($port) = (unpack_sockaddr_in($r->connection->local_addr));
	if ($port eq LOCAL_PORT) {
		## local request, so force server to consider all classes safe
		$SERVER->dispatch_with->{$namespace} = $class;
		$r->notes(api_local_call => 1);
	} else {
		## remote request, so require authentication (i.e. HSPC-SID)
		if ($class ne 'HSPC::API') {
			## deny classes not from API namespace
			die HSPC::API->fault('ClassDenied', "Denied access to class ($class).")
				unless index($class, 'HSPC::API') == $[;
			## deny methods starting with underscore (private methods)
			die HSPC::API->fault('MethodDenied', "Denied access to method ($method).")
				if substr($method, $[, 1) eq '_';
		}
	}

	## explicitly deny this package from dispatching
	delete $SERVER->dispatch_with->{'HSPC/XMLAPI'};

	## perform initialization tasks for API requests
	if (index($class, 'HSPC::API') == $[) {
		if ($class ne 'HSPC::API') {
			## require session ID on either HTTP or SOAP headers
			die HSPC::API->fault('AuthzError', 'Session ID is missing from both HTTP and SOAP headers or wrong.')
				unless $sid;
			## require valid instance of security object (i.e. authentication)
			die HSPC::API->fault('AuthenRequired', 'Authentication required.')
				unless $ENV{security_obj}->is_signed_in
					&& $ENV{security_obj}->account_no;
		}

		## log call
		my $logger = HSPC::Logger->new(category => 'xml_api');
		$logger->info("API ($version) request: ${class}->$method("
			. $class->_dump(method => $method, value => [$request->paramsin])
			. ')');

		## check required API version
		die HSPC::API->fault('WrongAPIVersion', 'API version is missing from namespace.')
			unless $version;
		eval {
			$class->VERSION($version);
		} or die HSPC::API->fault('WrongAPIVersion', "API version ($version) is not supported.");

		## tell SOAP server to always dispatch to API methods
		$SERVER->dispatch_with->{$namespace} = $class
			unless $class eq 'HSPC::API::Core';

		## tell serializer current class name
		$r->notes(api_class => $class);
		$r->notes(api_version => $version);
	}

	## check if method in request *should be* cached and named parameters are used, if any
	if (
		my $request_cache_schemes = &CACHE_REQUESTS->{"$class $version $method"}
			and !@parameters || ref $parameters[0] eq 'HASH'
	) {
		## throw away undefined request parameters
		my %parameters = @parameters
			? (map {$_ => $parameters[0]->{$_}}
				grep {defined $parameters[0]->{$_}} keys %{$parameters[0]})
			: ();

		## look through possible caching schemes
		foreach my $request_cache_scheme (@$request_cache_schemes) {
			my ($parameters_req, %parameters_seen) = ($request_cache_scheme->{parameters});

			## skip if unexpected parameter found or value does not patch the pattern
			next
				if grep {
					$parameters_seen{$_} = 1;
					!exists $parameters_req->{$_}
						|| $parameters_req->{$_} ne $parameters{$_}
							&& defined $parameters_req->{$_};
				} keys %parameters;

			## skip if required parameters are missing
			next
				if grep {
					!$parameters_seen{$_}
						&& defined $parameters_req->{$_};
				} keys %$parameters_req;

			## request scheme found, prepare key for cache
			local $Storable::canonical = 1;
			my $request_cache_key = join(' ',
				$ENV{security_obj}->account_no,
				$ENV{system_obj}->locale_lang,
				$class,
				$version,
				$method,
				(%parameters ? freeze(\%parameters) : ()));

			my $logger = HSPC::Logger->new(category => 'xml_api');
			$logger->debug("API request is *cache driven*, will be stored/restored by key [$request_cache_key]");

			## tell serializer cache key to use for request
			$r->pnotes(api_cache => {
				key		=> $request_cache_key,
				expires	=> $request_cache_scheme->{expires},
			});

			## check if response is cached
			if (my $response_cached = _cache_get($request_cache_key)) {
				## put response to request notes for serializer to use
				$r->notes(api_response_cached => $response_cached->{value});

				## tell SOAP handler to dispatch to "fake" method (just empty)
				$SERVER->dispatch_with->{'HSPC/XMLAPI'} = __PACKAGE__;
				return 'HSPC/XMLAPI', '_cache_update_on_request';
			}

			last;
		}
	}

	## return nothing to let superclass proceed with default parameters
	return $namespace, $method;
});

## set custom parser
$SERVER->deserializer->parser(HSPC::XMLAPI::Parser->new);

## set custom serializer (see definition above)
$SERVER->serializer(HSPC::XMLAPI::Serializer->new);
delete $SERVER->serializer->typelookup->{anyURI};

## ignore SOAPAction HTTP headers (obsolete in SOAP 1.2)
$SERVER->on_action(sub { 1 });

## set dispatching directory and default packages to dispatch to
$SERVER->dispatch_to(
	## packages location (lookup by path to already loaded HSPC::API)
	substr($INC{'HSPC/API.pm'}, $[, -length('/HSPC/API.pm')),
);

sub handler {
	my $r = shift;

	## temporary override 'die'
	my $die_handler = *CORE::GLOBAL::die;
	*CORE::GLOBAL::die = *HSPC::XMLAPI::Serializer::on_die;

	## handle SOAP request
	my $results = $SERVER->SUPER::handler($r, @_);

	## restore original 'die'
	*CORE::GLOBAL::die = $die_handler || *CORE::die;

	return $results;
}

1;

=pod

=head1 STATUS

B<XMLAPI_Implement> - B<Parallels Business Automation - Standard XML API> Gate I<Implementation>.

=head1 IMPLEMENTATION

=head2 PROTOCOL

B<Parallels Business Automation - Standard XML API> Gate is based on B<SOAP> protocol, currently mantained
by B<World Wide Web Consortium> at http://www.w3c.org and supported by most
of modern programming languages as framework for messages exchange and
remote method calls.

B<Parallels Business Automation - Standard XML API> Gate is implemented as B<mod_perl> handler and inherits
from C<SOAP::Transport::HTTP::Apache>, i.e. based on functionality provided by
C<SOAP::Lite> module available from B<CPAN>. Please, refer to C<SOAP::Lite>
documentation for general information and to this file for implementation
details and examples.

Module namespaces are package names with "::" replaced by "/" - see examples.

=head2 SECURITY

There are two different strategies used by B<Parallels Business Automation - Standard XML API> Gate
in defining security requirements:

=over 4

=item *

for requests coming from B<local> machine (directly to backend server without
involving frontend, i.e. originating from the same address space and
using http://localhost:8080 as B<Parallels Business Automation - Standard XML API> Gate proxy URL);

=item *

for requests coming from B<remote> machines (using frontend for proxying
requests to backend).

=back

Safe packages and methods:

=over 4

=item *

B<local> requests: I<all> packages are considered safe and I<all> their methods
are public;

=item *

B<remote> requests: only packages with namespace starting with C<HSPC/API>
are considered safe.

=back

Authentication and sessions handling:

=over 4

=item *

B<local> requests: authentication by password is possible, but not required,
enough for successful authentication are account number or server name;

=item *

B<remote> requests: authentication by password is required.

=back

Authentication is done with call to C<session_open()> interface in C<HSPC/API>
namespace and relies on functionality provided by B<Security Manager>.

Interfaces in C<HSPC/API> namespace:

=over 4

=item C<session_open()>

B<Parameters:> C<account_no>, C<server_name>, C<e-mail>, C<password>

Performs authentication with given parameters (I<required> for B<remote> requests
and I<optional> for B<local>, except C<account_no> or C<server_name>)
and initialize session.

If C<account_no> is given as C<0>, empty or not given, first account which user
has roles in is chosen automatically, but in this case C<e-mail> and C<password>
must be given as well.

If C<server_name> is given and C<account_no> is given as C<0>, empty or not given,
C<account_no> is being located by vendor's server name located in C<server_name>
parameter.

B<Returns> either unique value to be used as C<HSPC-SID> in further requests
(see examples of clients) or SOAP fault envelope with error message.

=item C<session_close()>

Performs cleanup of session identified by C<HSPC-SID> header.

B<Returns> C<undef> or SOAP fault envelope with error message.

=back

=head2 CONFIGURATION

B<Parallels Business Automation - Standard XML API> Gate intended for requests from both B<local> and
B<remote> machines is preconfigured at C</hspc/xml-api> location.

=over 4

=item backend

C</etc/hspcd/conf/hspc_xml-api.conf>:

 <Location /hspc/xml-api>
 	SetHandler perl-script
 	PerlHandler HSPC::XMLAPI
 	Order Allow,Deny
 	Allow from all
 </Location>

=item frontend

C</etc/httpd/conf/hspc_frontend.conf>:

 <Location /hspc/xml-api>
 	Order Deny,Allow
 	Deny from all
 </Location>
 ...
 <VirtualHost _default_:443>
 	...
 	Include "conf/hspc_ssl.conf"
 	...
 </VirtualHost>

C</etc/httpd/conf/hspc_ssl.conf>:

 <Location /hspc/xml-api>
 	Order Allow,Deny
 	# Replace 'none' with list of allowed hosts
 	# to open XML API Gate for remote requests
 	Allow from none
 </Location>

Security limitation is set by explicitly allowing C</hspc/xml-api> location for
B<HTTPS> connections (no hosts allowed by default!) and denying for B<HTTP>
connections, so that plain text SOAP envelopes couldn't be read by intruders.

=back

B<Parallels Business Automation - Standard XML API> Gate could be opened at another locations as well by
configuring backend and frontend server in the same fashion as described above,
i.e. by adding more C<Location> blocks to backend and frontend servers'
configurations.

=head2 SERVERS

Exported methods of packages providing API through B<Parallels Business Automation - Standard XML API> Gate
should rely on the following rules:

=over 4

=item *

in order to be available for remote requests, package should be named including
C<HSPC::API::> prefix and have its version set:

 our $VERSION = 1.0;

=item *

first parameter of each call to exported method is B<always> package name,
not blessed reference or whatever;

=item *

C<$ENV{session}> is valid I<only> for requests including session ID returned
by C<session_open()> call, i.e. could be valid for B<local> and always valid
for B<remote> requests;

=item *

C<$ENV{security_obj}> is valid I<only> for requests including session ID
and contains valid account and user IDs identified by call to C<session_open()>;

=item *

C<die> with error message to immediately return it in SOAP fault envelope with
message as description, using the call like this:

 ## return fault with:
 ## - error code 'ErrorCode'
 ## - error message
 die HSPC::API->fault('ErrorCode', 'Error description.');

Notes for PHP Store:

=over 4

=item *

If error code B<does> start with C<User> prefix, its description is shown to
PHP Store visitor, so it B<must> be localized:

 die HSPC::API->fault('UserPassword', string('passwords_do_not_match'));

=item *

If error code B<does not> start with C<User> prefix, its description is not shown
to PHP Store visitor and is only logged to vendor's local log file, so it
B<must not> be localized:

 die HSPC::API->fault('AuthenRequired', 'Authentication required.');

=back

=item *

feel free to return any data structures that you can theoretically serialize
to XML - and do not expect an object to arrive at remote side by just returning
its blessed reference (guess why it's just ridiculous);

=item *

use C<SOAP::Data> functionality for fine tuning serialization of results.

=back

=head3 Examples

=over 4

=item *

C<HSPC/Test.pm> (local requests):

 package HSPC::Test;
 use strict;
 use Data::Dumper;
 ## returns dump of parameters list, including class name
 sub method {
 	return Dumper(\@_);
 }
 1;

=item *

C<HSPC/API/Test.pm> (remote requests):

 package HSPC::API::Test;
 use strict;
 our $VERSION = 1.0;
 ## gets/sets parameter with key passed as a parameter
 sub param {
 	my (undef, $key, $value) = @_;
 	return defined $value
 		? $ENV{session}->{$key} = $value
 		: $ENV{session}->{$key};
 }
 1;

=back

=head2 CLIENTS

In order to initialize stateful communication with B<Parallels Business Automation - Standard XML API> Gate,
first call C<session_open()> in C<HSPC/API> namespace to receive C<HSPC-SID>
value and then add C<HSPC-SID> to either B<HTTP> or B<SOAP> headers to each
request before sending SOAP envelope.

=head3 Examples

=over 4

=item *

C<local.pl>:

 use SOAP::Lite;
 use strict;
 my $result = SOAP::Lite
 	->proxy('http://localhost:8080/hspc/xml-api') ## Gate URL
 	->ns('HSPC/Test') ## package namespace
 	->method ## method name
 		('param1', {param2 => 'test', param3 => [1, 2, 3]}, 0); ## parameters
 print $result->fault
 	? 'Fault: ' . $result->faultstring
 	: 'Result: ' . $result->result;

=item *

C<local.php>:

 <?
 require_once('nusoap.php');
 $client = new soapclient('http://localhost:8080/hspc/xml-api'); // Gate URL
 $result = $client->call(
 	'method', // method name
 	array ("param1", array ("param2" => "test", "param3" => array (1, 2, 3)), 0), // parameters
 	'HSPC/Test' // package namespace
 );
 if ($client->fault)
 	die("Fault: {$client->faultstring}");
 echo $result;
 ?>

=item *

C<remote.pl>:

 use strict;
 use SOAP::Lite;
 my $client = SOAP::Lite
 	->proxy('https://192.168.0.100/hspc/xml-api')
 	->on_fault(sub { die 'Error: ' . (ref $_[1]
 		? $_[1]->faultcode . ': ' . $_[1]->faultstring
 		: $_[0]->transport->is_success ? $_[1] : $_[0]->transport->status) });
 ## pass authentication and receive session ID
 my $sid = $client->ns('HSPC/API/1.0')->session_open({
 	email => 'email@provider.com', password => 'password'
 })->result->{session_id};
 ## put session ID to outgoing requests' HTTP headers
 $client->transport->http_request->header('HSPC-SID' => $sid);
 ## make session-dependent calls
 $client->ns('HSPC/API/Test/1.0');
 $client->param('key' => 'value');
 print $client->param('key')->result;
 $client->ns('HSPC/API/1.0')->session_close;

=item *

C<remote.php>:

 <?
 require_once('nusoap.php');
 $client = new soapclient('https://192.168.0.100/hspc/xml-api'); // Gate URL
 ## pass authentication and receive session ID
 $sid_result = $client->call('session_open', array (
 	array ('email' => 'root@23.c.sw.ru', 'password' => '1q2w3e')
 ), 'HSPC/API/1.0');
 $sid = $sid_result['session_id'];
 if ($client->fault)
 	die("Fault: {$client->faultstring}");
 ## put session ID to outgoing requests' SOAP headers
 $client->setHeaders("<HSPC-SID>$sid</HSPC-SID>");
 ## make session-dependent calls
 $client->call('param', array ('key', 'value'), 'HSPC/API/Test/1.0');
 if ($client->fault)
 	die("Fault: {$client->faultstring}");
 echo $client->call('param', array ('key'), 'HSPC/API/Test/1.0') . "\n";
 if ($client->fault)
 	die("Fault: {$client->faultstring}");
 $client->call('session_close', undef, 'HSPC/API/1.0');
 if ($client->fault)
 	die("Fault: {$client->faultstring}");
 ?>

=back

Examples for remote calls imply B<Parallels Business Automation - Standard> to be accessible at B<192.168.0.100> -
replace it with IP address of real installation for examples to work.

=head1 AUTHOR

Nikita Dedik E<lt>ndedik@swsoft.comE<gt>

=head1 COPYRIGHT

Copyright 2009 Parallels

=cut
